/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode;
import BOIException.ParameterException;
import BOIException.ErrorException;
import BOIException.WarningException;
import BOIFile.LoadFile;


/**
 * Main class of BOICode application. BOICode is a JPEG2000 coder that allows multiple parameterizations.
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class BOICode{

	/**
	 * Main method of BOIcode application. It takes program arguments, loads image and runs BOI coder.
	 *
	 * @param args an array of strings that contains program's parameters
	 */
	public static void main(String[] args){
		//Parse arguments
		CoderParser parser = null;
		try{
			parser = new CoderParser(args);
		}catch(ErrorException e){
			System.out.println("RUN ERROR:");
			e.printStackTrace();
			System.exit(1);
		}catch(ParameterException e){
			System.out.println("ARGUMENTS ERROR: " +  e.getMessage());
			System.exit(2);
		}

		//Image load
		String imageFile = parser.getImageFile();
		LoadFile image = null;
		try{
			if(imageFile.endsWith(".raw") || imageFile.endsWith(".img") || imageFile.endsWith(".RAW") || imageFile.endsWith(".IMG")){
				int[] imageGeometry = parser.getImageGeometry();

				//Check parameters of image geometry
				if((imageGeometry[0] <= 0) || (imageGeometry[1] <= 0) || (imageGeometry[2] <= 0)){
					throw new WarningException("Image dimensions in \".raw\" or \".img\" data files must be positive (\"-h\" displays help).");
				}
				if((imageGeometry[3] < 0) || (imageGeometry[3] > 7)){
					throw new WarningException("Image type in \".raw\" or \".img\" data must be between 0 to 7 (\"-h\" displays help).");
				}
				if((imageGeometry[4] != 0) && (imageGeometry[4] != 1)){
					throw new WarningException("Image byte order  in \".raw\" or \".img\" data must be 0 or 1 (\"-h\" displays help).");
				}
				if((imageGeometry[5] != 0) && (imageGeometry[5] != 1)){
					throw new WarningException("Image RGB specification in \".raw\" or \".img\" data must be between 0 or 1 (\"-h\" displays help).");
				}
				image = new LoadFile(imageFile, imageGeometry[0], imageGeometry[1], imageGeometry[2], imageGeometry[3], imageGeometry[4], imageGeometry[5] == 0 ? false: true);
			}else{
				image = new LoadFile(imageFile);
			}
		}catch(WarningException e){
			System.out.println("IMAGE LOAD ERROR: " + e.getMessage());
			System.exit(3);
		}
		int zSize = image.getImage().length;

		//Get arguments from parser
		String COutFile = parser.getCOutFile();
		if(COutFile.compareTo("") == 0){
			int dotPos = imageFile.lastIndexOf(".");
			if(dotPos >= 0){
				COutFile = imageFile.substring(0, imageFile.lastIndexOf("."));
			}else{
				COutFile = imageFile;
			}
		}
		int CFileHeader = parser.getCFileHeader();
		int CFileProgression = parser.getCFileProgression();
		int CTType = parser.getCTType();
		int[] CTComponents = parser.getCTComponents();
		int[] WTTypes = parser.getWTTypes();
		int[] WTLevels = parser.getWTLevels();
		int[] QTypes = parser.getQTypes();
		int QGuardBits = parser.getQGuardBits();
		int[] BDBlockWidths = parser.getBDBlockWidths();
		int[] BDBlockHeights = parser.getBDBlockHeights();
		int[] BDPrecinctWidths = parser.getBDPrecinctWidths();
		int[] BDPrecinctHeights = parser.getBDPrecinctHeights();
		int LCType = parser.getLCType();
		int LCTargetNumLayers = parser.getLCTargetNumLayers();
		int LCForceSingleLayer = parser.getLCForceSingleLayer();
		int LCTargetNumBytes = parser.getLCTargetNumBytes();
		boolean[] CVerbose = parser.getCVerbose();
		int CThreadNumber = parser.getCThreadNumber();
		parser = null;

		//BOI coder
		Coder BOIC = new Coder(image.getImage(), image.getTypes(), image.getRGBComponents());
		image = null;
		try{
			BOIC.setParameters(COutFile, CFileHeader, CFileProgression, CTType, CTComponents, WTTypes, WTLevels, QTypes, QGuardBits, BDBlockWidths, BDBlockHeights, BDPrecinctWidths, BDPrecinctHeights, LCType, LCTargetNumLayers, LCForceSingleLayer, LCTargetNumBytes, CVerbose, CThreadNumber);
		}catch(ParameterException e){
			System.out.println("PARAMETERS ERROR: " + e.getMessage());
			System.exit(4);
		}
		try{
			BOIC.run();
		}catch(ErrorException e){
			System.out.println("RUN ERROR:");
			e.printStackTrace();
			System.exit(5);
		}catch(Exception e){
			System.out.println("RUN LANGUAGE ERROR:");
			e.printStackTrace();
			System.exit(6);
		}
	}
}