/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.Code;
import BOIException.WarningException;
import BOIStream.ByteStream;


/**
 * This class receives a code-block samples and encodes them producing a JPEG2000 bytestream. Typically it is used in the following way:<br>
 * constructor(receiveing code-block parameters)<br>
 * runAll<br>
 * get functions<br>
 * You can also use this class to increasingly encode the coding passes: you only have to use the codeNextCodingPass function as times as coding passes exist.<br>
 *
 * ATTENTION: In the framework of the JPEG2000 many different code-blocks have to be encoded. It is more efficient to create only one object of this class and change the code-block:<br>
 * constructor(void)<br>
 * swapCodeBlock<br>
 * runAll<br>
 * getFunctions<br>
 * swapCodeBlock<br>
 * runAll<br>
 * getFunctions<br>
 * ...<br>
 * (of course, the function runAll can be substituted for multiple call of the codeNextCodingPass function) If this class is used in this last way it is easy to incorporate a multithreading support for the encoding of code-blocks.<br>
 *
 * Since version 1.3 this class uses an optimized way to calculate coefficient contexts. Although it achieves good results in terms of computational complexity, the source code is not very understandable (when computing the coefficient contexts). File BPECoder.java.1_2 contains the 1.2 version of this class, that has a more understandable source code. Apart of this, both classes are exactly equal.
 * Since version 1.4, this class estimate distortion decreases rather than actually compute them.
 *
 * @author Francesc Auli-Llinas
 * @version 1.4
 */
public class BPECoder{

	/**
	 * ByteStream generated by the fractional bit plane coding.
	 * <p>
	 * The content must be understood as a ByteStream.
	 */
	ByteStream CDByteStream = null;

	/**
	 * Original block samples (index meaning [y][x]).
	 * <p>
	 * All values allowed.
	 */
	int[][] blockSamples = null;

	/**
	 * Block height.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int ySize;

	/**
	 * Block width.
	 * <p>
	 * Negative values are not allowed for this field.
	 */
	int xSize;

	/**
	 * Image WT subband.
	 * <p>
	 * Valid values: 0(LL), 1(HL), 2(LH), 3(HH).<br>
	 * LL HL<br>
	 * LH HH
	 */
	int subband;

	/**
	 * Coding gain (needed for RD operations).
	 * <p>
	 * It is commonly the L2norm of the wavelet filter bank.
	 */
	float RDAdjust;

	/**
	 * Most Significance Bit Plane of the codeblock.
	 * <p>
	 * Negative values not allowed.
	 */
	int MSBPlane;

	/**
	 * Most Significance Bit Plane of the subband.
	 * <p>
	 * Negative values not allowed.
	 */
	int subbandMSBPlane;

	/**
	 * Error decrease of each truncation point (index meaning [bitPlane][0(SPP),1(MRP),2(CP)] -bitPlane==0 is MSBP-).
	 * <p>
	 * Negative values not allowed.
	 */
	float[][] errorsDec = null;

	/**
	 * Number of bytes of each coding pass (index meaning [bitPlane][0(SPP),1(MRP),2(CP)] -bitPlane==0 is MSBP-).
	 * <p>
	 * Negative values not allowed.
	 */
	int[][] lengths = null;

	/**
	 * Definition in {@link BOIDecode.Decode.BPEDecoder#intProcessing}
	 */
	boolean intProcessing;

	/**
	 * Map where the contexts of the coefficients are stored. This map is used to extract the context of coefficient when they have to be encoded. It works in the following way: at the begining all the contextMap is initialized to 0. When a coefficient is found significant, it updates all its neighboors to the next context status and so on. Thanks to this mechanism, less operations have to be performed when a looking for the context of a coefficient although the source code is not so... clean (see function updateMap).
	 * <p>
	 * The value means the context of the sample. Valid contexts are: 0 - 8 when the coefficient is still non-significant, 15 - 17 when the coefficient is significant and it is going to be refined. These values can be positive or negative to indicate if they have to be visited by the significancePropoagationPass or cleanupPass functions. SPP only visits coefficients with contexts > 0 and <= 8, CP only visits coefficients with contexts >= -8 <= 0.
	 */
	private byte[][] contextMap = null;

	/**
	 * Variable used to calcuate the number of significance/refinement coefficients encoded in the current coding pass.
	 * <p>
	 * Negative values not allowed.
	 */
	private long numSIGS;

	/**
	 * Current bit plane encoded.
	 * <p>
	 * 0 is the Less Significant Bit Plane.
	 */
	private int currentBitPlane;

	/**
	 * Current coding pass encoded.
	 * <p>
	 * 0 is the SPP, 1 is MRP, and 2 is CP.
	 */
	private int currentCodingPass;

	/**
	 * Current mask of the bit plane.
	 * <p>
	 * The content is 2^currentBitPlane
	 */
	private int currentBitPlaneMask = 0;

	/**
	 * Lookup table with pre-computed distortion decreases for significance bits (DWT). Each index means [MSBP of codeblock][significanceBitPlane, starting from the MSBP to 0].
	 * <p>
	 * Distortion estimators for bitplane image coding
	 */
	static final double[][] DDsigDWT =
		{{2.2},
		{8.309,2.2},
		{32.006,8.516,2.2},
		{124.285,33.235,8.6,2.2},
		{484.79,130.276,33.764,8.645,2.2},
		{1896.367,512.088,132.941,34.065,8.673,2.2},
		{7432.682,2016.805,524.441,134.495,34.261,8.693,2.2},
		{29173.83,7954.264,2071.793,531.763,135.525,34.398,8.707,2.2},
		{114675.393,31405.77,8193.412,2104.735,536.678,136.261,34.5,8.718,2.2},
		{451529.261,124106.269,32430.478,8337.68,2127.054,540.228,136.814,34.579,8.727,2.2},
		{1780357.486,490774.667,128452.419,33051.529,8436.095,2143.308,542.923,137.246,34.642,8.734,2.2},
		{7028004.063,1941879.313,509072.88,131094.594,33477.331,8508.214,2155.73,545.042,137.592,34.694,8.74,2.2},
		{2.7770403353E7,7687299.959,2018492.357,520218.361,132912.902,33790.862,8563.635,2165.559,546.755,137.877,34.736,8.745,2.2},
		{1.09823853493E8,3.0444265834E7,8006692.41,2065208.089,527909.937,134256.802,34032.857,8607.702,2173.545,548.169,138.115,34.772,8.749,2.2},
		{4.3463715618E8,1.20612454002E8,3.1771204745E7,8201533.751,2097513.496,533611.511,135297.746,34226.036,8643.656,2180.168,549.357,138.317,34.803,8.752,2.2},
		{1.721207660629E9,4.77984025824E8,1.2610985714E8,3.2580664315E7,8336474.516,2121516.422,538040.495,136131.426,34384.215,8673.592,2185.756,550.37,138.491,34.83,8.756,2.2},
		{6.819987064572E9,1.894875211306E9,5.00706273262E8,1.29462040961E8,3.3141865136E7,8436921.101,2140205.828,541597.243,136816.109,34516.346,8698.932,2190.535,551.243,138.642,34.853,8.758,2.2},
		{2.7036516817167E10,7.515366574585E9,1.988484416752E9,5.14552132164E8,1.3178776819E8,3.3560215942E7,8515282.64,2155248.64,544525.705,137389.606,34628.512,8720.674,2194.672,552.005,138.774,34.874,8.761,2.2,},
		{1.07228981649241E11,2.9819797306265E10,7.898691952953E9,2.045547604049E9,5.24162417916E8,1.33523475915E8,3.388710333E7,8578474.944,2167660.882,546984.355,137877.664,34725.006,8739.544,2198.289,552.675,138.891,34.892,8.763,2.2},
		{4.2545008864916E11,1.18365686643067E11,3.1381402166102E10,8.133428414717E9,2.085162516189E9,5.31340952867E8,1.34881482591E8,3.4151136693E7,8630712.54,2178102.818,549081.254,138298.49,34808.955,8756.083,2201.479,553.269,138.996,34.908,8.765,2.2},
		{1.701800354596641E12,4.69999967391564E11,1.24699712856448E11,3.2345469721841E10,8.296390433995E9,2.114773082849E9,5.36963455078E8,1.35979860612E8,3.4369741299E7,8674734.382,2187025.084,550892.899,138665.379,34882.696,8770.703,2204.314,553.8,139.089,34.923,8.766,2.2},
		{6.807201418386566E12,1.866840130968202E12,4.95595501628209E11,1.28653634177634E11,3.3014658072112E10,8.418257742309E9,2.137985624798E9,5.41516250999E8,1.36890486716E8,3.4554243283E7,8712411.274,2194747.171,552475.254,138988.267,34948.008,8783.722,2206.85,554.276,139.173,34.936,8.768,2.2},
		{2.722880567354626E13,7.417233121711878E12,1.969934520286498E12,5.1179189070047E11,1.31397412343691E11,3.3515264195062E10,8.513860846563E9,2.156800318431E9,5.4529522861E8,1.37660073556E8,3.471238059E7,8745070.483,2201502.876,553870.223,139274.76,35006.278,8795.393,2209.133,554.707,139.25,34.948,8.769,2.2},
		{1.0891522269418505E14,2.9477585552721934E13,7.831290279100486E12,2.036208042754559E12,5.23026805992856E11,1.33450401036099E11,3.3908208794768E10,8.591415281241E9,2.17243281611E9,5.48492597689E8,1.38320518895E8,3.4849645094E7,8773683.977,2207467.558,555109.92,139530.783,35058.599,8805.916,2211.199,555.098,139.319,34.959,8.771,2.2}};

	/**
	 * Lookup table with pre-computed distortion decreases for significance bits (IWT). Each index means [MSBP of codeblock][significanceBitPlane, starting from the MSBP to 0].
	 * <p>
	 * Distortion estimators for bitplane image coding
	 */
	static final double[][] DDsigIWT =
		{{1.0},
		{5.649,1.0},
		{26.804,5.756,1.0},
		{113.321,27.806,5.798,1.0},
		{463.499,119.367,28.232,5.821,1.0},
		{1854.089,490.235,121.864,28.472,5.836,1.0},
		{7346.624,1971.814,502.042,123.31,28.628,5.846,1.0},
		{29001.708,7864.559,2026.781,508.975,124.263,28.738,5.853,1.0},
		{114332.361,31233.261,8104.241,2060.019,513.602,124.942,28.819,5.858,1.0},
		{450884.45,123747.65,32251.638,8247.547,2081.676,516.933,125.452,28.882,5.863,1.0},
		{1779051.674,490079.233,128090.244,32869.687,8346.213,2097.38,520.066,125.849,28.932,5.866,1.0},
		{7025459.581,1940525.689,508357.649,130738.71,33297.998,8416.211,2109.345,522.135,126.168,28.972,5.869,1.0},
		{2.7765258384E7,7684641.469,2017114.41,519486.673,132551.112,33609.51,8472.918,2118.79,523.805,126.43,29.006,5.872,1.0},
		{1.09813435406E8,3.0438816039E7,8003815.822,2063811.293,527186.712,133886.669,33851.373,8515.893,2127.704,525.182,126.648,29.035,5.874,1.0},
		{4.34616210325E8,1.20601467423E8,3.1765734786E7,8198641.29,2096051.842,532882.886,134928.03,34039.308,8550.882,2134.189,526.338,126.833,29.059,5.876,1.0},
		{1.721166203018E9,4.77962613107E8,1.26098907021E8,3.2574989411E7,8333677.027,2120082.696,537309.204,135766.996,34199.738,8579.967,2139.653,527.322,126.992,29.08,5.877,1.0},
		{6.819907102182E9,1.89483098154E9,5.0068471896E8,1.29450561481E8,3.3136194439E7,8434010.656,2138744.25,540856.425,136447.111,34328.436,8607.106,2144.322,528.17,127.131,29.099,5.878,1.0},
		{2.7036356505896E10,7.515281629659E9,1.988440884778E9,5.14530169059E8,1.31776125878E8,3.355435159E7,8512325.524,2153794.309,543791.489,137018.918,34443.73,8628.406,2148.36,528.91,127.252,29.115,5.88,1.0},
		{1.07228663837554E11,2.9819621885973E10,7.898603337938E9,2.045503651002E9,5.24139432268E8,1.33512205355E8,3.3881322039E7,8575617.134,2166220.483,546240.979,137507.312,34537.958,8646.873,2151.888,529.56,127.359,29.129,5.881,1.0},
		{4.25449454210249E11,1.18365340373307E11,3.1381231249951E10,8.133340581359E9,2.085117767825E9,5.31318462851E8,1.34870033699E8,3.4145263895E7,8627743.967,2176613.459,548337.709,137929.822,34625.022,8663.046,2154.997,530.136,127.454,29.142,5.882,1.0},
		{1.701799090643189E12,4.69999287428171E11,1.2469937221169E11,3.2345288078231E10,8.296299066003E9,2.114727833231E9,5.36940887791E8,1.35968197747E8,3.4363963498E7,8671752.484,2185525.983,550155.431,138299.203,34697.271,8677.331,2157.759,530.65,127.54,29.154,5.883,1.0},
		{6.807198785790226E12,1.866838746241229E12,4.95594802740441E11,1.28653284784268E11,3.3014476586212E10,8.418165544493E9,2.137939293531E9,5.41492650376E8,1.36878934785E8,3.4548461791E7,8709459.71,2193266.486,551725.142,138615.263,34761.207,8690.045,2160.228,531.112,127.617,29.164,5.883,1.0},
		{2.722880048715442E13,7.417230431426064E12,1.969933145693421E12,5.11791203239648E11,1.31397043157807E11,3.3515085244137E10,8.513770811721E9,2.156754830827E9,5.45272237941E8,1.37648316455E8,3.4706560504E7,8742158.387,2200011.833,553134.598,138905.892,34818.207,8701.434,2162.45,531.529,127.687,29.174,5.884,1.0},
		{1.0891521213904192E14,2.9477580106193387E13,7.831287596544856E12,2.036206617875154E12,5.23026083140078E11,1.33450043511634E11,3.3908024320017E10,8.591320823955E9,2.172386905908E9,5.48469148618E8,1.38308953622E8,3.4843851088E7,8770697.829,2205983.896,554369.497,139156.732,34873.861,8711.698,2164.46,531.907,127.75,29.183,5.885,1.0}};

	/**
	 * Lookup table with pre-computed distortion decreases for refinements bits (DWT). Each index means [MSBP of codeblock][significanceBitPlane, starting from the MSBP to 0].
	 * <p>
	 * Distortion estimators for bitplane image coding
	 */
	static final double[][] DDrefDWT =
		{{0.0},
		{0.0,0.24},
		{0.0,0.91,0.245},
		{0.0,3.41,0.958,0.247},
		{0.0,12.678,3.728,0.973,0.247},
		{0.0,46.734,14.458,3.825,0.98,0.248},
		{0.0,170.69,55.939,15.012,3.871,0.983,0.248},
		{0.0,616.771,215.976,58.835,15.276,3.897,0.986,0.248},
		{0.0,2212.404,832.194,230.325,60.226,15.428,3.914,0.987,0.248},
		{0.0,7939.545,3200.127,900.764,237.259,61.032,15.526,3.925,0.989,0.248},
		{0.0,28482.96,12280.235,3519.457,934.05,241.3,61.554,15.593,3.933,0.989,0.248},
		{0.0,102065.512,47021.348,13738.87,3674.982,953.52,243.923,61.916,15.643,3.94,0.99,0.248},
		{0.0,365039.012,179624.637,53585.041,14451.052,3766.228,966.197,245.75,62.181,15.68,3.945,0.991,0.249},
		{0.0,1302039.473,684439.683,208810.451,56795.796,14869.901,3825.799,975.053,247.089,62.383,15.71,3.949,0.991,0.249},
		{0.0,4627765.65,2600757.303,812962.018,223107.16,58687.799,15143.95,3867.506,981.556,248.108,62.54,15.733,3.952,0.992,0.249},
		{0.0,1.6374886237E7,9852250.308,3162179.054,875982.914,231545.075,59927.999,15336.196,3898.193,986.515,248.907,62.667,15.752,3.955,0.992,0.249},
		{0.0,5.7620861923E7,3.7232316606E7,1.2288130908E7,3437669.139,913225.476,237084.685,60799.462,15477.88,3921.625,990.407,249.549,62.77,15.768,3.957,0.992,0.249},
		{0.0,2.01380752002E8,1.40696476455E8,4.7703434514E7,1.348397947E7,3600639.574,937707.575,240982.876,61442.682,15586.228,3940.044,993.535,250.074,62.856,15.782,3.959,0.992,0.249},
		{0.0,6.97911101002E8,5.31649394598E8,1.84993198827E8,5.2863399195E7,1.4191989666E7,3707889.508,954957.032,243863.897,61935.211,15671.5,3954.863,996.098,250.511,62.929,15.793,3.961,0.993,0.249},
		{0.0,2.393504389659E9,2.008581469748E9,7.16602804774E8,2.07143375105E8,5.5920413524E7,1.4658360879E7,3783536.928,967720.402,246072.589,62323.291,15740.178,3967.015,998.232,250.88,62.991,15.803,3.962,0.993,0.249},
		{0.0,9.306628705087E9,7.586141851809E9,2.772609060582E9,8.11260883597E8,2.20273012049E8,5.7935674013E7,1.4987620291E7,3839568.036,977515.653,247814.722,62636.159,15796.548,3977.142,1000.034,251.195,63.044,15.812,3.964,0.993,0.249},
		{0.0,3.6169942110706E10,2.8639582592424E10,1.0714030864267E10,3.175535359268E9,8.67392264733E8,2.28934168158E8,5.935961617E7,1.5231719767E7,3882610.291,985249.246,249220.547,62893.185,15843.557,3985.696,1001.573,251.467,63.091,15.82,3.965,0.993,0.249},
		{0.0,1.40507198359744E11,1.08062241305777E11,4.134599792856E10,1.242317712941E10,3.414537994151E9,9.04440867578E8,2.35058303462E8,6.0416123282E7,1.5419395031E7,3916623.36,991495.369,250376.408,63107.688,15883.292,3993.007,1002.902,251.703,63.132,15.826,3.966,0.993,0.249},
		{0.0,5.45557499389015E11,4.0746456882556E11,1.59326469486607E11,4.8573428645023E10,1.343719341536E10,3.57236125052E9,9.30653476176E8,2.3960541383E8,6.1229038436E7,1.556781926E7,3944116.268,996634.915,251341.74,63289.113,15917.271,3999.321,1004.059,251.911,63.168,15.832,3.967,0.993,0.249}};

	/**
	 * Lookup table with pre-computed distortion decreases for refinements bits (IWT). Each index means [MSBP of codeblock][significanceBitPlane, starting from the MSBP to 0].
	 * <p>
	 * Distortion estimators for bitplane image coding
	 */
	static final double[][] DDrefIWT =
		{{0.0},
		{0.0,0.412},
		{0.0,0.776,0.439},
		{0.0,3.934,0.842,0.45},
		{0.0,14.007,3.717,0.871,0.455},
		{0.0,53.601,14.953,3.756,0.888,0.459},
		{0.0,205.026,59.732,15.396,3.782,0.899,0.461},
		{0.0,780.839,232.894,60.754,15.684,3.8,0.906,0.463},
		{0.0,2968.185,907.254,240.799,61.044,15.889,3.813,0.912,0.465},
		{0.0,11190.457,3563.171,950.275,244.63,61.587,16.044,3.824,0.917,0.466},
		{0.0,42389.095,13910.628,3751.49,971.796,246.24,62.625,15.77,3.833,0.92,0.467},
		{0.0,160056.497,54383.353,14799.407,3842.405,979.518,248.511,63.06,15.77,3.84,0.923,0.467},
		{0.0,603344.06,212553.225,58308.874,15228.862,3898.089,989.409,248.442,63.418,15.772,3.845,0.926,0.468},
		{0.0,2269870.984,829954.92,230122.149,60252.686,15466.045,3938.337,993.595,250.705,62.944,15.775,3.85,0.928,0.468},
		{0.0,8524869.784,3239340.284,906793.684,238641.627,61419.421,15633.627,3957.578,999.57,251.711,63.042,15.777,3.855,0.929,0.469},
		{0.0,3.1957155926E7,1.2633182482E7,3574166.896,944775.058,243684.742,62145.555,15755.561,3974.954,1000.212,252.579,63.129,15.78,3.858,0.931,0.469},
		{0.0,1.19551482429E8,4.9244848013E7,1.4082927655E7,3739373.559,967020.735,247049.044,62675.302,15841.36,3988.095,1006.386,251.805,63.206,15.783,3.862,0.932,0.47},
		{0.0,4.46247883053E8,1.91888444283E8,5.5478974027E7,1.4797647008E7,3837026.526,981793.29,249426.602,63064.318,15900.745,4003.155,1005.85,252.21,63.275,15.785,3.865,0.933,0.47},
		{0.0,1.661601827683E9,7.47496330263E8,2.18497153465E8,5.8555617526E7,1.5222680705E7,3901329.468,992221.787,251118.683,63343.427,15957.428,4009.496,1008.281,252.573,63.338,15.788,3.867,0.934,0.47},
		{0.0,6.17058171596E9,2.910960060176E9,8.60317398209E8,2.31676424897E8,6.0384909878E7,1.5502063877E7,3946805.734,999840.779,252460.304,63604.087,16000.34,4018.883,1007.411,252.901,63.394,15.79,3.87,0.935,0.47},
		{0.0,2.4741596228503E10,1.1332613093352E10,3.386656358453E9,9.16504990037E8,2.39516260077E8,6.1592644525E7,1.5699645406E7,3980716.519,1005671.505,253506.535,63803.761,16033.146,4022.082,1008.922,253.198,63.445,15.792,3.872,0.936,0.471},
		{0.0,9.9239225439867E10,4.4105461467254E10,1.3328524812441E10,3.625196874404E9,9.49949438808E8,2.44697759054E8,6.244647883E7,1.5847015066E7,4006553.196,1010359.971,254354.348,63935.218,16061.703,4029.593,1010.297,253.468,63.491,15.794,3.874,0.937,0.471},
		{0.0,3.9817154203567E11,1.7160240128272E11,5.2442926779035E10,1.433754464314E10,3.767314980163E9,9.72081351093E8,2.48367226838E8,6.3081087487E7,1.59593506E7,4027098.136,1014181.775,255018.11,64084.5,16081.448,4030.365,1011.553,253.715,63.534,15.796,3.875,0.938,0.471},
		{0.0,1.597964232707309E12,6.67455592878735E11,2.06293165284679E11,5.6697613367393E10,1.4939147927264E10,3.861415420932E9,9.87765242075E8,2.51096687713E8,6.3569322674E7,1.6049078541E7,4043573.514,1017188.931,255647.659,64184.653,16099.308,4040.349,1010.133,253.942,63.573,15.798,3.877,0.938,0.471}};

	/**
	 * Coder used to output the ByteStream.
	 * <p>
	 * This Coder must be compatible with the abstract class Coder.
	 */
	private Coder blockCoder = new MQCoder();

	/**
	 * Stripe height.
	 * <p>
	 * Negative values not allowed for this field.
	 */
	private final static int stripeHeight = 4;

	/**
	 * Number of bits needed to flush the singificant coefficient in the stripe (used when the run mode is broken).
	 * <p>
	 * This should be log_2(stripeHeight)
	 */
	private final static int numBitsRunModeBreak = (int) Math.ceil(Math.log(stripeHeight) / Math.log(2));


	/**
	 * Constructor that does not receive anything. Before use the functions of codeNextCodingPass the function swapCodeBlock must be used.
	 */
	public BPECoder(){}

	/**
	 * Constructor that receives block samples and its subband and initializes all needed structures.
	 *
	 * @param blockSamples definition in {@link #blockSamples}
	 * @param MSBPlane definition in {@link #MSBPlane}
	 * @param subband definition in {@link #subband}
	 * @param subbandMSBPlane definition in {@link #subbandMSBPlane}
	 * @param RDAdjust definition in {@link #RDAdjust}
	 * @param intProcessing definition in {@link BOIDecode.Decode.BPEDecoder#intProcessing}
	 *
	 * @throws WarningException when the MSBPlane is higher than 24 (max allowed due to distortion estimators LUTs)
	 */
	public BPECoder(int[][] blockSamples, int MSBPlane, int subband, int subbandMSBPlane, float RDAdjust, boolean intProcessing) throws WarningException{
		swapCodeBlock(blockSamples, MSBPlane, subband, subbandMSBPlane, RDAdjust, intProcessing);
	}

	/**
	 * Swaps the code-block encoded in this class. Some initilizations needed.
	 *
	 * @param blockSamples definition in {@link #blockSamples}
	 * @param MSBPlane definition in {@link #MSBPlane}
	 * @param subband definition in {@link #subband}
	 * @param subbandMSBPlane definition in {@link #subbandMSBPlane}
	 * @param RDAdjust definition in {@link #RDAdjust}
	 * @param intProcessing definition in {@link BOIDecode.Decode.BPEDecoder#intProcessing}
	 *
	 * @throws WarningException when the MSBPlane is higher than 24 (max allowed due to distortion estimators LUTs)
	 */
	public void swapCodeBlock(int[][] blockSamples, int MSBPlane, int subband, int subbandMSBPlane, float RDAdjust, boolean intProcessing) throws WarningException{
		//Data copy
		this.blockSamples = blockSamples;
		if(subbandMSBPlane >= DDsigDWT.length){
			throw new WarningException("subbandMSBPlane higher than the allowed in the distortion estimation LUTs.");
		}
		this.MSBPlane = MSBPlane;
		this.subband = subband;
		this.subbandMSBPlane = subbandMSBPlane;
		this.RDAdjust = RDAdjust;
		this.intProcessing = intProcessing;

		//Size set
		ySize = blockSamples.length;
		xSize = blockSamples[0].length;

		//Set the number of bitPlanes
		int bitPlanes = MSBPlane + 1;

		//Memory allocation
		contextMap = new byte[ySize][xSize];
		errorsDec = new float[bitPlanes][3];
		lengths = new int[bitPlanes][3];
		CDByteStream = new ByteStream();
		blockCoder.swapOutputByteStream(CDByteStream);

		//Set blockCoder (we do not create another object, we reutilitze the already created coder)
		blockCoder.restart();
		blockCoder.reset();

		//Initialize current variables
		currentBitPlane = this.MSBPlane;
		currentCodingPass = 2; //CP
	}

	/**
	 * Code the image samples (all bit planes) and produces the bit stream of JPEG2000.
	 *
	 * @throws WarningException when reached the last coding pass or problems with the outputByteStream
	 */
	public void runAll() throws WarningException{
		while(currentBitPlane >= 0){
			codeNextCodingPass();
		}
	}

	/**
	 * Encodes the next coding pass.
	 *
	 * @throws WarningException when reached the last coding pass or problems with the outputByteStream
	 */
	public void codeNextCodingPass() throws WarningException{
		if(currentBitPlane < 0){
			throw new WarningException("Reached last bit plane.");
		}

		//Change the outputByteStream of the blockCoder for the RESTART coding variation - disabled
		//blockCoder.swapOutputByteStream(CDByteStream[MSBPlane-currentBitPlane][currentCodingPass]);
		//blockCoder.restart();

		//Initialize current variables
		currentBitPlaneMask = (1 << currentBitPlane); //2^bitPlane

		//System.out.println("\nBP: " + currentBitPlane + " CP: " + currentCodingPass);
		//Encodes the current coding pass
		switch(currentCodingPass){
		case 0: //SPP
			significancePropagationPass();
			//Distortion decrease
			if(!intProcessing){
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDsigDWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}else{
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDsigIWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}
			break;
		case 1: //MRP
			magnitudeRefinementPass();
			//Distortion decrease
			if(!intProcessing){
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDrefDWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}else{
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDrefIWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}
			break;
		case 2: //CP
			cleanupPass();
			//Distortion decrease
			if(!intProcessing){
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDsigDWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}else{
				errorsDec[MSBPlane-currentBitPlane][currentCodingPass] = (float) ((double)numSIGS * DDsigIWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] * (double)(RDAdjust * RDAdjust));
			}
			break;
		}

		//Terminate the outputByteStream -- needed at each coding pass only when using the RESTART coding variation
		//blockCoder.terminate();
		if((currentBitPlane == 0) && (currentCodingPass == 2)){
			blockCoder.terminate();
			//Length of the coding pass
			lengths[MSBPlane-currentBitPlane][currentCodingPass] = CDByteStream.getNumBytes();
			renormalizeLengths();
		}else{
			//Length of the coding pass (current length +2 to make sure internal registers of the MQ coder are flushed)
			lengths[MSBPlane-currentBitPlane][currentCodingPass] = CDByteStream.getNumBytes() + blockCoder.remainingBytes();
		}

		//Prepares the current variables for the next coding pass
		if(currentCodingPass == 2){
			currentCodingPass = 0;
			currentBitPlane--;
		}else{
			currentCodingPass++;
		}
	}

	/**
	 * Check that lengths are consistent.
	 */
	private void renormalizeLengths() throws WarningException{
		//Make sure the previous potential truncation point does not finish with and 0xFF byte, which can be excluded
		int currTMPBitPlane = 0;
		int currTMPCodingPass = 1;
		int nextTMPBitPlane = 0;
		int nextTMPCodingPass = 2;
		while((currTMPBitPlane < MSBPlane) || ((currTMPBitPlane == MSBPlane) && (currTMPCodingPass == 2))){
			if(lengths[MSBPlane-currTMPBitPlane][currTMPCodingPass] > lengths[MSBPlane-nextTMPBitPlane][nextTMPCodingPass]){
				lengths[MSBPlane-currTMPBitPlane][currTMPCodingPass] = lengths[MSBPlane-nextTMPBitPlane][nextTMPCodingPass];
			}
			if((lengths[MSBPlane-currTMPBitPlane][currTMPCodingPass] > 0) && (CDByteStream.getByte(lengths[MSBPlane-currTMPBitPlane][currTMPCodingPass] - 1) == (byte)0xFF)){
				lengths[MSBPlane-currTMPBitPlane][currTMPCodingPass]--;
			}
			nextTMPCodingPass = currTMPCodingPass;
			nextTMPBitPlane = currTMPBitPlane;
			if(currTMPCodingPass == 0){
				currTMPCodingPass = 2;
				currTMPBitPlane++;
			}else{
				currTMPCodingPass--;
			}
		}
	}


	/**
	 * Search the bit plane where the most significant bit is allocated.
	 *
	 * @param blockSamples definition in {@link #blockSamples}
	 * @return the MSB bit plane of all block samples (least significant bit plane is 0)
	 */
	static public int MSBPlane(int[][] blockSamples){
		long maxValue = -1;
		int ySize = blockSamples.length;
		int xSize = blockSamples[0].length;
		for(int y = 0; y < ySize; y++){
		for(int x = 0; x < xSize; x++){
			if(Math.abs(blockSamples[y][x]) > maxValue){
				maxValue = Math.abs(blockSamples[y][x]);
			}
		}}
		int bp = (int) (Math.log( (float) maxValue) / Math.log(2.0D));
		return(bp <= 0 ? 1: bp);
	}

	/**
	 * Calculates the bit sign and context from a sample (called when the sample becomes significant).
	 *
	 * @param x position of the sample
	 * @param y position of the sample
	 */
	private void signEncode(int x, int y){
		int up    = y-1 <  0     ? 0 : contextMap[y-1][x] >= -8 && contextMap[y-1][x] <= 8 ? 0 : blockSamples[y-1][x] < 0 ? -1 : 1;
		int down  = y+1 >= ySize ? 0 : contextMap[y+1][x] >= -8 && contextMap[y+1][x] <= 8 ? 0 : blockSamples[y+1][x] < 0 ? -1 : 1;
		int left  = x-1 <  0     ? 0 : contextMap[y][x-1] >= -8 && contextMap[y][x-1] <= 8 ? 0 : blockSamples[y][x-1] < 0 ? -1 : 1;
		int right = x+1 >= xSize ? 0 : contextMap[y][x+1] >= -8 && contextMap[y][x+1] <= 8 ? 0 : blockSamples[y][x+1] < 0 ? -1 : 1;

		int v = up+down    == 0 ? 0 : up+down    < 0 ? -1 : 1;
		int h = left+right == 0 ? 0 : left+right < 0 ? -1 : 1;

		int signContext = 0;
		switch(h){
		case -1:
			switch(v){
			case -1:
				signContext = -14;
				break;
			case 0:
				signContext = -13;
				break;
			case 1:
				signContext = -12;
				break;
			}
			break;
		case 0:
			switch(v){
			case -1:
				signContext = -11;
				break;
			case 0:
				signContext = 10;
				break;
			case 1:
				signContext = 11;
				break;
			}
			break;
		case 1:
			switch(v){
			case -1:
				signContext = 12;
				break;
			case 0:
				signContext = 13;
				break;
			case 1:
				signContext = 14;
				break;
			}
			break;
		}

		boolean sign = ((blockSamples[y][x] < 0 ? -1 : 1) * (signContext < 0 ? -1 : 1)) < 0 ? true : false;
		blockCoder.codeBit(sign, Math.abs(signContext));
		//System.out.print((sign ? "1": "0") + "(" + Math.abs(signContext) + ")");
	}

	/**
	 * Significance propagation pass of JPEG2000 coding scheme.<br>
	 * The main idea of this pass is that is more probable that neighbourhoods of significance samples will be significance in next bit plane. This algorithm searchs which of these neighbourhoods samples are significant in the current bit plane.
	 */
	private void significancePropagationPass(){
		//Used to compute distortion decreases
		numSIGS = 0;
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
			for(int y = yBegin; y < yEnd; y++){
				//Sample coding
				//contextMap shows the samples that we must visit (contexts from 1 to 8)
				if((contextMap[y][x] > 0) && (contextMap[y][x] <= 8)){
					boolean significant = (Math.abs(blockSamples[y][x]) & currentBitPlaneMask) > 0;

					if(significant){
						//System.out.print("1(" + contextMap[y][x] + ")");
						blockCoder.codeBit(true, contextMap[y][x]);
						signEncode(x, y);

						//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
						updateMap(x, y, false);

						//Used to compute distortion decreases
						numSIGS++;

					}else{
						//System.out.print("0(" + contextMap[y][x] + ")");
						blockCoder.codeBit(false, contextMap[y][x]);
					}
				}
			}}
			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Magnitude refinement pass of JPEG2000 coding scheme.<br>
	 * This pass searchs the significance samples that in the current bit plane have a refinement bit.
	 */
	private void magnitudeRefinementPass(){
		//Used to compute distortion decreases
		numSIGS = 0;
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
			for(int y = yBegin; y < yEnd; y++){
				//The coefficient has been significant in the previous bit plane
				if(Math.abs(contextMap[y][x]) >= 15){
					if(contextMap[y][x] > 0){
						boolean refinement = (Math.abs(blockSamples[y][x]) & currentBitPlaneMask) > 0;
						blockCoder.codeBit(refinement, contextMap[y][x]);
						contextMap[y][x] = 17;

						//Used to compute distortion decreases
						numSIGS++;

					}else{
						//The coefficient is significant in the current bit plane
						contextMap[y][x] = (byte) -contextMap[y][x];
					}
				}
			}}

			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Cleanup pass of JPEG2000 coding scheme.<br>
	 * This pass scans all non visited samples to find out which are significance in the current bit plane. It has a run mode started under special conditions that optimizes the coder.
	 */
	private void cleanupPass(){
		//Used to compute distortion decreases
		numSIGS = 0;
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
				//Test if we can code the column in run mode
				boolean runMode = true;
				if(yEnd - yBegin == stripeHeight){
					//Check if exists a sample with some significant neighbourhood or the column should not be completely scanned. If one of these conditions is true then the run mode cannot be started
					int y = yBegin;
					do{
						if(contextMap[y][x] != 0){
							runMode = false;
						}
						y++;
					}while((runMode) && (y < yEnd));
				}else{
					runMode = false;
				}

				int y = yBegin;
				//If we can start the run mode, we check if the column has any significant coefficient
				if(runMode){
					boolean significant = false;
					do{
						significant = (Math.abs(blockSamples[y][x]) & currentBitPlaneMask) > 0;
						y++;
					}while((!significant) && (y < yEnd));
					if(significant){
						//Break the run mode
						//System.out.print("1(9)X(18)X(18)");
						blockCoder.codeBit(true, 9);
						int nSample = (y-1) % stripeHeight;
						for(int bitRunModeBreak = numBitsRunModeBreak-1; bitRunModeBreak >= 0; bitRunModeBreak--){
							boolean bit = (nSample & (1 << bitRunModeBreak)) > 0;
							blockCoder.codeBit(bit, 18);
						}

						//Sign of the coefficient
						signEncode(x, y-1);

						//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
						updateMap(x, y-1, true);

						//Used to compute distortion decreases
						numSIGS++;

					}else{
						//Column entirely encoded in run mode
						//System.out.print("0(9)");
						blockCoder.codeBit(false, 9);
					}
				}

				//Encode each coefficient
				while(y < yEnd){
					//contextMap shows the samples that we must visit (only these ones that are not visited yet in the current bit plane)
					if((contextMap[y][x] >= -8) && (contextMap[y][x] <= 0)){
						boolean significant = (Math.abs(blockSamples[y][x]) & currentBitPlaneMask) > 0;
						//In the next bit plane this coefficient must be visitied by the SPP
						contextMap[y][x] = (byte) -contextMap[y][x];
						if(significant){
							//System.out.print("1(" + contextMap[y][x] + ")");
							blockCoder.codeBit(true, contextMap[y][x]);
							signEncode(x, y);

							//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
							updateMap(x, y, true);

							//Used to compute distortion decreases
							numSIGS++;

						}else{
							//System.out.print("0(" + contextMap[y][x] + ")");
							blockCoder.codeBit(false, contextMap[y][x]);
						}
					}
					y++;
				}
			}
			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Turn neighbourhoods map samples of (x,y) from 0 to 1 if they can be visited in the current scan.
	 *
	 * @param x x position of the sample
	 * @param y y position of the sample
	 * @param CP indicates if the updating is performed by the Cleanup Pass (true, otherwise false)
	*/
	private void updateMap(int x, int y, boolean CP){
		//Context of the current coefficient
		int yNeighBegin = y-1 >= 0 ? y-1: 0;
		int yNeighEnd = y+1 < ySize ? y+1: y;
		int xNeighBegin = x-1 >= 0 ? x-1: 0;
		int xNeighEnd = x+1 < xSize ? x+1: x;
		boolean significantNeighs = false;
		for(int yNeigh = yNeighBegin; yNeigh <= yNeighEnd; yNeigh++){
		for(int xNeigh = xNeighBegin; xNeigh <= xNeighEnd; xNeigh++){
			if(Math.abs(contextMap[yNeigh][xNeigh]) >= 15){
				if(Math.abs(contextMap[yNeigh][xNeigh]) == 15){
					contextMap[yNeigh][xNeigh] = contextMap[yNeigh][xNeigh] < 0 ? (byte) -16: (byte) 16;
				}
				significantNeighs = true;
			}
		}}
		if(significantNeighs){
			contextMap[y][x] = CP ? (byte) 16: (byte) -16;
		}else{
			contextMap[y][x] = CP ? (byte) 15: (byte) -15;
		}

		//Context of the neighboors
		switch(subband){
		case 0: //SUBBAND LL
		case 2: //SUBBAND LH
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 5: (byte) -5;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -8: (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 1: (byte) -1;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -5: (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -8: (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -1: (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -3: (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 3: (byte) -3;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			break;
		case 1: //SUBBAND HL
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 3: (byte) -3;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 1: (byte) -1;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -3: (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -1: (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -5: (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -8: (byte) 8;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 5: (byte) -5;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -8: (byte) 8;
					break;
				}
			}
			break;
		case 3: //SUBBAND HH
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 1: (byte) -1;
					break;
				case 1:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 3: (byte) -3;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 3: (byte) -3;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -3: (byte) 3;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -1: (byte) 1;
					break;
				case 1:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -3: (byte) 3;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 3: (byte) -3;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -3: (byte) 3;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -1: (byte) 1;
					break;
				case 1:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 1: (byte) -1;
					break;
				case 1:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			break;
		}
	}

	/**
	 * @return errors definition in {@link #errorsDec}
	 */
	public float[][] getErrorsDec(){
		return(errorsDec);
	}

	/**
	 * @return lengths definition in {@link #lengths}
	 */
	public int[][] getLengths(){
		return(lengths);
	}

	/**
	 * @return CDByteStream definition in {@link #CDByteStream}
	 */
	public ByteStream getByteStream(){
		return(CDByteStream);
	}

}

