/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.Code;
import BOIException.WarningException;
import BOIStream.ByteStream;


/**
 * This class encodes bits belonging to the fractional bit plane coder of JPEG2000, coding them with the MQ coder. This class allows the use of the MQ with the options: restart, reset, bypass. Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; codeBit<br>
 * &nbsp; codeBit<br>
 * &nbsp; terminateOptimal | terminate<br>
 * &nbsp; reset<br>
 * &nbsp; swapOutputByteStream<br>
 * &nbsp; restart<br>
 * &nbsp; codeBit<br>
 * &nbsp; codeBit<br>
 * &nbsp; ...<br>
 *
 * @author Francesc Auli-Llinas
 * @version 1.2
 */
public class MQCoder implements Coder{

	/**
	 * ByteStream where the generated bytes are flushed
	 *<p>
	 * The resulting ByteStream can be null when optimal terminations are used. Usually it will contain a part of the final JPEG2000 compliant codestream.
	 */
	ByteStream outputByteStream;

	/**
	 * Number of contexts (defined by JPEG2000).
	 * <p>
	 * Positive number.
	 */
	private final int numContexts = 19;

	/**
	 * MQ context state initialization (defined by JPEG2000 standard).
	 * <p>
	 * First array index represents the context and second is: (p.e. [2][0] is the initial state of context 2)
	 * <ul>
	 *   <li> 0 - Initial state for the context
	 *   <li> 1 - Initial expected symbol for the context
	 * </ul>
	 */
	private final int[][] stateSymbolInit = {
		{4,  0}, //context  0
		{0,  0}, //context  1
		{0,  0}, //context  2
		{0,  0}, //context  3
		{0,  0}, //context  4
		{0,  0}, //context  5
		{0,  0}, //context  6
		{0,  0}, //context  7
		{0,  0}, //context  8
		{3,  0}, //context  9
		{0,  0}, //context 10
		{0,  0}, //context 11
		{0,  0}, //context 12
		{0,  0}, //context 13
		{0,  0}, //context 14
		{0,  0}, //context 15
		{0,  0}, //context 16
		{0,  0}, //context 17
		{46, 0}  //context 18
	};

	/**
	 * MQ probability state transition (defined by JPEG2000 standard).
	 * <p>
	 * First array index represents the current state and second is: (i.e. [1][0] is the transition to do when code a MPS symbol and the state of the context is 1)
	 * <ul>
	 *   <li> 0 - MPS transition
	 *   <li> 1 - LPS transition
	 *   <li> 2 - If 1 change MPS by LPS, nothing otherwise
	 *   <li> 3 - estimation (probability is--> 0xXXX / (2^16 * alpha) -alpha usually is 0.708-)
	 * </ul>
	 */
	private final int[][] stateTrans = {
		{ 1,  1, 1, 0x5601}, //state  0
		{ 2,  6, 0, 0x3401}, //state  1
		{ 3,  9, 0, 0x1801}, //state  2
		{ 4, 12, 0, 0x0AC1}, //state  3
		{ 5, 29, 0, 0x0521}, //state  4
		{38, 33, 0, 0x0221}, //state  5
		{ 7,  6, 1, 0x5601}, //state  6
		{ 8, 14, 0, 0x5401}, //state  7
		{ 9, 14, 0, 0x4801}, //state  8
		{10, 14, 0, 0x3801}, //state  9
		{11, 17, 0, 0x3001}, //state 10
		{12, 18, 0, 0x2401}, //state 11
		{13, 20, 0, 0x1C01}, //state 12
		{29, 21, 0, 0x1601}, //state 13
		{15, 14, 1, 0x5601}, //state 14
		{16, 14, 0, 0x5401}, //state 15
		{17, 15, 0, 0x5101}, //state 16
		{18, 16, 0, 0x4801}, //state 17
		{19, 17, 0, 0x3801}, //state 18
		{20, 18, 0, 0x3401}, //state 19
		{21, 19, 0, 0x3001}, //state 20
		{22, 19, 0, 0x2801}, //state 21
		{23, 20, 0, 0x2401}, //state 22
		{24, 21, 0, 0x2201}, //state 23
		{25, 22, 0, 0x1C01}, //state 24
		{26, 23, 0, 0x1801}, //state 25
		{27, 24, 0, 0x1601}, //state 26
		{28, 25, 0, 0x1401}, //state 27
		{29, 26, 0, 0x1201}, //state 28
		{30, 27, 0, 0x1101}, //state 29
		{31, 28, 0, 0x0AC1}, //state 30
		{32, 29, 0, 0x09C1}, //state 31
		{33, 30, 0, 0x08A1}, //state 32
		{34, 31, 0, 0x0521}, //state 33
		{35, 32, 0, 0x0441}, //state 34
		{36, 33, 0, 0x02A1}, //state 35
		{37, 34, 0, 0x0221}, //state 36
		{38, 35, 0, 0x0141}, //state 37
		{39, 36, 0, 0x0111}, //state 38
		{40, 37, 0, 0x0085}, //state 39
		{41, 38, 0, 0x0049}, //state 40
		{42, 39, 0, 0x0025}, //state 41
		{43, 40, 0, 0x0015}, //state 42
		{44, 41, 0, 0x0009}, //state 43
		{45, 42, 0, 0x0005}, //state 44
		{45, 43, 0, 0x0001}, //state 45
		{46, 46, 0, 0x5601}  //state 46
	};

	/**
	 * Interval range.
	 * <p>
	 * From right to left: 8 register bits, 3 spacer bits, 8 partial code byte, 1 carry bit.
	 */
	private int A;

	/**
	 * Lower down interval.
	 * <p>
	 * From right to left: 8 register bits, 3 spacer bits, 8 partial code byte, 1 carry bit.
	 */
	private int C;

	/**
	 * Number of bits to transfer (down counter).
	 * <p>
	 * Usually it is initialized to 8, unless carry situations that it initilized to 7. Then count down until 0.
	 */
	private int t;

	/**
	 * Byte to transfer.
	 * <p>
	 * Byte flushed to the outputByteStream.
	 */
	private int Tr;

	/**
	 * Number of bytes transferred to the outputByteStream.
	 * <p>
	 * Byte flushed to the output ByteStream.
	 */
	private int L;

	/**
	 * State and Most Probable Symbol of each context.
	 * <p>
	 * Array index is the context and the State [context][0] or MPS [context][1]. It must correspond to some stateTrans.
	 */
	private int[][] statesMPS = null;


	/**
	 * Constructor that initializes internal registers of MQ coder but does not initialize any output ByteStream. Before use the MQ coder the function swapOutputByteStream should be called.
	 */
	public MQCoder(){
		//Initialize context status
		statesMPS = new int[numContexts][2];
		reset();

		//Initialize internal registrers
		restart();
	}

	/**
	 * Constructor that initializes internal registers of MQ coder.
	 *
	 * @param outputByteStream ByteStream where the bytes are flushed
	 */
	public MQCoder(ByteStream outputByteStream){
		//Copy the destination ByteStream
		swapOutputByteStream(outputByteStream);
		//Initialize context status
		statesMPS = new int[numContexts][2];
		reset();
		//Initialize internal registrers
		restart();
	}

	/**
	 * Code a bit using the context 18 (equivalent probabilities).
	 *
	 * @param bit bit to encode
	 */
	public void codeBit(boolean bit){
		codeBit(bit, (byte) 18);
	}

	/**
	 * Code a bit.
	 *
	 * @param bit bit to encode
	 * @param context context the bit
	 */
	public void codeBit(boolean bit, int context){
		int x = bit ? 1 : 0;
		int s = statesMPS[context][1];
		int p = stateTrans[statesMPS[context][0]][3];

		//OPTIMIZATION OF JPEG2000 BOOK (p 646) by Taubman/Marcellin
		A -= p;
		if(x == s){ //MPS coding
			if(A >= (1 << 15)){
				C += p;
			}else{
				if(A < p){
					A = p;
				}else{
					C += p;
				}
				statesMPS[context][0] = stateTrans[statesMPS[context][0]][0];
				while(A < (1 << 15)){
					A <<= 1;
					C <<= 1;
					t--;
					if(t == 0){
						transferByte();
					}
				}
			}
		}else{ //LPS coding
			if(A < p){
				C += p;
			}else{
				A = p;
			}
			if(stateTrans[statesMPS[context][0]][2] == 1){
				statesMPS[context][1] = statesMPS[context][1] == 0 ? 1: 0;
			}
			statesMPS[context][0] = stateTrans[statesMPS[context][0]][1];
			while(A < (1 << 15)){
				A <<= 1;
				C <<= 1;
				t--;
				if(t == 0){
					transferByte();
				}
			}
		}

		/* "NORMAL" ALGORITHM
		//Conditional exchange of MPS and LPS in order to associate the highest length interval (p or A) to the MPS.
		if(A < p){
			s = 1 - s;
		}
		if(x == s){
			//Assign MPS the upper sub-interval
			C += p;
		}else{
			//Assign LPS the lower sub-interval
			A = p;
		}

		//Adaptative algorithm. Obtain new value of the state associate to the context. Moreover, if the symbol is a LPS, it may switch the values of MPS/LPS (depends on the value of stateTrans[statesMPS[context][0]][2]
		if(A < ((int) 1 << 15)){
			if(x == statesMPS[context][1]){ //The symbol was a real MPS
				statesMPS[context][0] = stateTrans[statesMPS[context][0]][0];
			}else{ //The symbol was a real LPS
				//Switch MPS/LPS if necessary
				if(stateTrans[statesMPS[context][0]][2] == 1){
					statesMPS[context][1] = statesMPS[context][1] == 0 ? 1: 0;
				}
				statesMPS[context][0] = stateTrans[statesMPS[context][0]][1];
			}
		}

		//Perform renormalization shift. All left zeroes in A (16 bits) are bits of C that should be transmitted. Bits are written in the output buffer only when the down counter t is zero.
		while(A < ((int) 1 << 15)){
			A <<= 1;
			C <<= 1;
			t--;
			if(t == 0){
				//There is a full byte to be transmitted
				transferByte();
			}
		}
		*/
	}

	/**
	 * This function transfer a byte to the outputByteStream taken its content from C.
	 */
	private void transferByte(){
		if(Tr == 0xFF){ //Bit stuff
			outputByteStream.addByte((byte) Tr);
			L++;
			Tr = (C >>> 20); //Put C_msbs to Tr
			C &= (~0xFFF00000); //Put 0 to C_msbs
			t = 7;
		}else{
			if(C >= 0x08000000){
				//Propagate any carry bit from C into Tr
				Tr += 0x01;
				C &= (~0xF8000000); //Reset the carry bit
			}
			if(L >= 0){
				outputByteStream.addByte((byte) Tr);
			}
			L++;
			if(Tr == 0xFF){ //Bit stuff
				//Even though it may not be a bit carry in this case, we shoud write, after a FFh value, a byte like c000bbbb, because the decoder con not tell the difference between both cases
				Tr = (C >>> 20); //Put C_msbs to Tr
				C &= (~0xFFF00000); //Put 0 to C_msbs
				t = 7;
			}else{
				Tr = (C >>> 19); //Put C_partial to Tr
				C &= (~0xFFF80000); //Put 0 to C_partial
				t = 8;
			}
		}
	}

	/**
	 * Swaps the current output ByteStream. Before to call this function you should terminate the last ByteStream with the functions terminateEasy or terminateOptimal. If you are using the RESTART mode of JPEG2000, after the calling this function you should call the restart function.
	 *
	 * @param outputByteStream destination ByteStream
	 */
	public void swapOutputByteStream(ByteStream outputByteStream){
		this.outputByteStream = outputByteStream;
	}

	/**
	 * Reset the internal contexts of the MQ coder, i.e. the probability state of the Most Probable Symbols (MPS) are reset to their initial value.
	 */
	public void reset(){
		for(int i = 0; i < numContexts; i++){
			statesMPS[i][0] = stateSymbolInit[i][0];
			statesMPS[i][1] = stateSymbolInit[i][1];
		}
	}

	/**
	 * Restart the value of the internal variables A, C, ... of the MQ coder, in other words, restarts the MQ coder althought the context probabilities are maintained. This function does NOT terminate the current ByteStream (the codestream should be terminated before to call this function using the terminate functions).
	 */
	public void restart(){
		A  = 0x8000;
		C  = 0;
		t = 12;
		Tr = (byte) 0;
		L = -1;
	}

	/**
	 * Compute the number of bytes belonging to the currently encoded data needed to flush the internal registers. This function is useful to determine potential truncation points for coding passes.
	 *
	 * @return number of bytes
	 */
	public int remainingBytes(){
		int nBits = 27 - 15 - t;
		return(((nBits + 8) / 7) + 1); //ceil(nBits / 7) + 1 //Assume the worst case
	}

	/**
	 * Finishes the current outputByteStream using the termination optimal, used by default. After calling this function you should call the restart function or swapOutputByteStream+restart.
	 *
	 * @throws WarningException when some problem with the manipulation of the ByteStream happens
	 */
	public void terminate() throws WarningException{
		terminateOptimal();
	}

	/**
	 * Finishes the current outputByteStream using easy termination, i.e. flushing all the remaining bits of the C register to the output ByteStream. After calling this function you should call the restart function or swapOutputByteStream+restart.
	 */
	public void terminateEasy(){
		//Easy MQ codeword termination
		int nBits = 27 - 15 - t;
		C <<= t;
		while(nBits > 0){
			transferByte();
			nBits -= t;
			C <<= t;
		}
		transferByte();

		//If last byte is 0xFF, delete it because MQ cannot terminate in this way (decoder puts all these bits to 1)
		if(t == 7){
			outputByteStream.deleteEndBytes(1);
		}
	}

	/**
	 * Finishes the current outputByteStream using the optimal termination. After calling this function you should call the restart function or swapOutputByteStream+restart.
	 *
	 * @throws WarningException when some problem with the manipulation of the ByteStream happens
	 */
	public void terminateOptimal() throws WarningException{
		//Normalization state variables
		int NZTr = Tr;
		int NZt = t;
		int NZC = C;
		int NZA = A;
		int NZL = L;

		//Length without easy termination
		int lengthEmptyTermination = outputByteStream.getNumBytes();

		//First we must terminate the stream in an easy way
		terminateEasy();

		int necessaryBytes = minFlush(NZTr, NZt, NZC, NZA, NZL, lengthEmptyTermination);

		//outputByteStream length with the optimal termination
		int lengthOptimalTermination = lengthEmptyTermination + necessaryBytes;

		//Take into acount that 0xff and 0xff 0x7f should not happen when terminating with the easy mode, but checked again
		if(lengthOptimalTermination >= 1){
		if(outputByteStream.getByte(lengthOptimalTermination - 1) == 0xFF){
				lengthOptimalTermination--;
		}}
		boolean elimination;
		do{
			elimination = false;
			if(lengthOptimalTermination >= 2){
			if((outputByteStream.getByte(lengthOptimalTermination - 2) == 0xFF) && (outputByteStream.getByte(lengthOptimalTermination - 1) == 0x7F)){
				lengthOptimalTermination -= 2;
				elimination = true;
			}}
		}while(elimination);

		//Remove unwanted bytes
		outputByteStream.deleteEndBytes(outputByteStream.getNumBytes() - lengthOptimalTermination);
	}

	/**
	 * Determine the minimum number of bytes needed to terminate the flushed ByteStream assuring a complete recovering (MQ decoder fills the ByteStreams with 0x00 bytes). Algorithm extract from book of JPEG2000 by Taubman and Marcellin.
	 *
	 * @param NZTr Tr register for the normaliztion
	 * @param NZt t register for the normaliztion
	 * @param NZC C register for the normaliztion
	 * @param NZA A register for the normaliztion
	 * @param NZL number of bytes flushed
	 * @param lengthEmptyTermination length bytes used by the easy termination
	 * @return the number of bytes that should be flushed to terminate the ByteStream optimally
	 *
	 * @throws WarningException when some problem with the manipulation of the ByteStream happens
	 */
	private int minFlush(int NZTr, int NZt, int NZC, int NZA, int NZL, int lengthEmptyTermination) throws WarningException{
		//Real counterparts
		long Cr = ((long) NZTr << 27) + ((long) NZC << NZt);
		long Ar = (long) NZA << NZt;

		//Decoder
		long Rf = 0;
		int s = 8;
		int Sf = 35;

		//What we are looking for
		int necessaryBytes = 0;

		//If we have necessaryBytes greater than that we cut more bytes than we have
		int maxNecessaryBytes = 5;
		int cutZone = outputByteStream.getNumBytes() - lengthEmptyTermination;
		if(maxNecessaryBytes > cutZone){
			maxNecessaryBytes = cutZone;
		}

		if(lengthEmptyTermination == 0){
			//Make sure A and C are normalitzed
			if((((Cr >> 32) & 0xFF) == 0x00) && (NZL == -1)){
				Cr = Cr << 8;
				Ar = Ar << 8;
			}
		}

		//Check if we are out of bounds to know if we should continue (out of bounds => next byte)
		while((necessaryBytes < maxNecessaryBytes) && ((Rf + ((long) 1 << Sf) - 1 < Cr) || (Rf + ((long) 1 << Sf) - 1 >= Cr + Ar))){
			necessaryBytes++;
			if (necessaryBytes <= 4){
				//Decode one byte more
				Sf -= s;
				long b = outputByteStream.getByte(lengthEmptyTermination + necessaryBytes - 1);
				if(b < 0){
					b += 256;
				}
				Rf += b << Sf;
				if(b == 0xFF){
					s = 7;
				}else{
					s = 8;
				}
			}
		}
		return(necessaryBytes);
	}

}
