/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode;
import BOIException.ParameterException;
import BOIException.ErrorException;
import BOIParser.ArgumentsParser;

import java.lang.reflect.Method;


/**
 * Arguments parser for BOI coder (extended from ArgumentsParser).
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class CoderParser extends ArgumentsParser{

	//ARGUMENTS SPECIFICATION
	String[][] coderArguments = {
		{"-i", "--inputImage", "{string}", "", "1", "1",
			"Input image. Valid formats are: pgm, ppm, pbm, jpg, tiff, png, bmp, gif, fpx. If image is raw data file extension must be \".raw\" or \".img\" and \"-ig\" parameter is mandatory."
		},
		{"-ig", "--inputImageGeometry", "{int int int int boolean}", "", "0", "1",
			"Geometry of raw image data. Parameters are:\n    1- zSize (number of image components)\n    2- ySize (image height)\n    3- xSize (image width)\n    4- data type. Possible values are:\n      0- boolean (1 byte)\n      1- unsigned int (1 byte)\n      2- unsigned int (2 bytes)\n      3- signed int (2 bytes)\n      4- signed int (4 bytes) ATTENTION: see \"-rt\" parameter\n      5- signed int (8 bytes) ATTENTION: see \"-rt\" parameter\n      6- float (4 bytes) ATTENTION: see \"-rt\" parameter\n      7- double (8 bytes) ATTENTION: see \"-rt\" parameter\n    5- Byte order (0 if BIG ENDIAN, 1 if LITTLE ENDIAN)\n    6- 1 if 3 first components are RGB, 0 otherwise\n    Loading process can be checked by \"-vp\" parameter (see detailed explanation below)."
		},
		{"-o", "--outputFile", "{string}", "same as input with a new extension (see \"-of\" parameter)", "0", "1",
			"Output image file name (specified without extension)."
		},
		{"-oh", "--outputFileHeader", "{int}", CoderDefaultValues.CFileHeader + "", "0", "1",
			"Header type for the output file: \n    0- NOH Empty heading\n    1- JPC JPEG2000 codestream header\n    2- JP2 JPEG2000 codestream header with JP2 box\n    3- JPK BOI header"
		},
		{"-op", "--outputFileProgression", "{int}", CoderDefaultValues.CFileProgression + "", "0", "1",
			"Progression order for the output file: \n    0- LRCP Layer-Resolution-Component-Position\n    1- RLCP Resolution-Layer-Component-Position\n    2- RPCL Resolution-Position-Component-Layer\n    3- PCRL Position-Component-Resolution-Layer\n    4- CPRL Component-Position-Resolution-Layer"
		},
		{"-ct", "--colourTransformType", "{int}", "when first 3 components are RGB, if waveletTransformType is 0 or 1: 1; if waveletTransformType is 2: 2. Otherwise 0.", "0", "1",
			"Colour transform type:\n    0- No colour transform\n    1- Reversible Colour Transform\n    2- Irreversible Colour Transform"
		},
		{"-cc", "--colourTransformComponents", "{int int int}", "if image is RGB with known type (jpeg, bmp, ...) 0 1 2 components have been taken", "0", "1",
			"Components taken to perform colour transformation. First specified must be the Red component, second Green and third Blue."
		},
		{"-wt", "--waveletTransformType", "{int[ int[ int[ ...]]]}", CoderDefaultValues.WTType + "", "0", "1",
			"Discrete wavelet transform type for each image component. First value is for the first component, second value for the second component and so on. If only one value is specified, wavelete transform type will be the same for all components. Valid values are:\n    0- No wavelet transform\n    1- Reversible 5/3 DWT\n    2- Irreversible 9/7 DWT"
		},
		{"-wl", "--waveletTransformLevels", "{int[ int[ int[ ...]]]}", "min(" + CoderDefaultValues.WTLevels + ", maximum WT levels that image size allows)", "0", "1",
			"Discrete wavelet transform levels for each image component. First value is for the first component, second value for the second component and so on. If only one value is specified, wavelet transform levels will be the same for all components."
		},
		{"-qt", "--quantizationType", "{int[ int[ int[ ...]]]}", "(0 if waveletTransformType (\"-wt\" parameter) is 1, 2 if waveletTransformType is 2)", "0", "1",
			"Quantization type for each image component. First value is for the first component, second for the second component and so on. If only one value is specified, type will be used for all image components. Valid values are:\n    0- Reversible quantization\n    1- Irreversible quantization (derived calculation)\n    2- Irreversible quantization (expounded calculation performed on L2-norms)"
		},
		{"-qg", "--guardBits", "{int}", CoderDefaultValues.QGuardBits + "", "0", "1",
			"Number of guard bits used in quantization step. It is advisable that this value be equal or greater than 1 to avoid possible out of range errors in the quantization process."
		},
		{"-bw", "--blockWidth", "{int[ int[ int[ ...]]]}", "min(" + CoderDefaultValues.BDBlockWidth + ", maximum width that subband size allows)", "0", "1",
			"Block width specification for each image component. Value is an exponent of 2. If only one value is specified, it will be used for all image components. If specified width is greater than subband width, it will take subband width."
		},
		{"-bh", "--blockHeight", "{int[ int[ int[ ...]]]}", "min(" + CoderDefaultValues.BDBlockHeight + ", maximum height that subband size allows)", "0", "1",
			"Block height specification for each image component. Value is an exponent of 2. If only one value is specified, it will be used for all image components. If specified height is greater than subband height, it will take subband height."
		},
		{"-pw", "--precinctWidth", "{int[ int[ int[ ...]]]}", "maximum allowed for image width", "0", "1",
			"Precinct width specification for each image component. Value is an exponent of 2. If only one value is specified, it will be used for all image components. The value specifies precinct width of the original image: then precinct size of each resolution level will be internally calculated. This value must be equal or greater than block width (\"-bw\" parameter) + 1."
		},
		{"-ph", "--precinctHeight", "{int[ int[ int[ ...]]]}", "maximum allowed for image height", "0", "1",
			"Precinct height specification for each image component. Value is an exponent of 2. If only one value is specified, it will be used for all image components. The value specifies precinct height of the original image: then precinct size of each resolution level will be internally calculated. This value must be equal or greater than block height (\"-bh\" parameter) + 1."
		},
		{"-lt", "--layerCalculationType", "{int}", CoderDefaultValues.LCType + "", "0", "1",
			"Calculation type to build quality layers and for rate allocation. There are three ways to perform this operation:\n    0- Generic: rate allocation and layers are built generically using rate distortion measures (optimal rate control strategy defined in EBCOT). Layers' size are calculated through an exponential function. Parameters layersNumber (\"-ln\") and targetBytes (\"-tb\") can be taken into account simultaneously. A bistream with the specified length containing the number of layers passed (or taken by default) will be generated. In this way, if we don't specify anything, only a single layer with maximum length will be built. If we only specify layersNumber, a maximum length file will be built containing the number of layers passed (each layer will contain approximately the same number of bytes). If we only specify targetBytes, a bitstream of desired length will be built with only one layer.\n    1- BitPlane: rate allocation and layers are constructed from MSB to LSB bit plane, without using BPE rate distortion measures. In this way, targetBytes and layersNumber are exclusive. If we specify targetBytes a bitstream with specified length will be built and it will contain as bit planes as possible. If we specify layersNumber, a bitstream with specified bit planes will be built (each one will contain all coding passes for each bit plane). With layerSingle parameter (\"-ls\") it is possible to separate bit plane coding passes in different layers (one layer for each bit plane).\n    2- CodingPasses: same as above but layers are constructed by coding passes (one layer per coding pass). ATTENTION: Use this mode is very efficient computationally than the others because CPI algorithms are using in the BPE+MQ coding stages (when \"-cpi\" == 1).\n    3- CodingPasses SPP and MRP+CP: same as above but layers are allocated at coding passes of type SPP or MRP+CP (one layer per coding passes SPP and one layer per coding passes MRP+CP)."
		},
		{"-ln", "--layersNumber", "{int}", CoderDefaultValues.LCTargetNumLayers + "", "0", "1",
			"Number of layers for the final bitstream (0 indicates that it doesn't matter). See \"-lt\" parameter for more information."
		},
		{"-ls", "--layerSingle", "{boolean}", (CoderDefaultValues.LCForceSingleLayer ? "1": "0") + "", "0", "1",
			"With \"-lt\" == 1 we can save each bit plane in a different layer (specify 0 in this parameter). The same for \"-lt\" == 2 (instead of bit planes is for coding passes). This value cannot be specified if we use \"-lt\" == 0. Valid values are 0 (NO forceSingleLayer) or 1 (forceSinglelayer)."
		},
		{"-tb", "--targetBytes", "{int}", CoderDefaultValues.LCTargetNumBytes +  "", "0", "1",
			"Number of target bytes for the compressed bitstream. Computations are done without taking into account headings, so file size will be slightly larger than this number of bytes (0 indicates that size file is not important -all compressed data will be stored-). See \"-lt\" parameter for more information."
		},
		{"-vc", "--verboseComputation", "{boolean}", (CoderDefaultValues.CVerbose[0] ? "1": "0") + " ", "0", "1",
			"Show some information about time and used memory for each compression stage. Value is a boolean: 0 indicates NO show and 1 indicates show."
		},
		{"-ve", "--verboseEncodingParameters", "{boolean}", (CoderDefaultValues.CVerbose[1] ? "1": "0") + " ", "0", "1",
			"Show BOI encoding parameters. Value is a boolean: 0 indicates NO show and 1 indicates show."
		},
		{"-tn", "--threadNumber", "{boolean}", CoderDefaultValues.CThreadNumber + " ", "0", "1",
			"Number of threads employed in the block coding stage."
		},
		{"-h", "--help", "", "", "0", "1",
			"Displays this help and exits program."
		}
	};

	//ARGUMENTS VARIABLES
	String imageFile = "";
	String COutFile = "";
	int[] imageGeometry = null;
	int CFileHeader = -1;
	int CFileProgression = -1;
	int CTType = -1;
	int[] CTComponents = null;
	int[] WTTypes = null;
	int[] WTLevels = null;
	int[] QTypes = null;
	int QGuardBits = -1;
	int[] BDBlockWidths = null;
	int[] BDBlockHeights = null;
	int[] BDPrecinctWidths = null;
	int[] BDPrecinctHeights = null;
	int LCType = -1;
	int LCTargetNumLayers = -1;
	int LCForceSingleLayer = -1;
	int LCTargetNumBytes = -1;
	int CThreadNumber = -1;
	boolean[] CVerbose = null;


	/**
	 * Receives program arguments and parses it, setting to arguments variables.
	 *
	 * @param arguments the array of strings passed at the command line
	 *
	 * @throws ParameterException when an invalid parsing is detected
	 * @throws ErrorException when some problem with method invocation occurs
	 */
	public CoderParser(String[] arguments) throws ParameterException, ErrorException{
		try{
			Method m = this.getClass().getMethod("parseArgument", new Class[] {int.class, String[].class});
			parse(coderArguments, arguments, this, m);
		}catch(NoSuchMethodException e){
			throw new ErrorException("Coder parser error invoking parse function.");
		}
	}

	/**
	 * Parse an argument using parse functions from super class and put its value/s to the desired variable. This function is called from parse function of the super class.
	 *
	 * @param argFound number of parameter (the index of the array coderArguments)
	 * @param options the command line options of the argument
	 *
	 * @throws ParameterException when some error about parameters passed (type, number of params, etc.) occurs
	 */
	public void parseArgument(int argFound, String[] options) throws ParameterException{
		switch(argFound){
		case  0: //-i  --inputImage
			imageFile = parseString(options);
			if(imageFile.endsWith(".raw") || imageFile.endsWith(".img")){
				coderArguments[1][4] = "1";
			}
			break;
		case  1: //-ig  --inputImageGeometry
			imageGeometry = parseIntegerArray(options, 6);
			break;
		case  2: //-o  --outputFile
			COutFile = parseString(options);
			break;
		case  3: //-oh  --outputFileHeader
			CFileHeader = parseIntegerPositive(options);
			break;
		case  4: //-op  --outputFileProgression
			CFileProgression = parseIntegerPositive(options);
			break;
		case  5: //-ct  --colourTransformType
			CTType = parseIntegerPositive(options);
			break;
		case  6: //-cc --colourTransformComponents
			CTComponents = parseIntegerArray(options, 3);
			break;
		case  7: //-wt  --waveletTransformType
			WTTypes = parseIntegerArray(options);
			break;
		case  8: //-wl --waveletTransformLevels
			WTLevels = parseIntegerArray(options);
			break;
		case  9: //-qt --quantizationType
			QTypes = parseIntegerArray(options);
			break;
		case 10: //-qg --guardBits
			QGuardBits = parseIntegerPositive(options);
			break;
		case 11: //-bw --blockWidth
			BDBlockWidths = parseIntegerArray(options);
			break;
		case 12: //-bh --blockHeight
			BDBlockHeights = parseIntegerArray(options);
			break;
		case 13: //-pw --precinctWidth
			BDPrecinctWidths = parseIntegerArray(options);
			break;
		case 14: //-ph --precinctHeight
			BDPrecinctHeights = parseIntegerArray(options);
			break;
		case 15: //-lt --layerCalculationType
			LCType = parseIntegerPositive(options);
			break;
		case 16: //-ln --layersNumber
			LCTargetNumLayers = parseIntegerPositive(options);
			break;
		case 17: //-ls --layerSingle
			LCForceSingleLayer = parseBoolean(options) ? 1: 0;
			break;
		case 18: //-tb --targetBytes
			LCTargetNumBytes = parseIntegerPositive(options);
			break;
		case 19: //-vc  --verboseComputation
			if(CVerbose == null){
				CVerbose = CoderDefaultValues.CVerbose;
			}
			CVerbose[0] = parseBoolean(options);
			break;
		case 20: //-ve  --verboseParameters
			if(CVerbose == null){
				CVerbose = CoderDefaultValues.CVerbose;
			}
			CVerbose[1] = parseBoolean(options);
			break;
		case 21: //-tn  --threadNumber
			CThreadNumber = parseIntegerPositive(options);
			break;
		case 22: //-h  --help
			System.out.println("BOI Coder 1.6");
			showArgsInfo();
			//showArgsInfoLatexTable();
			System.exit(0);
			break;
		}
	}

	//ARGUMENTS GET FUNCTIONS
	public String getImageFile(){
		return(imageFile);
	}
	public String getCOutFile(){
		return(COutFile);
	}
	public int[] getImageGeometry(){
		return(imageGeometry);
	}
	public int getCFileHeader(){
		return(CFileHeader);
	}
	public int getCFileProgression(){
		return(CFileProgression);
	}
	public int getCTType(){
		return(CTType);
	}
	public int[] getCTComponents(){
		return(CTComponents);
	}
	public int[] getWTTypes(){
		return(WTTypes);
	}
	public int[] getWTLevels(){
		return(WTLevels);
	}
	public int[] getQTypes(){
		return(QTypes);
	}
	public int getQGuardBits(){
		return(QGuardBits);
	}
	public int[] getBDBlockWidths(){
		return(BDBlockWidths);
	}
	public int[] getBDBlockHeights(){
		return(BDBlockHeights);
	}
	public int[] getBDPrecinctWidths(){
		return(BDPrecinctWidths);
	}
	public int[] getBDPrecinctHeights(){
		return(BDPrecinctHeights);
	}
	public int getLCType(){
		return(LCType);
	}
	public int getLCTargetNumLayers(){
		return(LCTargetNumLayers);
	}
	public int getLCForceSingleLayer(){
		return(LCForceSingleLayer);
	}
	public int getLCTargetNumBytes(){
		return(LCTargetNumBytes);
	}
	public int getCThreadNumber(){
		return(CThreadNumber);
	}
	public boolean[] getCVerbose(){
		return(CVerbose);
	}

	//OTHER MISCELLANEOUS FUNCTIONS
	/**
	 * This function converts an array of integer that represents components (p.e. if array is 0,2 indicates component 0 and 2) to an array of booleans that indicates if a component is marked or not (p.e. 0,2,4 is true,false,true,false,true) .
	 *
	 * @param intComponents an array of integers representing selected components
	 * @param minSize minium array size of components
	 * @return an array of booleans representing the components passed
	 */
	static public boolean[] integerToBooleanComponents(int[] intComponents, int minSize){
		boolean[] boolComponents = null;

		if(intComponents != null){
			int maxComponent = 0;
			for(int i = 0; i < intComponents.length; i++){
				if(maxComponent < intComponents[i]){
					maxComponent = intComponents[i];
				}
			}
			boolComponents = new boolean[maxComponent+1 > minSize? maxComponent+1: minSize];
			for(int i = 0; i < boolComponents.length; i++){
				boolComponents[i] = false;
			}
			for(int i = 0; i < intComponents.length; i++){
				boolComponents[intComponents[i]] = true;
			}
		}
		return(boolComponents);
	}

}

