/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOICode.FileWrite;
import BOIException.WarningException;
import BOIStream.BitStream;
import BOIStream.ByteStream;


/**
 * This class generates JPK headers from the BOI coder. This heading is useful for some parameters/options allowed in BOI and not allowed in JPEG2000 standard headings. Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; run<br>
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class JPKHeading{

	/**
	 * All following variables have the following structure:
	 *
	 * type nameVar; //variable to be saved at the heading
	 * final type nameVar_BITS; //number of bits allowed for this variable in the heading - its range will be from 0 to 2^nameVar_BITS, otherwise a WarningException will be thrown and heading will not be generated
	 */

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;
	final int zSize_BITS = 14;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;
	final int ySize_BITS = 20;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;
	final int xSize_BITS = 20;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift} (see source code)
	 */
	boolean[] LSSignedComponents = null;
	//final int LSSignedComponents_BITS = 1;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;
	final int CTType_BITS = 4;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 */
	int[] CTComponents = null;
	final int CTComponents_BITS = zSize_BITS;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	int[] WTTypes = null;
	final int WTTypes_BITS = 4;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	int[] WTLevels = null;
	final int WTLevels_BITS = 5; //log_2(max(ySize_BITS, xSize_BITS))

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	int[] QTypes = null;
	final int QTypes_BITS = 4;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;
	final int QComponentsBits_BITS = 6;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QExponents}
	 */
	int[][][] QExponents = null;
	final int QExponents_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QMantisas}
	 */
	int[][][] QMantisas = null;
	final int QMantisas_BITS = 12;

	/**
	 * Definition in {@link BOICode.Transform.Quantization} (see source code)
	 */
	int QGuardBits;
	final int QGuardBits_BITS = 4;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	int[] BDBlockWidths = null;
	final int BDBlockWidths_BITS = 5; //log_2(max(ySize_BITS, xSize_BITS))

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	int[] BDBlockHeights = null;
	final int BDBlockHeights_BITS = 5; //log_2(max(ySize_BITS, xSize_BITS))

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 */
	int[][] BDResolutionPrecinctWidths = null;
	final int BDResolutionPrecinctWidths_BITS = 5; //log_2(max(ySize_BITS, xSize_BITS))

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 */
	int[][] BDResolutionPrecinctHeights = null;
	final int BDResolutionPrecinctHeights_BITS = 5; //log_2(max(ySize_BITS, xSize_BITS))

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 */
	int LCAchievedNumLayers;
	final int LCAchievedNumLayers_BITS = 12;

	/**
	 * Definition in {@link BOICode.Coder#CFileProgression}
	 */
	int FWProgressionOrder;
	final int FWProgressionOrder_BITS = 4;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	boolean[] FWPacketHeaders = null;
	//final int FWPacketHeaders_BITS = 1;


	/**
	 * Constructor of JPKHeading. It receives the information about the compressed image needed to be put in JPKHeading.
	 *
	 * @param zSize definition in {@link BOICode.Coder#zSize}
	 * @param ySize definition in {@link BOICode.Coder#ySize}
	 * @param xSize definition in {@link BOICode.Coder#xSize}
	 * @param LSSignedComponents definition in {@link BOICode.Transform.LevelShift} (see source code)
	 * @param CTType definition in {@link BOICode.Transform.ColourTransform#CTType}
	 * @param CTComponents definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 * @param WTTypes definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 * @param WTLevels definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 * @param QTypes definition in {@link BOICode.Transform.Quantization#QTypes}
	 * @param QComponentsBits definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 * @param QExponents definition in {@link BOICode.Transform.Quantization#QExponents}
	 * @param QMantisas definition in {@link BOICode.Transform.Quantization#QMantisas}
	 * @param QGuardBits definition in {@link BOICode.Transform.Quantization} (see source code)
	 * @param BDBlockWidths definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 * @param BDBlockHeights definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 * @param BDResolutionPrecinctWidths definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 * @param BDResolutionPrecinctHeights definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 * @param LCAchievedNumLayers definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 * @param FWProgressionOrder definition in {@link BOICode.Coder#CFileProgression}
	 * @param FWPacketHeaders definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	public JPKHeading(
	int zSize,
	int ySize,
	int xSize,
	boolean[] LSSignedComponents,
	int CTType,
	int[] CTComponents,
	int[] WTTypes,
	int[] WTLevels,
	int[] QTypes,
	int[] QComponentsBits,
	int[][][] QExponents,
	int[][][] QMantisas,
	int QGuardBits,
	int[] BDBlockWidths,
	int[] BDBlockHeights,
	int[][] BDResolutionPrecinctWidths,
	int[][] BDResolutionPrecinctHeights,
	int LCAchievedNumLayers,
	int FWProgressionOrder,
	boolean[] FWPacketHeaders
	){
		//Parameters copy
		this.zSize = zSize;
		this.ySize = ySize;
		this.xSize = xSize;
		this.LSSignedComponents = LSSignedComponents;
		this.CTType = CTType;
		this.CTComponents = CTComponents;
		this.WTTypes = WTTypes;
		this.WTLevels = WTLevels;
		this.QTypes = QTypes;
		this.QComponentsBits = QComponentsBits;
		this.QExponents = QExponents;
		this.QMantisas = QMantisas;
		this.QGuardBits = QGuardBits;
		this.BDBlockWidths = BDBlockWidths;
		this.BDBlockHeights = BDBlockHeights;
		this.BDResolutionPrecinctWidths = BDResolutionPrecinctWidths;
		this.BDResolutionPrecinctHeights = BDResolutionPrecinctHeights;
		this.LCAchievedNumLayers = LCAchievedNumLayers;
		this.FWProgressionOrder = FWProgressionOrder;
		this.FWPacketHeaders = FWPacketHeaders;
	}

	/**
	 * Generates the JPK heading.
	 *
	 * @return the generate ByteStream
	 * @throws WarningException when the heading can not be generated due to some variable exceeds the maximum allowed range
	 */
	public ByteStream run() throws WarningException{
		BitStream JPKHeading = new BitStream();

		//zSize
		if((zSize < 0) || (zSize >= (int)Math.pow(2, zSize_BITS))){
			throw new WarningException("Wrong zSize.");
		}
		JPKHeading.addBits(zSize, zSize_BITS);

		//ySize
		if((ySize < 0) || (ySize >= (int)Math.pow(2, ySize_BITS))){
			throw new WarningException("Wrong ySize.");
		}
		JPKHeading.addBits(ySize, ySize_BITS);

		//xSize
		if((xSize < 0) || (xSize >= (int)Math.pow(2, xSize_BITS))){
			throw new WarningException("Wrong xSize.");
		}
		JPKHeading.addBits(xSize, xSize_BITS);

		//LSSignedComponents
		//Only put all components LSSignedComponents in heading if all of them are not equal
		boolean LSSignedComponents_ALL = true;
		for(int z = 0; z < zSize; z++){
			if(!LSSignedComponents[z]){
				LSSignedComponents_ALL = false;
			}
		}
		JPKHeading.addBit(LSSignedComponents_ALL);
		if(!LSSignedComponents_ALL){
			for(int z = 0; z < zSize; z++){
				JPKHeading.addBit(LSSignedComponents[z]);
			}
		}

		//CTType
		if((CTType < 0) || (CTType >= (int)Math.pow(2, CTType_BITS))){
			throw new WarningException("Wrong CTType.");
		}
		JPKHeading.addBits(CTType, CTType_BITS);

		//CTComponents
		for(int i = 0; i < CTComponents.length; i++){
			if((CTComponents[i] < 0) || (CTComponents[i] >= (int)Math.pow(2, CTComponents_BITS))){
				throw new WarningException("Wrong CTComponents.");
			}
			JPKHeading.addBits(CTComponents[i], CTComponents_BITS);
		}

		//WTTypes
		//Only put all WTTypes in heading if they are different
		boolean WTypes_ALL = true;
		int WTType = WTTypes[0];
		for(int z = 1; z < zSize; z++){
			if(WTType != WTTypes[z]){
				WTypes_ALL = false;
			}
		}
		JPKHeading.addBit(WTypes_ALL);
		if(WTypes_ALL){
			if((WTTypes[0] < 0) || (WTTypes[0] >= (int)Math.pow(2, WTTypes_BITS))){
				throw new WarningException("Wrong WTTypes.");
			}
			JPKHeading.addBits(WTTypes[0], WTTypes_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((WTTypes[z] < 0) || (WTTypes[z] >= (int)Math.pow(2, WTTypes_BITS))){
					throw new WarningException("Wrong WTTypes.");
				}
				JPKHeading.addBits(WTTypes[z], WTTypes_BITS);
			}
		}

		//WTLevels
		//Only put all WTLevels in heading if they are different
		boolean WTLevels_ALL = true;
		int WTLevel = WTLevels[0];
		for(int z = 1; z < zSize; z++){
			if(WTLevel != WTLevels[z]){
				WTLevels_ALL = false;
			}
		}
		JPKHeading.addBit(WTLevels_ALL);
		if(WTLevels_ALL){
			if((WTLevels[0] < 0) || (WTLevels[0] >= (int)Math.pow(2, WTLevels_BITS))){
				throw new WarningException("Wrong WTLevels.");
			}
			JPKHeading.addBits(WTLevels[0], WTLevels_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((WTLevels[z] < 0) || (WTLevels[z] >= (int)Math.pow(2, WTLevels_BITS))){
					throw new WarningException("Wrong WTLevels.");
				}
				JPKHeading.addBits(WTLevels[z], WTLevels_BITS);
			}
		}

		//QTypes
		//Only put all QTypes in heading if they are different
		boolean QTypes_ALL = true;
		int QType = QTypes[0];
		for(int z = 1; z < zSize; z++){
			if(QType != QTypes[z]){
				QTypes_ALL = false;
			}
		}
		JPKHeading.addBit(QTypes_ALL);
		if(QTypes_ALL){
			if((QTypes[0] < 0) || (QTypes[0] >= (int)Math.pow(2, QTypes_BITS))){
				throw new WarningException("Wrong QTypes.");
			}
			JPKHeading.addBits(QTypes[0], QTypes_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((QTypes[z] < 0) || (QTypes[z] >= (int)Math.pow(2, QTypes_BITS))){
					throw new WarningException("Wrong QTypes.");
				}
				JPKHeading.addBits(QTypes[z], QTypes_BITS);
			}
		}

		//QComponentsBits
		//Only put all QComponentsBits in heading if they are different
		boolean QComponentsBits_ALL = true;
		int QComponentBits = QComponentsBits[0];
		for(int z = 1; z < zSize; z++){
			if(QComponentBits != QComponentsBits[z]){
				QComponentsBits_ALL = false;
			}
		}
		JPKHeading.addBit(QComponentsBits_ALL);
		if(QComponentsBits_ALL){
			if((QComponentsBits[0] < 0) || (QComponentsBits[0] >= (int)Math.pow(2, QComponentsBits_BITS))){
				throw new WarningException("Wrong QComponentsBits.");
			}
			JPKHeading.addBits(QComponentsBits[0], QComponentsBits_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((QComponentsBits[z] < 0) || (QComponentsBits[z] >= (int)Math.pow(2, QComponentsBits_BITS))){
					throw new WarningException("Wrong QComponentsBits.");
				}
				JPKHeading.addBits(QComponentsBits[z], QComponentsBits_BITS);
			}
		}

		//QExponents
		//Only put all QExponents in heading if they are different
		boolean QExponents_ALL = true;
		if(WTLevels_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
			for(int subband = 0; subband < QExponents[0][rLevel].length; subband++){
				int QExponent = QExponents[0][rLevel][subband];
				for(int z = 1; z < zSize; z++){
					if(QExponent != QExponents[z][rLevel][subband]){
						QExponents_ALL = false;
					}
				}
			}}
		}else{
			QExponents_ALL = false;
		}
		JPKHeading.addBit(QExponents_ALL);
		if(QExponents_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
			for(int subband = 0; subband < QExponents[0][rLevel].length; subband++){
				if((QExponents[0][rLevel][subband] < 0) || (QExponents[0][rLevel][subband] >= (int)Math.pow(2, QExponents_BITS))){
					throw new WarningException("Wrong QExponents.");
				}
				JPKHeading.addBits(QExponents[0][rLevel][subband], QExponents_BITS);
			}}
		}else{
			for(int z = 0; z < zSize; z++){
			for(int rLevel = 0; rLevel <= WTLevels[z]; rLevel++){
			for(int subband = 0; subband < QExponents[z][rLevel].length; subband++){
				if((QExponents[z][rLevel][subband] < 0) || (QExponents[z][rLevel][subband] >= (int)Math.pow(2, QExponents_BITS))){
					throw new WarningException("Wrong QExponents.");
				}
				JPKHeading.addBits(QExponents[z][rLevel][subband], QExponents_BITS);
			}}}
		}

		//QMantisas
		//Only put all QMantisas in heading if they are different
		boolean QMantisas_ALL = true;
		if(WTLevels_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
			for(int subband = 0; subband < QMantisas[0][rLevel].length; subband++){
				int QMantisa = QMantisas[0][rLevel][subband];
				for(int z = 1; z < zSize; z++){
					if(QMantisa != QMantisas[z][rLevel][subband]){
						QMantisas_ALL = false;
					}
				}
			}}
		}else{
			QMantisas_ALL = false;
		}
		JPKHeading.addBit(QMantisas_ALL);
		if(QMantisas_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
			for(int subband = 0; subband < QMantisas[0][rLevel].length; subband++){
				if((QMantisas[0][rLevel][subband] < 0) || (QMantisas[0][rLevel][subband] >= (int)Math.pow(2, QMantisas_BITS))){
					throw new WarningException("Wrong QMantisas.");
				}
				JPKHeading.addBits(QMantisas[0][rLevel][subband], QMantisas_BITS);
			}}
		}else{
			for(int z = 0; z < zSize; z++){
			for(int rLevel = 0; rLevel <= WTLevels[z]; rLevel++){
			for(int subband = 0; subband < QMantisas[z][rLevel].length; subband++){
				if((QMantisas[z][rLevel][subband] < 0) || (QMantisas[z][rLevel][subband] >= (int)Math.pow(2, QMantisas_BITS))){
					throw new WarningException("Wrong QMantisas.");
				}
				JPKHeading.addBits(QMantisas[z][rLevel][subband], QMantisas_BITS);
			}}}
		}

		//QGuardBits
		if((QGuardBits < 0) || (QGuardBits >= (int)Math.pow(2, QGuardBits_BITS))){
			throw new WarningException("Wrong QGuardBits.");
		}
		JPKHeading.addBits(QGuardBits, QGuardBits_BITS);

		//BDBlockWidths
		//Only put all BDBlockWidths in heading if they are different
		boolean BDBlockWidths_ALL = true;
		int BDBlockWidth = BDBlockWidths[0];
		for(int z = 1; z < zSize; z++){
			if(BDBlockWidth != BDBlockWidths[z]){
				BDBlockWidths_ALL = false;
			}
		}
		JPKHeading.addBit(BDBlockWidths_ALL);
		if(BDBlockWidths_ALL){
			if((BDBlockWidths[0] < 0) || (BDBlockWidths[0] >= (int)Math.pow(2, BDBlockWidths_BITS))){
				throw new WarningException("Wrong BDBlockWidths.");
			}
			JPKHeading.addBits(BDBlockWidths[0], BDBlockWidths_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((BDBlockWidths[z] < 0) || (BDBlockWidths[z] >= (int)Math.pow(2, BDBlockWidths_BITS))){
					throw new WarningException("Wrong BDBlockWidths.");
				}
				JPKHeading.addBits(BDBlockWidths[z], BDBlockWidths_BITS);
			}
		}

		//BDBlockHeights
		//Only put all BDBlockHeights in heading if they are different
		boolean BDBlockHeights_ALL = true;
		int BDBlockHeight = BDBlockHeights[0];
		for(int z = 1; z < zSize; z++){
			if(BDBlockHeight != BDBlockHeights[z]){
				BDBlockHeights_ALL = false;
			}
		}
		JPKHeading.addBit(BDBlockHeights_ALL);
		if(BDBlockHeights_ALL){
			if((BDBlockHeights[0] < 0) || (BDBlockHeights[0] >= (int)Math.pow(2, BDBlockHeights_BITS))){
				throw new WarningException("Wrong BDBlockHeights.");
			}
			JPKHeading.addBits(BDBlockHeights[0], BDBlockHeights_BITS);
		}else{
			for(int z = 0; z < zSize; z++){
				if((BDBlockHeights[z] < 0) || (BDBlockHeights[z] >= (int)Math.pow(2, BDBlockHeights_BITS))){
					throw new WarningException("Wrong BDBlockHeights.");
				}
				JPKHeading.addBits(BDBlockHeights[z], BDBlockHeights_BITS);
			}
		}

		//BDResolutionPrecinctWidths
		//Only put all BDResolutionPrecinctWidths in heading if they are different
		boolean BDResolutionPrecinctWidths_ALL = true;
		if(WTLevels_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
				int BDResolutionPrecinctWidth = BDResolutionPrecinctWidths[0][rLevel];
				for(int z = 1; z < zSize; z++){
					if(BDResolutionPrecinctWidth != BDResolutionPrecinctWidths[z][rLevel]){
						BDResolutionPrecinctWidths_ALL = false;
					}
				}
			}
		}else{
			BDResolutionPrecinctWidths_ALL = false;
		}
		JPKHeading.addBit(BDResolutionPrecinctWidths_ALL);
		if(BDResolutionPrecinctWidths_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
				if((BDResolutionPrecinctWidths[0][rLevel] < 0) || (BDResolutionPrecinctWidths[0][rLevel] >= (int)Math.pow(2, BDResolutionPrecinctWidths_BITS))){
					throw new WarningException("Wrong BDResolutionPrecinctWidths.");
				}
				JPKHeading.addBits(BDResolutionPrecinctWidths[0][rLevel], BDResolutionPrecinctWidths_BITS);
			}
		}else{
			for(int z = 0; z < zSize; z++){
			for(int rLevel = 0; rLevel <= WTLevels[z]; rLevel++){
				if((BDResolutionPrecinctWidths[z][rLevel] < 0) || (BDResolutionPrecinctWidths[z][rLevel] >= (int)Math.pow(2, BDResolutionPrecinctWidths_BITS))){
					throw new WarningException("Wrong BDResolutionPrecinctWidths.");
				}
				JPKHeading.addBits(BDResolutionPrecinctWidths[z][rLevel], BDResolutionPrecinctWidths_BITS);
			}}
		}

		//BDResolutionPrecinctHeights
		//Only put all BDResolutionPrecinctHeights in heading if they are different
		boolean BDResolutionPrecinctHeights_ALL = true;
		if(WTLevels_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
				int BDResolutionPrecinctHeight = BDResolutionPrecinctHeights[0][rLevel];
				for(int z = 1; z < zSize; z++){
					if(BDResolutionPrecinctHeight != BDResolutionPrecinctHeights[z][rLevel]){
						BDResolutionPrecinctHeights_ALL = false;
					}
				}
			}
		}else{
			BDResolutionPrecinctHeights_ALL = false;
		}
		JPKHeading.addBit(BDResolutionPrecinctHeights_ALL);
		if(BDResolutionPrecinctHeights_ALL){
			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
				if((BDResolutionPrecinctHeights[0][rLevel] < 0) || (BDResolutionPrecinctHeights[0][rLevel] >= (int)Math.pow(2, BDResolutionPrecinctHeights_BITS))){
					throw new WarningException("Wrong BDResolutionPrecinctHeights.");
				}
				JPKHeading.addBits(BDResolutionPrecinctHeights[0][rLevel], BDResolutionPrecinctHeights_BITS);
			}
		}else{
			for(int z = 0; z < zSize; z++){
			for(int rLevel = 0; rLevel <= WTLevels[z]; rLevel++){
				if((BDResolutionPrecinctHeights[z][rLevel] < 0) || (BDResolutionPrecinctHeights[z][rLevel] >= (int)Math.pow(2, BDResolutionPrecinctHeights_BITS))){
					throw new WarningException("Wrong BDResolutionPrecinctHeights.");
				}
				JPKHeading.addBits(BDResolutionPrecinctHeights[z][rLevel], BDResolutionPrecinctHeights_BITS);
			}}
		}

		//LCAchievedNumLayers
		if((LCAchievedNumLayers < 0) || (LCAchievedNumLayers >= (int)Math.pow(2, LCAchievedNumLayers_BITS))){
			throw new WarningException("Wrong LCAchievedNumLayers.");
		}
		JPKHeading.addBits(LCAchievedNumLayers, LCAchievedNumLayers_BITS);

		//FWProgressionOrder
		if((FWProgressionOrder < 0) || (FWProgressionOrder >= (int)Math.pow(2, FWProgressionOrder_BITS))){
			throw new WarningException("Wrong FWProgressionOrder.");
		}
		JPKHeading.addBits(FWProgressionOrder, FWProgressionOrder_BITS);

		//FWPacketHeaders
		JPKHeading.addBit(FWPacketHeaders[0]);
		JPKHeading.addBit(FWPacketHeaders[1]);

		//Change BitStream to ByteStream and return it
		return(new ByteStream(JPKHeading.getBitStream(), (int) JPKHeading.getNumBytes()));
	}

}
