/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode;
import BOIException.ErrorException;
import BOIException.ParameterException;
import BOIException.WarningException;
import BOIFile.SaveFile;


/**
 * Main class of BOIDecode application. BOIDecode is a JPEG2000 decoder.
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class BOIDecode{

	/**
	 * Main method of BOIDecode application. It takes program arguments, loads image and runs BOI decoder and saves recovered image to a file.
	 *
	 * @param args an array of strings that contains program parameters
	 */
	public static void main(String[] args){
		//Parse arguments
		DecoderParser parser = null;
		try{
			parser = new DecoderParser(args);
		}catch(ErrorException e){
			System.out.println("RUN ERROR:");
			e.printStackTrace();
			System.exit(1);
		}catch(ParameterException e){
			System.out.println("ARGUMENTS ERROR: " +  e.getMessage());
			System.exit(2);
		}

		//Get arguments from parser
		String DInFile = parser.getDInFile();
		String[] imageFiles = parser.getImageFiles();
		int[] imageGeometry = parser.getImageGeometry();
		boolean[] CVerbose = parser.getCVerbose();
		int CThreadNumber = parser.getCThreadNumber();

		//Bit depth of image components
		int[] QComponentsBits = null;

		//BOI decoder
		float[][][] imageSamples = null;
		Decoder BOID = new Decoder(DInFile);
		try{
			BOID.setParameters(CVerbose, CThreadNumber);
		}catch(ParameterException e){
			System.out.println("PARAMETERS ERROR: " + e.getMessage());
			System.exit(3);
		}
		try{
			imageSamples = BOID.run();
			QComponentsBits = BOID.getQComponentsBits();
		}catch(ErrorException e){
			System.out.println("RUN ERROR:");
			e.printStackTrace();
			System.exit(4);
		}catch(Exception e){
			System.out.println("RUN LANGUAGE ERROR:");
			e.printStackTrace();
			System.exit(5);
		}
		//Free unused memory
		BOID = null;

		//Save recovered samples
		for(int file = 0; file < imageFiles.length; file++){
			try{
				SaveFile.SaveFileExtension(imageSamples, QComponentsBits, imageFiles[file], imageGeometry, 0);
			}catch(WarningException e){
				System.out.println("IMAGE SAVE WARNING: " + e.getMessage());
				System.out.println("FILE \"" + imageFiles[file] + "\" CANNOT BE SAVED.");
			}
		}
	}

}
