/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode.Decode;
import BOIException.ErrorException;
import BOIException.WarningException;
import BOIStream.ByteStream;


/**
 * This class receives an array of ByteStreams belonging to a code-block and decodes them obtaining the original image samples. Typically it is used in the following way:<br>
 * constructor(receiveing the bytestreams)<br>
 * runAll<br>
 * getBlockSamples<br>
 *
 * You can also use this class to increasingly decode the coding passes: you only have to use the decodeNextCodingPass function as times as coding passes contains the code-block.<br>
 * ATTENTION: In the framework of the JPEG2000 many different code-blocks have to be encoded. It is more efficient to create only one object of this class and change the code-block:<br>
 * constructor(void)<br>
 * swapCodeBlock<br>
 * runAll<br>
 * getBlockSamples<br>
 * swapCodeBlock<br>
 * runAll<br>
 * getBlockSamples<br>
 * ...<br>
 * (of course, the function runAll can be substituted for multiple call of the decodeNextCodingPass function) If this class is used in this last way it is easy to incorporate a multithreading support for the decoding of code-blocks.<br>
 *
 * Since version 1.3 this class uses an optimized way to calculate coefficient contexts. Although it achieves good results in terms of computational complexity, the source code is not very understandable (when computing the coefficient contexts). File BPEDecoder.java.1_2 contains the 1.2 version of this class, that has a more understandable source code. Apart of this, both classes are exactly equal.
 * Since version 1.4, this class deconstructs coefficients using variable BDReconstruction values.
 *
 * @author Francesc Auli-Llinas
 * @version 1.4
 */
public class BPEDecoder{

	/**
	 * Definition in {@link BOICode.Code.BPECoder#CDByteStream}
	 */
	ByteStream CDByteStream = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#blockSamples}
	 */
	float[][] blockSamples = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#ySize}
	 */
	int ySize;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#xSize}
	 */
	int xSize;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#subband}
	 */
	int subband;

	/**
	 * Most Significance Bit Plane of the subband.
	 * <p>
	 * Negative values not allowed.
	 */
	int subbandMSBPlane;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#MSBPlane}
	 */
	int MSBPlane;

	/**
	 * Definition in {@link BOIDecode.Decode.BlockBuild#BCCodingPasses}
	 */
	int codingPasses;

	/**
	 * Whether to use integer processing or not. Integer processing is commonly employed when the Integer Wavelet Transform is used. The main difference with respect to the normal processing is that with integer processing the decoded coefficients are rounded to the nearest integer rather than to leave them as a floating point number (which is caused when BDreconstruction is a value different from 0.5).
	 * <p>
	 * true indicates that integer processing is prefered.
	 */
	boolean intProcessing;

	/**
	 * Reconstruction parameter used to adjust the decoded coefficients when the DWT is used (estimated values through Distortion estimators for bitplane image coding).
	 * <p>
	 * Recommened values between 0 to 1.
	 */
	static final float[][] BDReconstructionDWT =
		{{0.475f},
		{0.415f,0.475f},
		{0.38f,0.44f,0.475f},
		{0.355f,0.415f,0.45f,0.475f},
		{0.335f,0.395f,0.431f,0.456f,0.475f},
		{0.319f,0.38f,0.415f,0.44f,0.459f,0.475f},
		{0.306f,0.366f,0.401f,0.426f,0.446f,0.462f,0.475f},
		{0.294f,0.355f,0.39f,0.415f,0.434f,0.45f,0.463f,0.475f},
		{0.284f,0.344f,0.38f,0.405f,0.424f,0.44f,0.453f,0.465f,0.475f},
		{0.275f,0.335f,0.37f,0.395f,0.415f,0.431f,0.444f,0.456f,0.466f,0.475f},
		{0.267f,0.327f,0.362f,0.387f,0.407f,0.422f,0.436f,0.447f,0.458f,0.467f,0.475f},
		{0.259f,0.319f,0.355f,0.38f,0.399f,0.415f,0.428f,0.44f,0.45f,0.459f,0.467f,0.475f},
		{0.252f,0.312f,0.348f,0.373f,0.392f,0.408f,0.421f,0.433f,0.443f,0.452f,0.46f,0.468f,0.475f},
		{0.246f,0.306f,0.341f,0.366f,0.386f,0.401f,0.415f,0.426f,0.437f,0.446f,0.454f,0.462f,0.469f,0.475f},
		{0.24f,0.3f,0.335f,0.36f,0.38f,0.395f,0.409f,0.42f,0.431f,0.44f,0.448f,0.456f,0.463f,0.469f,0.475f},
		{0.234f,0.294f,0.33f,0.355f,0.374f,0.39f,0.403f,0.415f,0.425f,0.434f,0.442f,0.45f,0.457f,0.463f,0.469f,0.475f},
		{0.229f,0.289f,0.324f,0.349f,0.369f,0.385f,0.398f,0.41f,0.42f,0.429f,0.437f,0.445f,0.452f,0.458f,0.464f,0.47f,0.475f},
		{0.224f,0.284f,0.319f,0.344f,0.364f,0.38f,0.393f,0.405f,0.415f,0.424f,0.432f,0.44f,0.447f,0.453f,0.459f,0.465f,0.47f,0.475f},
		{0.219f,0.279f,0.315f,0.34f,0.359f,0.375f,0.388f,0.4f,0.41f,0.419f,0.428f,0.435f,0.442f,0.448f,0.454f,0.46f,0.465f,0.47f,0.475f},
		{0.215f,0.275f,0.31f,0.335f,0.355f,0.37f,0.384f,0.395f,0.406f,0.415f,0.423f,0.431f,0.438f,0.444f,0.45f,0.456f,0.461f,0.466f,0.471f,0.475f},
		{0.215f,0.271f,0.306f,0.331f,0.35f,0.366f,0.38f,0.391f,0.401f,0.411f,0.419f,0.426f,0.433f,0.44f,0.446f,0.451f,0.457f,0.462f,0.466f,0.471f,0.475f},
		{0.215f,0.267f,0.302f,0.327f,0.346f,0.362f,0.376f,0.387f,0.397f,0.407f,0.415f,0.422f,0.429f,0.436f,0.442f,0.447f,0.453f,0.458f,0.462f,0.467f,0.471f,0.475f},
		{0.215f,0.263f,0.298f,0.323f,0.342f,0.358f,0.372f,0.383f,0.394f,0.403f,0.411f,0.418f,0.425f,0.432f,0.438f,0.443f,0.449f,0.454f,0.458f,0.463f,0.467f,0.471f,0.475f},
		{0.215f,0.259f,0.294f,0.319f,0.339f,0.355f,0.368f,0.38f,0.39f,0.399f,0.407f,0.415f,0.422f,0.428f,0.434f,0.44f,0.445f,0.45f,0.455f,0.459f,0.463f,0.467f,0.471f,0.475f}};

	/**
	 * Reconstruction parameter used to adjust the decoded coefficients when the IWT is used (estimated values through Distortion estimators for bitplane image coding).
	 * <p>
	 * Recommened values between 0 to 1.
	 */
	static final float[][] BDReconstructionIWT =
		{{0.0f},
		{0.415f,0.0f},
		{0.38f,0.44f,0.0f},
		{0.355f,0.415f,0.45f,0.0f},
		{0.335f,0.395f,0.431f,0.456f,0.0f},
		{0.319f,0.38f,0.415f,0.44f,0.459f,0.0f},
		{0.306f,0.366f,0.401f,0.426f,0.446f,0.462f,0.0f},
		{0.294f,0.355f,0.39f,0.415f,0.434f,0.45f,0.463f,0.0f},
		{0.284f,0.344f,0.38f,0.405f,0.424f,0.44f,0.453f,0.465f,0.0f},
		{0.275f,0.335f,0.37f,0.395f,0.415f,0.431f,0.444f,0.456f,0.466f,0.0f},
		{0.267f,0.327f,0.362f,0.387f,0.407f,0.422f,0.436f,0.447f,0.458f,0.467f,0.0f},
		{0.259f,0.319f,0.355f,0.38f,0.399f,0.415f,0.428f,0.44f,0.45f,0.459f,0.467f,0.0f},
		{0.252f,0.312f,0.348f,0.373f,0.392f,0.408f,0.421f,0.433f,0.443f,0.452f,0.46f,0.468f,0.0f},
		{0.246f,0.306f,0.341f,0.366f,0.386f,0.401f,0.415f,0.426f,0.437f,0.446f,0.454f,0.462f,0.469f,0.0f},
		{0.24f,0.3f,0.335f,0.36f,0.38f,0.395f,0.409f,0.42f,0.431f,0.44f,0.448f,0.456f,0.463f,0.469f,0.0f},
		{0.234f,0.294f,0.33f,0.355f,0.374f,0.39f,0.403f,0.415f,0.425f,0.434f,0.442f,0.45f,0.457f,0.463f,0.469f,0.0f},
		{0.229f,0.289f,0.324f,0.349f,0.369f,0.385f,0.398f,0.41f,0.42f,0.429f,0.437f,0.445f,0.452f,0.458f,0.464f,0.47f,0.0f},
		{0.224f,0.284f,0.319f,0.344f,0.364f,0.38f,0.393f,0.405f,0.415f,0.424f,0.432f,0.44f,0.447f,0.453f,0.459f,0.465f,0.47f,0.0f},
		{0.219f,0.279f,0.315f,0.34f,0.359f,0.375f,0.388f,0.4f,0.41f,0.419f,0.428f,0.435f,0.442f,0.448f,0.454f,0.46f,0.465f,0.47f,0.0f},
		{0.215f,0.275f,0.31f,0.335f,0.355f,0.37f,0.384f,0.395f,0.406f,0.415f,0.423f,0.431f,0.438f,0.444f,0.45f,0.456f,0.461f,0.466f,0.471f,0.0f},
		{0.215f,0.271f,0.306f,0.331f,0.35f,0.366f,0.38f,0.391f,0.401f,0.411f,0.419f,0.426f,0.433f,0.44f,0.446f,0.451f,0.457f,0.462f,0.466f,0.471f,0.0f},
		{0.215f,0.267f,0.302f,0.327f,0.346f,0.362f,0.376f,0.387f,0.397f,0.407f,0.415f,0.422f,0.429f,0.436f,0.442f,0.447f,0.453f,0.458f,0.462f,0.467f,0.471f,0.0f},
		{0.215f,0.263f,0.298f,0.323f,0.342f,0.358f,0.372f,0.383f,0.394f,0.403f,0.411f,0.418f,0.425f,0.432f,0.438f,0.443f,0.449f,0.454f,0.458f,0.463f,0.467f,0.471f,0.0f},
		{0.215f,0.259f,0.294f,0.319f,0.339f,0.355f,0.368f,0.38f,0.39f,0.399f,0.407f,0.415f,0.422f,0.428f,0.434f,0.44f,0.445f,0.45f,0.455f,0.459f,0.463f,0.467f,0.471f,0.0f}};

	/**
	 * Definition in {@link BOICode.Code.BPECoder#contextMap}
	 */
	private byte[][] contextMap = null;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#currentBitPlane}
	 */
	private int currentBitPlane;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#currentCodingPass}
	 */
	private int currentCodingPass;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#currentBitPlaneMask}
	 */
	private int currentBitPlaneMask = 0;

	/**
	 * Number of decoded coding passes.
	 * <p>
	 * Positive values.
	 */
	private int decodedPasses = 0;

	/**
	 * Inverse mask of the currentBitPlane. This is, if the original mask is 00100, this inverse is 11100.
	 * <p>
	 * The content is ~(2^(currentBitPlane-1) - 1)
	 */
	private int currentBitPlaneMaskNull = 0;

	/**
	 * Reconsuctrion value for the non transmitted bit of a coefficient.
	 * <p>
	 * Value computed using the above LUTs.
	 */
	private float deqReconstruction = 0.0F;

	/**
	 * Decoder to read from the inputByteStream.
	 * <p>
	 * This Decoder must be compatible with the abstract class Decoder.
	 */
	private Decoder blockDecoder = new MQDecoder();

	/**
	 * Definition in {@link BOICode.Code.BPECoder#stripeHeight}
	 */
	private final int stripeHeight = 4;

	/**
	 * Definition in {@link BOICode.Code.BPECoder#numBitsRunModeBreak}
	 */
	private final int numBitsRunModeBreak = (int) Math.ceil(Math.log(stripeHeight) / Math.log(2));


	/**
	 * Constructor that does not receive anything. Before use the functions of decodeNextCodingPass the function swapCodeBlock must be used.
	 */
	public BPEDecoder(){}

	/**
	 * Constructor that receives block coded bit streams, block sizes, most significant bit plane and block subband.  This information is absolutely necessary to decode bitstream.
	 *
	 * @param CDByteStream definition in {@link #CDByteStream}
	 * @param blockSamples definition in {@link #blockSamples}
	 * @param MSBPlane definition in {@link BOICode.Code.BPECoder#MSBPlane}
	 * @param codingPasses definition in {@link BOIDecode.Decode.BlockBuild#BCCodingPasses}
	 * @param subband definition in {@link #subband}
	 * @param subbandMSBPlane definition in {@link #subbandMSBPlane}
	 * @param intProcessing definition in {@link #intProcessing}
	 *
	 * @throws ErrorException problems with the inputByteStream
	 */
	public BPEDecoder(ByteStream CDByteStream, float[][] blockSamples, int MSBPlane, int codingPasses, int subband, int subbandMSBPlane, boolean intProcessing) throws ErrorException{
		swapCodeBlock(CDByteStream, blockSamples, MSBPlane, codingPasses, subband, subbandMSBPlane, intProcessing);
	}

	/**
	 * Swaps the code-block decoded by this class. Some initializations performed.
	 *
	 * @param CDByteStream definition in {@link #CDByteStream}
	 * @param blockSamples definition in {@link #blockSamples}
	 * @param MSBPlane definition in {@link BOICode.Code.BPECoder#MSBPlane}
	 * @param codingPasses definition in {@link BOIDecode.Decode.BlockBuild#BCCodingPasses}
	 * @param subband definition in {@link #subband}
	 * @param subbandMSBPlane definition in {@link #subbandMSBPlane}
	 * @param intProcessing definition in {@link #intProcessing}
	 *
	 * @throws ErrorException problems with the inputByteStream
	 */
	public void swapCodeBlock(ByteStream CDByteStream, float[][] blockSamples, int MSBPlane, int codingPasses, int subband, int subbandMSBPlane, boolean intProcessing) throws ErrorException{
		//Some checks
		int numBitPlanes = MSBPlane + 1;
		if(MSBPlane < 0){
			throw new ErrorException("MSBPlane cannot be negative.");
		}
		if(subbandMSBPlane >= BDReconstructionDWT.length){
			throw new ErrorException("subbandMSBPlane higher than the allowed in the distortion estimation LUTs.");
		}

		//Data copy
		this.blockSamples = blockSamples;
		this.MSBPlane = MSBPlane;
		this.codingPasses = codingPasses;
		this.subband = subband;
		this.subbandMSBPlane = subbandMSBPlane;
		this.intProcessing = intProcessing;
		this.CDByteStream = CDByteStream;

		this.ySize = blockSamples.length;
		this.xSize = blockSamples[0].length;

		//Memory allocation
		contextMap = new byte[ySize][xSize];

		//Set blockDecoder (we do not create another object, we reutilize the already created decoder)
		blockDecoder.swapInputByteStream(CDByteStream);
		blockDecoder.restart();
		blockDecoder.reset();

		//Initialize current variables
		currentBitPlane = this.MSBPlane;
		currentCodingPass = 2; //CP
		decodedPasses = 0;
	}

	/**
	 * Decode the image samples (all bit planes).
	 *
	 * @throws WarningException when reached the last coding pass
	 * @throws ErrorException problems with the inputByteStream
	 */
	public void runAll() throws WarningException, ErrorException{
		while((currentBitPlane >= 0) && (decodedPasses < codingPasses)){
			decodeNextCodingPass();
		}
	}

	/**
	 * Decodes the next coding pass.
	 *
	 * @throws WarningException when reached the last coding pass
	 * @throws ErrorException problems with the inputByteStream
	 */
	public void decodeNextCodingPass() throws WarningException, ErrorException{
		if(currentBitPlane < 0){
			throw new WarningException("Reached last bit plane.");
		}

		currentBitPlaneMask = (1 << currentBitPlane); //2^bitPlane
		currentBitPlaneMaskNull = ~(currentBitPlaneMask-1);
		//deqReconstruction = (currentBitPlaneMask >>> 1);
		if(!intProcessing){
			deqReconstruction = (float)currentBitPlaneMask * BDReconstructionDWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane];
		}else{
			//Wondering why currentBitPlaneMask - 1 ? Becasue valid dequantization range in integer transforms is [2^P , (2^P) -1], whereas in lineal transforms is [2^P , 2^P).
			deqReconstruction = Math.round ( (float)(currentBitPlaneMask-1) * BDReconstructionIWT[subbandMSBPlane][subbandMSBPlane-currentBitPlane] );
		}

		//System.out.println("\nBP: " + currentBitPlane + " CP: " + currentCodingPass);
		//Decodes the current coding pass
		switch(currentCodingPass){
		case 0: //SPP
			significancePropagationPass();
			break;
		case 1: //MRP
			magnitudeRefinementPass();
			break;
		case 2: //CP
			cleanupPass();
			break;
		}

		//Prepares the current variables for the next coding pass
		if(currentCodingPass == 2){
			currentCodingPass = 0;
			currentBitPlane--;
		}else{
			currentCodingPass++;
		}
		decodedPasses++;
	}

	/**
	 * Same function as {@link BOICode.Code.BPECoder#signEncode}
	 *
	 * @throws ErrorException when end of bitstream is reached
	 */
	private void signDecode(int x, int y) throws ErrorException{
		int up    = y-1 <  0     ? 0 : contextMap[y-1][x] >= -8 && contextMap[y-1][x] <= 8 ? 0 : blockSamples[y-1][x] < 0 ? -1 : 1;
		int down  = y+1 >= ySize ? 0 : contextMap[y+1][x] >= -8 && contextMap[y+1][x] <= 8 ? 0 : blockSamples[y+1][x] < 0 ? -1 : 1;
		int left  = x-1 <  0     ? 0 : contextMap[y][x-1] >= -8 && contextMap[y][x-1] <= 8 ? 0 : blockSamples[y][x-1] < 0 ? -1 : 1;
		int right = x+1 >= xSize ? 0 : contextMap[y][x+1] >= -8 && contextMap[y][x+1] <= 8 ? 0 : blockSamples[y][x+1] < 0 ? -1 : 1;

		int v = up+down    == 0 ? 0 : up+down    < 0 ? -1 : 1;
		int h = left+right == 0 ? 0 : left+right < 0 ? -1 : 1;

		int signContext = 0;
		switch(h){
		case -1:
			switch(v){
			case -1:
				signContext = -14;
				break;
			case 0:
				signContext = -13;
				break;
			case 1:
				signContext = -12;
				break;
			}
			break;
		case 0:
			switch(v){
			case -1:
				signContext = -11;
				break;
			case 0:
				signContext = 10;
				break;
			case 1:
				signContext = 11;
				break;
			}
			break;
		case 1:
			switch(v){
			case -1:
				signContext = 12;
				break;
			case 0:
				signContext = 13;
				break;
			case 1:
				signContext = 14;
				break;
			}
			break;
		}

		boolean sign = blockDecoder.decodeBit(Math.abs(signContext));
		if(signContext > 0 == sign){
			blockSamples[y][x] = -blockSamples[y][x];
		}
		//System.out.print((sign ? "1": "0") + "(" + Math.abs(context) + ")");
	}

	/**
	 * Same function as {@link BOICode.Code.BPECoder#significancePropagationPass}
	 * Minor changes due to differences between coding and decoding.
	 *
	 * @throws ErrorException when end of bitstream reached
	 */
	private void significancePropagationPass() throws ErrorException{
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
			for(int y = yBegin; y < yEnd; y++){
				//Sample decoding
				//contextMap shows the samples that we must visit (only these ones that have significant neighbourhoods)
				if((contextMap[y][x] > 0) && (contextMap[y][x] <= 8)){
					boolean decodedBit = blockDecoder.decodeBit(contextMap[y][x]);
					//System.out.print((decodedBit ? "1(" : "0(") + contextSearch(x, y, 0) + ")");
					if(decodedBit){
						blockSamples[y][x] = currentBitPlaneMask + deqReconstruction;
						signDecode(x, y);

						//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
						updateMap(x, y, false);
					}
				}
			}}
			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Same function as {@link BOICode.Code.BPECoder#magnitudeRefinementPass}
	 * Minor changes due to differences between coding and decoding.
	 *
	 * @throws ErrorException when end of bitstream reached
	 */
	private void magnitudeRefinementPass() throws ErrorException{
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
			for(int y = yBegin; y < yEnd; y++){
				//The coefficient has been significant in the previous bit plane
				if(Math.abs(contextMap[y][x]) >= 15){
					if(contextMap[y][x] > 0){
						boolean decodedBit = blockDecoder.decodeBit(contextMap[y][x]);
						if(decodedBit){
							if(blockSamples[y][x] < 0){
								//blockSamples[y][x] = -((Math.abs(blockSamples[y][x]) | currentBitPlaneMask) | deqReconstruction);
								blockSamples[y][x] = -( (float)((Math.abs((int)blockSamples[y][x]) & currentBitPlaneMaskNull) | currentBitPlaneMask) + deqReconstruction);
							}else{
								//blockSamples[y][x] = ((blockSamples[y][x] | currentBitPlaneMask) | deqReconstruction);
								blockSamples[y][x] = ( (float)(((int)blockSamples[y][x] & currentBitPlaneMaskNull) | currentBitPlaneMask) + deqReconstruction);
							}
						}else{
							if(blockSamples[y][x] < 0){
								//blockSamples[y][x] = -((Math.abs(blockSamples[y][x]) & ~currentBitPlaneMask) | deqReconstruction);
								blockSamples[y][x] = -( (float)((Math.abs((int)blockSamples[y][x]) & currentBitPlaneMaskNull) & ~currentBitPlaneMask) + deqReconstruction);
							}else{
								//blockSamples[y][x] = ((blockSamples[y][x] & ~currentBitPlaneMask) | deqReconstruction);
								blockSamples[y][x] = ( (float)(((int)blockSamples[y][x] & currentBitPlaneMaskNull) & ~currentBitPlaneMask) + deqReconstruction);
							}
						}
						//System.out.print((decodedBit ? "1(" : "0(") + context + ")");
						contextMap[y][x] = 17;

					}else{
						//The coefficient is significant in the current bit plane
						contextMap[y][x] = (byte) -contextMap[y][x];
					}
				}
			}}
			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Same function as {@link BOICode.Code.BPECoder#cleanupPass}
	 * Minor changes due to differences between coding and decoding.
	 *
	 * @throws ErrorException when end of bitstream reached
	 */
	private void cleanupPass() throws ErrorException{
		//First stripe boundaries
		int yBegin = 0;
		int yEnd = stripeHeight < ySize ? stripeHeight: ySize;
		//Stripes
		do{
			for(int x = 0; x < xSize; x++){
				//Test if we can code the column in run mode
				boolean runMode = true;
				if(yEnd - yBegin == stripeHeight){
					//Check if exists a sample with some significant neighbourhood or the column should not be completely scanned. If one of these conditions is true then the run mode cannot be started
					int y = yBegin;
					do{
						if(contextMap[y][x] != 0){
							runMode = false;
						}
						y++;
					}while((runMode) && (y < yEnd));
				}else{
					runMode = false;
				}

				int y = yBegin;
				//If we can start the run mode, we check if the column has any significant coefficient
				if(runMode){
					boolean decodedBit = blockDecoder.decodeBit(9);
					//System.out.print((decodedBit ? "1(" : "0(") + "9)");
					if(decodedBit){
						int significantCoefficient = 0;
						for(int bitRunModeBreak = numBitsRunModeBreak-1; bitRunModeBreak >= 0; bitRunModeBreak--){
							significantCoefficient |= ((blockDecoder.decodeBit(18) ? 1: 0) << bitRunModeBreak);
							//System.out.print("X(18)");
						}
						y += significantCoefficient;
						blockSamples[y][x] = currentBitPlaneMask + deqReconstruction;
						signDecode(x, y);

						//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
						updateMap(x, y, true);
					}else{
						y = yEnd;
					}
				}

				//Decode each coefficient
				while(y < yEnd){
					//contextMap shows the samples that we must visit (only these ones that are not visited yet in the current bit plane)
					if((contextMap[y][x] >= -8) && (contextMap[y][x] <= 0)){
						//In the next bit plane this coefficient must be visitied by the SPP
						contextMap[y][x] = (byte) -contextMap[y][x];
						boolean decodedBit = blockDecoder.decodeBit(contextMap[y][x]);
						//System.out.print((decodedBit ? "1(" : "0(") + contextSearch(x, y, 0) + ")");

						if(decodedBit){
							blockSamples[y][x] = currentBitPlaneMask + deqReconstruction;
							signDecode(x, y);

							//If the sample is significant we have to update contextMap to indicate that the neighbouhoods in the scan must be visited. We will also update the context of the current coefficient.
							updateMap(x, y, true);
						}
					}
					y++;
				}
			}
			//Next stripe boundaries
			yBegin = yEnd;
			yEnd = yBegin + stripeHeight;
			//Reaching the block boundary
			if(yEnd > ySize){
				yEnd = ySize;
			}
		}while(yBegin < ySize);
	}

	/**
	 * Turn neighbourhoods map samples of (x,y) from 0 to 1 if they can be visited in the current scan.
	 *
	 * @param x x position of the sample
	 * @param y y position of the sample
	 * @param CP indicates if the updating is performed by the Cleanup Pass (true, otherwise false)
	*/
	private void updateMap(int x, int y, boolean CP){
		//Context of the current coefficient
		int yNeighBegin = y-1 >= 0 ? y-1: 0;
		int yNeighEnd = y+1 < ySize ? y+1: y;
		int xNeighBegin = x-1 >= 0 ? x-1: 0;
		int xNeighEnd = x+1 < xSize ? x+1: x;
		boolean significantNeighs = false;
		for(int yNeigh = yNeighBegin; yNeigh <= yNeighEnd; yNeigh++){
		for(int xNeigh = xNeighBegin; xNeigh <= xNeighEnd; xNeigh++){
			if(Math.abs(contextMap[yNeigh][xNeigh]) >= 15){
				if(Math.abs(contextMap[yNeigh][xNeigh]) == 15){
					contextMap[yNeigh][xNeigh] = contextMap[yNeigh][xNeigh] < 0 ? (byte) -16: (byte) 16;
				}
				significantNeighs = true;
			}
		}}
		if(significantNeighs){
			contextMap[y][x] = CP ? (byte) 16: (byte) -16;
		}else{
			contextMap[y][x] = CP ? (byte) 15: (byte) -15;
		}

		//Context of the neighboors
		switch(subband){
		case 0: //SUBBAND LL
		case 2: //SUBBAND LH
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 5: (byte) -5;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -8: (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 1: (byte) -1;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -5: (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -8: (byte) 8;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -1: (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -3: (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 3: (byte) -3;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			break;
		case 1: //SUBBAND HL
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 3: (byte) -3;
					break;
				case 1:
				case 2:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 1: (byte) -1;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -3: (byte) 3;
					break;
				case 1:
				case 2:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -3: (byte) 3;
					break;
				case 3:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 5:
				case 6:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -1: (byte) 1;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 1: (byte) -1;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -1: (byte) 1;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -2: (byte) 2;
						break;
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -5: (byte) 5;
					break;
				case 1:
				case 2:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -8: (byte) 8;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 5: (byte) -5;
					break;
				case 1:
				case 2:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -6: (byte) 6;
					break;
				case 3:
				case 4:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				case 5:
				case 6:
				case 7:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -8: (byte) 8;
					break;
				}
			}
			break;
		case 3: //SUBBAND HH
			if(x-1 >= 0){
				//LEFT
				switch(Math.abs(contextMap[y][x-1])){
				case 0:
					contextMap[y][x-1] = CP ? (byte) 1: (byte) -1;
					break;
				case 1:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y][x-1] = contextMap[y][x-1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-LEFT
					switch(Math.abs(contextMap[y-1][x-1])){
					case 0:
						contextMap[y-1][x-1] = CP ? (byte) 3: (byte) -3;
						break;
					case 1:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x-1] = contextMap[y-1][x-1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-LEFT
					switch(Math.abs(contextMap[y+1][x-1])){
					case 0:
						if(CP){
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) 3: (byte) -3;
						}else{
							contextMap[y+1][x-1] = (y+1)%stripeHeight != 0 ? (byte) -3: (byte) 3;
						}
						break;
					case 1:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x-1] = contextMap[y+1][x-1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
			}
			if(x+1 < xSize){
				//RIGHT
				switch(Math.abs(contextMap[y][x+1])){
				case 0:
					contextMap[y][x+1] = CP ? (byte) -1: (byte) 1;
					break;
				case 1:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y][x+1] = contextMap[y][x+1] < 0 ? (byte) -7: (byte) 7;
					break;
				}
				if(y-1 >= 0){
					//UP-RIGHT
					switch(Math.abs(contextMap[y-1][x+1])){
					case 0:
						if(CP){
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) -3: (byte) 3;
						}else{
							contextMap[y-1][x+1] = y%stripeHeight != 0 ? (byte) 3: (byte) -3;
						}
						break;
					case 1:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y-1][x+1] = contextMap[y-1][x+1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
				if(y+1 < ySize){
					//DOWN-RIGHT
					switch(Math.abs(contextMap[y+1][x+1])){
					case 0:
						contextMap[y+1][x+1] = CP ? (byte) -3: (byte) 3;
						break;
					case 1:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -4: (byte) 4;
						break;
					case 2:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -5: (byte) 5;
						break;
					case 3:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -6: (byte) 6;
						break;
					case 4:
					case 5:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -7: (byte) 7;
						break;
					case 6:
					case 7:
						contextMap[y+1][x+1] = contextMap[y+1][x+1] < 0 ? (byte) -8: (byte) 8;
						break;
					}
				}
			}
			//DOWN
			if(y+1 < ySize){
				switch(Math.abs(contextMap[y+1][x])){
				case 0:
					contextMap[y+1][x] = CP ? (byte) -1: (byte) 1;
					break;
				case 1:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y+1][x] = contextMap[y+1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			//UP
			if(y-1 >= 0){
				switch(Math.abs(contextMap[y-1][x])){
				case 0:
					contextMap[y-1][x] = CP ? (byte) 1: (byte) -1;
					break;
				case 1:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -2: (byte) 2;
					break;
				case 3:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -4: (byte) 4;
					break;
				case 4:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -5: (byte) 5;
					break;
				case 6:
					contextMap[y-1][x] = contextMap[y-1][x] < 0 ? (byte) -7: (byte) 7;
					break;
				}
			}
			break;
		}
	}

	/**
	 * @return blockSamples definition in {@link #blockSamples}
	 */
	public float[][] getBlockSamples(){
		return(blockSamples);
	}
}

