/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode;
import BOIException.ErrorException;
import BOIException.ParameterException;
import BOIStream.ByteStream;
import BOIDecode.FileRead.Deheading;
import BOIDecode.FileRead.FileRead;
import BOIDecode.FileRead.JP2Deheading;
import BOIDecode.FileRead.JPCDeheading;
import BOIDecode.FileRead.JPKDeheading;
import BOIDecode.FileRead.PrecinctDivision;
import BOIDecode.Decode.BlockBuild;
import BOIDecode.Decode.BlockDecode;
import BOIDecode.Detransform.ColourDetransform;
import BOIDecode.Detransform.Dequantization;
import BOIDecode.Detransform.ImageBuild;
import BOIDecode.Detransform.LevelUnshift;
import BOIDecode.Detransform.RangeCheck;
import BOIDecode.Detransform.WaveletDetransform;

import java.io.RandomAccessFile;
import java.io.IOException;


/**
 * Main class of BOIDecoder application. It receives a BOI compressed bitstream and recovers original image.
 * Usage example:<br>
 * &nbsp; construct<br>
 * &nbsp; run<br>
 *
 * @author Francesc Auli-Llinas
 * @version 1.0
 */
public class Decoder{

	/**
	 * File where heading information and image data are.
	 * <p>
	 * Filename of the file.
	 */
	String DInFile;

	/**
	 * Header type:<br>
	 *  <ul>
	 *    <li> 0- NOH Empty headings
	 *    <li> 1- JPC JPEG2000 codestream header
	 *    <li> 2- JP2 JPEG2000 codestream header with JP2 box
	 *    <li> 3- JPK BOI JPEG2000 codestream header
	 *  </ul>
	 */
	int DFileHeader;

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift} (see source code)
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 */
	int[] CTComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	int[] WTTypes = null;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	int[] WTLevels = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	int[] QTypes = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QExponents}
	 */
	int[][][] QExponents = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QMantisas}
	 */
	int[][][] QMantisas = null;

	/**
	 * Definition in {@link BOICode.Transform.Quantization} (see source code)
	 */
	int QGuardBits;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	int[] BDBlockWidths = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	int[] BDBlockHeights = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 */
	int[][] BDResolutionPrecinctWidths = null;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 */
	int[][] BDResolutionPrecinctHeights = null;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 */
	int LCAchievedNumLayers;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWProgressionOrder}
	 */
	int FWProgressionOrder;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	boolean[] FWPacketHeaders = null;

	/**
	 * Show some information about decompression process. Each value is boolean (false means NO show, true means show).
	 * <p>
	 * Each value represents the following:<br>
	 *  <ul>
	 *    <li> 0- Information about time and used memory for each compression stage
	 *    <li> 1- BOI parameters read from file
	 *  </ul>
	 */
	boolean[] CVerbose = null;

	/**
	 * Number of threads employed in the block coding stage.
	 * <p>
	 * Positive values
	 */
	int CThreadNumber = -1;

	//INTERNAL VARIABLES

	/**
	 * To know if parameters are set.
	 * <p>
	 * True indicates that they are set otherwise false.
	 */
	boolean parametersSet = false;

	/**
	 * Used for verbose information (time for stage).
	 * <p>
	 * 0 is initial time.
	 */
	long initStageTime = 0;

	/**
	 * Used for verbose information (total time).
	 * <p>
	 * 0 is initial time.
	 */
	long initTime = 0;

	/**
	 * Constructor of BOIDecoder. It receives image file name.
	 *
	 * @param DInFile definition in {@link #DInFile}
	 */
	public Decoder(String DInFile){
		//Parameters copy
		this.DInFile = DInFile;
	}

	/**
	 * Set the parameters used to perform the BOI compression and check validity. If a parameter is not initialized, this class initalizes it with default values.
	 *
	 * @param P_CVerbose definition in {@link #CVerbose}
	 * @param P_CThreadNumber definition in this {@link #CThreadNumber}
	 *
	 * @throws ParameterException when some parameter is not valid
	 */
	public void setParameters(boolean[] P_CVerbose, int P_CThreadNumber) throws ParameterException{
		parametersSet = true;

		//FILE TYPE
		String fileName = DInFile.substring(DInFile.lastIndexOf("/") + 1, DInFile.length()).toUpperCase();
		DFileHeader = 1; //By default .jpc
		if(fileName.indexOf(".JPC") >= 0){
			DFileHeader = 1;
		}else if(fileName.indexOf(".JP2") >= 0){
			DFileHeader = 2;
		}else if(fileName.indexOf(".JPK") >= 0){
			DFileHeader = 3;
		}else{
			throw new ParameterException("File has unrecognized heading.");
		}

		//VERBOSE INFORMATION
		if(P_CVerbose != null){
			CVerbose = P_CVerbose;
		}else{
			CVerbose = DecoderDefaultValues.CVerbose;
		}

		//NUMBER OF THREADS
		if(P_CThreadNumber > 0){
			CThreadNumber = P_CThreadNumber;
		}else{
			CThreadNumber = DecoderDefaultValues.CThreadNumber;
		}
	}

	/**
	 * Runs decoder algorithm to recover the image.
	 *
	 * @return imageSamplesFloat definition in {@link BOICode.Coder#imageSamplesFloat}
	 *
	 * @throws ErrorException when decompression cannot be performed
	 */
	public float[][][] run() throws ErrorException{
		//If parameters are not set run cannot be executed
		if(!parametersSet){
			throw new ErrorException("Parameters not set.");
		}

		//File to read from
		RandomAccessFile inFile;
		//Image samples in different types
		float[][][]       imageSamplesFloat = null;
		float[][][][][][][] imageBlocks = null;
		//Definition in BOICoder/FileWrite/PrecinctBuild
		ByteStream[][][][][][][] PBByteStreamsLayers = null;
		int[][][][][][]          PBMostBitPlanesNull = null;
		int[][][][][][]          PBFirstLayer = null;
		int[][][][][][]          PDCodingPasses = null;
		int[][]                  BDBlocksPerPrecinctWidths = null;
		int[][]                  BDBlocksPerPrecinctHeights = null;
		//Definition in BOICoder/Code/CodeBlockBP
		ByteStream[][][][][] BCByteStreams = null;
		int[][][][][]        BCMSBPlanes = null;
		int[][][][][]        BCCodingPasses = null;
		//Heading
		Deheading Heading = null;

		showTimeMemory("STARTING DECOMPRESSION...");

		//Open file
		try{
			inFile = new RandomAccessFile(DInFile, "r");
		}catch(IOException e){
			throw new ErrorException("I/O error opening file.");
		}

		//READ HEADING
		switch(DFileHeader){
		case 1: //JPC
			Heading = new JPCDeheading(inFile);
			break;
		case 2: //JP2
			JP2Deheading JP2 = new JP2Deheading(inFile);
			JP2.run();
			Heading = new JPCDeheading(inFile);
			break;
		case 3: //BOI
			Heading = new JPKDeheading(inFile);
			break;
		}
		Heading.run();

		//READ PARAMETERS FROM HEADING
		zSize = Heading.getzSize();
		ySize = Heading.getySize();
		xSize = Heading.getxSize();
		LSSignedComponents = Heading.getLSSignedComponents();
		CTType = Heading.getCTType();
		CTComponents = Heading.getCTComponents();
		WTTypes = Heading.getWTTypes();
		WTLevels = Heading.getWTLevels();
		QTypes = Heading.getQTypes();
		QComponentsBits = Heading.getQComponentsBits();
		QExponents = Heading.getQExponents();
		QMantisas = Heading.getQMantisas();
		QGuardBits = Heading.getQGuardBits();
		BDBlockWidths = Heading.getBDBlockWidths();
		BDBlockHeights = Heading.getBDBlockHeights();
		BDResolutionPrecinctWidths = Heading.getBDResolutionPrecinctWidths();
		BDResolutionPrecinctHeights = Heading.getBDResolutionPrecinctHeights();
		LCAchievedNumLayers = Heading.getLCAchievedNumLayers();
		FWProgressionOrder = Heading.getFWProgressionOrder();
		FWPacketHeaders = Heading.getFWPacketHeaders();
		//Free unused memory
		Heading = null;
		//Show statistics
		showTimeMemory("main headers reading");
		showParameters();

		//PRECINCT DIVISION
		PrecinctDivision pd = new PrecinctDivision(zSize, ySize, xSize, WTLevels, BDBlockWidths, BDBlockHeights, BDResolutionPrecinctWidths, BDResolutionPrecinctHeights, LCAchievedNumLayers);
		pd.run();
		PBByteStreamsLayers = pd.getByteStreamsLayers();
		PBMostBitPlanesNull = pd.getMostBitPlanesNull();
		PBFirstLayer = pd.getFirstLayer();
		PDCodingPasses = pd.getCodingPasses();
		BDBlocksPerPrecinctWidths = pd.getBlocksPerPrecinctWidths();
		BDBlocksPerPrecinctHeights = pd.getBlocksPerPrecinctHeights();
		//Free unused memory
		pd = null;
		//Show statistics
		showTimeMemory("precinct division");

		//FILE READ
		FileRead fr = new FileRead(zSize, ySize, xSize,  BDBlockWidths, BDBlockHeights, BDBlocksPerPrecinctWidths, BDBlocksPerPrecinctHeights, FWProgressionOrder, FWPacketHeaders, PBByteStreamsLayers, PBMostBitPlanesNull, PBFirstLayer, PDCodingPasses, inFile);
		fr.run();
		PBByteStreamsLayers = fr.getPBByteStreamsLayers();
		PBMostBitPlanesNull = fr.getPBMostBitPlanesNull();
		PBFirstLayer = fr.getPBFirstLayer();
		//Free unused memory and close file
		try{
			inFile.close();
		}catch(IOException e){
			throw new ErrorException("I/O error closing file.");
		}
		fr = null;
		//Show statistics
		showTimeMemory("file read");

		//BLOCK BUILD
		BlockBuild bb = new BlockBuild(zSize, ySize, xSize, WTLevels, BDBlockWidths, BDBlockHeights, BDBlocksPerPrecinctWidths, BDBlocksPerPrecinctHeights, PBByteStreamsLayers, QGuardBits, QExponents, PBMostBitPlanesNull, PBFirstLayer, PDCodingPasses);
		bb.run();
		BCByteStreams = bb.getBCByteStreams();
		BCMSBPlanes = bb.getBCMSBPlanes();
		BCCodingPasses = bb.getBCCodingPasses();
		imageBlocks = bb.getImageBlocks();
		//Free unused memory
		bb = null;
		BDBlocksPerPrecinctWidths = null;
		BDBlocksPerPrecinctHeights = null;
		PBMostBitPlanesNull = null;
		PBFirstLayer = null;
		PDCodingPasses = null;
		//Show statistics
		showTimeMemory("block build");

		//BLOCK DECODE
		BlockDecode.initialize(BCByteStreams, BCMSBPlanes, BCCodingPasses, imageBlocks, WTTypes, CThreadNumber);
		BlockDecode[] bd = new BlockDecode[CThreadNumber];
		for(int thread = 0; thread < CThreadNumber; thread++){
			bd[thread] = new BlockDecode(thread);
		}
		Thread[] threads = new Thread[CThreadNumber];
		for(int thread = 0; thread < CThreadNumber; thread++){
			threads[thread] = new Thread(bd[thread]);
			threads[thread].start();
		}
		for(int thread = 0; thread < CThreadNumber; thread++){
			try{
				threads[thread].join();
			}catch(InterruptedException e){
				throw new ErrorException(e.getMessage());
			}
		}
		//Free unused memory
		bd = null;
		PBByteStreamsLayers = null;
		BCByteStreams = null;
		BCMSBPlanes = null;
		BCCodingPasses = null;
		//Show statistics
		showTimeMemory("block decode");

		//IMAGE BUILD
		ImageBuild ib = new ImageBuild(zSize, ySize, xSize, imageBlocks);
		imageSamplesFloat = ib.run();
		//Free unused memory
		ib = null;
		imageBlocks = null;
		//Show statistics
		showTimeMemory("image build");

		//DEQUANTIZATION
		Dequantization d = new Dequantization(imageSamplesFloat);
		d.setParameters(WTLevels, QTypes, QComponentsBits, QExponents, QMantisas);
		imageSamplesFloat = d.run();
		//Free unused memory
		d = null;
		//Show statistics
		showTimeMemory("dequantization");

		//DISCRETE WAVELET DETRANSFORM
		WaveletDetransform wd = new WaveletDetransform(imageSamplesFloat);
		wd.setParameters(WTTypes, WTLevels);
		imageSamplesFloat = wd.run();
		//Free unused memory
		wd = null;
		//Show statistics
		showTimeMemory("wavelet detransform");

		//COLOUR DETRANSFORM
		ColourDetransform cd = new ColourDetransform(imageSamplesFloat);
		cd.setParameters(CTType, CTComponents);
		imageSamplesFloat = cd.run();
		//Free unused memory
		cd = null;
		//Show statistics
		showTimeMemory("colour detransform");

		//LEVEL UNSHIFTING
		LevelUnshift lu = new LevelUnshift(imageSamplesFloat);
		lu.setParameters(1, QComponentsBits);
		imageSamplesFloat = lu.run();
		//Free unused memory
		lu = null;
		//Show statistics
		showTimeMemory("level unshifting");

		//RANGE CHECK
		RangeCheck rc = new RangeCheck(imageSamplesFloat);
		rc.setParameters(1, LSSignedComponents, QComponentsBits);
		imageSamplesFloat = rc.run();
		//Free unused memory
		rc = null;
		//Show statistics
		showTimeMemory("range check");

		//return recovered image samples
		return(imageSamplesFloat);
	}

	/**
	 * @return QExponents definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	public int[] getQComponentsBits(){
		return(QComponentsBits);
	}

	//////////////////////////////////////////////
	//FUNCTIONS TO SHOW SOME VERBOSE INFORMATION//
	//////////////////////////////////////////////

	/**
	 * Show some time and memory usage statisticals.
	 *
	 * @param stage string that will be displayed
	 */
	void showTimeMemory(String stage){
		if(CVerbose[0]){
			long actualTime = System.currentTimeMillis();
			if(initTime == 0){
				initTime = actualTime;
			}
			if(initStageTime == 0){
				initStageTime = actualTime;
			}

			//print times are not considered
			long totalMemory = Runtime.getRuntime().totalMemory() / 1048576;
			long usedMemory =  (Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory()) / 1048576;
			//((float) initTime - (float) actualTime)
			String durationStage = Float.toString((actualTime - initStageTime) / 1000F) + "000";
			durationStage = durationStage.substring(0, durationStage.lastIndexOf(".") + 4);
			String duration = Float.toString((actualTime - initTime) / 1000F) + "000";
			duration = duration.substring(0, duration.lastIndexOf(".") + 4);

			System.out.println("STAGE: " + stage);
			System.out.println("  Memory (USED/TOTAL): " + usedMemory + "/" + totalMemory + " MB");
			System.out.println("  Time (USED/TOTAL)  : " + durationStage + "/" + duration + " secs");

			initStageTime = System.currentTimeMillis();
		}
	}

	/**
	 * Show BOI image parameters read from file.
	 */
	void showParameters(){
		if(CVerbose[1]){
			System.out.print("BOI PARAMETERS");

			System.out.print("\n  zSize: " + zSize);
			System.out.print("\n  ySize: " + ySize);
			System.out.print("\n  xSize: " + xSize);

			System.out.print("\n  LSSignedComponents: ");
			for(int z = 0; z < LSSignedComponents.length; z++){
				System.out.print(LSSignedComponents[z] + " ");
			}

			System.out.print("\n  CTType: " + CTType);
			System.out.print("\n  CTComponents: ");
			for(int i = 0; i < CTComponents.length; i++){
				System.out.print(CTComponents[i] + " ");
			}

			System.out.print("\n  WTTypes: ");
			for(int z = 0; z < WTTypes.length; z++){
				System.out.print(WTTypes[z] + " ");
			}
			System.out.print("\n  WTLevels: ");
			for(int z = 0; z < WTLevels.length; z++){
				System.out.print(WTLevels[z] +  " ");
			}

			System.out.print("\n  QTypes: ");
			for(int z = 0; z < QTypes.length; z++){
				System.out.print(QTypes[z] + " ");
			}
			System.out.print("\n  QComponentsBits: ");
			for(int z = 0; z < QComponentsBits.length; z++){
				System.out.print(QComponentsBits[z] + " ");
			}
			for(int z = 0; z < QExponents.length; z++){
				System.out.print("\n  QExponents[" + z + "]: ");
			for(int rLevel = 0; rLevel < QExponents[z].length; rLevel++){
				System.out.print("{");
			for(int subband = 0; subband < QExponents[z][rLevel].length; subband++){
				System.out.print(QExponents[z][rLevel][subband] + ",");
			}
				System.out.print("\b} ");
			}}
			for(int z = 0; z < QMantisas.length; z++){
				System.out.print("\n  QMantisas[" + z + "]: ");
			for(int rLevel = 0; rLevel < QMantisas[z].length; rLevel++){
				System.out.print("{");
			for(int subband = 0; subband < QMantisas[z][rLevel].length; subband++){
				System.out.print(QMantisas[z][rLevel][subband] + ",");
			}
				System.out.print("\b} ");
			}}
			System.out.print("\n  QGuardBits: " + QGuardBits);

			System.out.print("\n  BDBlockWidths: ");
			for(int z = 0; z < BDBlockWidths.length; z++){
				System.out.print(BDBlockWidths[z] + " ");
			}
			System.out.print("\n  BDBlockHeights: ");
			for(int z = 0; z < BDBlockHeights.length; z++){
				System.out.print(BDBlockHeights[z] + " ");
			}

			for(int z = 0; z < BDResolutionPrecinctWidths.length; z++){
				System.out.print("\n  BDResolutionPrecinctWidths[" + z + "]: ");
			for(int rLevel = 0; rLevel < BDResolutionPrecinctWidths[z].length; rLevel++){
					System.out.print("{" + BDResolutionPrecinctWidths[z][rLevel] + "} ");
			}}

			for(int z = 0; z < BDResolutionPrecinctHeights.length; z++){
				System.out.print("\n  BDResolutionPrecinctHeights[" + z + "]: ");
			for(int rLevel = 0; rLevel < BDResolutionPrecinctHeights[z].length; rLevel++){
				System.out.print("{" + BDResolutionPrecinctHeights[z][rLevel] + "} ");
			}}

			System.out.print("\n  LCAchievedNumLayers: " + LCAchievedNumLayers);
			System.out.print("\n  FWProgressionOrder: " + FWProgressionOrder);
			System.out.print("\n  FWHeader: " + DFileHeader);
			System.out.print("\n  FWPacketHeaders: " + FWPacketHeaders[0] + " " + FWPacketHeaders[1]);
			System.out.print("\n");
		}
	}

}

