/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode.FileRead;
import BOIException.ErrorException;

import java.io.RandomAccessFile;
import java.util.Arrays;


/**
 * This class finds out the JPEG2000 parameters from the J2C headers.
 *
 * @author Joan Bartrina-Rapesta, Francesc Auli-Llinas
 * @version 1.0
 */
public class JP2Deheading{

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift} (see source code)
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;

	/**
	 * Definition in {@link BOIDecode.Decoder#DInFile}
	 */
	RandomAccessFile DInFile;

	//INTERNAL VARIABLES OF THE CLASS

	/**
	 * Boolean indicating if there are components with different bit depths from others
	 */
	boolean bppVariation;


	/**
	 * Constructor.
	 *
	 * @param DInFile definition in {@link BOIDecode.Decoder#DInFile}
	 */
	public JP2Deheading(RandomAccessFile DInFile){
		this.DInFile = DInFile;
	}

	/**
	 * Reads the boxes of the JP2 headers and checks their conformity.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	public void run() throws ErrorException{
		try{
		//Reads JPEG2000 Signature Box
		readSignatureBox();
		//Reads File Type Box
		readFileTypeBox();
		//Reads JP2 Header SuperBox
		readJP2HeaderSuperBox();
		//Reads Contiguous Codestream Box (only length and type; content is J2C)
		readContiguousCodestreamBox();
		}catch(Exception e){
			throw new ErrorException("Error reading JP2 Headers: " + e.toString());
		}
	}

	/**
	 * Reads the Signature Box and checks its conformity.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readSignatureBox() throws ErrorException{
		int SBLength = 12;
		boolean conformity = true;
		byte[] SignatureBox = new byte[SBLength];
		try{
		DInFile.read(SignatureBox);
		if (SignatureBox[0] != (byte)0x00){
			conformity = false;
		}else if (SignatureBox[1] != (byte)0x00){
			conformity = false;
		}else if (SignatureBox[2] != (byte)0x00){
			conformity = false;
		}else if (SignatureBox[3] != (byte)0x0C){
			conformity = false;
		}else if (SignatureBox[4] != (byte)0x6A){
			conformity = false;
		}else if (SignatureBox[5] != (byte)0x50){
			conformity = false;
		}else if (SignatureBox[6] != (byte)0x20){
			conformity = false;
		}else if (SignatureBox[7] != (byte)0x20){
			conformity = false;
		}else if (SignatureBox[8] != (byte)0x0D){
			conformity = false;
		}else if (SignatureBox[9] != (byte)0x0A){
			conformity = false;
		}else if (SignatureBox[10] != (byte)0x87){
			conformity = false;
		}else if (SignatureBox[11] != (byte)0x0A){
			conformity = false;
		}
		if (!conformity){
			throw new ErrorException("JPEG2000 Signature Box does not conform the standard.");
		}
		}catch(Exception e){
			throw new ErrorException("I/O Error: " + e);
		}
	}

	/**
	 * Reads the File Type Box and checks its conformity.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readFileTypeBox() throws ErrorException{
		boolean conformity = true;
		boolean compatibility = false;
		//length
		byte[] FTBLengthByte = new byte[4];
		try{
		DInFile.read(FTBLengthByte);
		int FTBLength = byteArray2int(FTBLengthByte);
		if ((FTBLength < 20) || (Math.IEEEremainder((double)(FTBLength-20), 4) != 0)){
			conformity = false;
		}
		else{
			//type
			byte[] FTBType = new byte[4];
			DInFile.read(FTBType);
			if (FTBType[0] != (byte)0x66){
				conformity = false;
			}else if (FTBType[1] != (byte)0x74){
				conformity = false;
			}else if (FTBType[2] != (byte)0x79){
				conformity = false;
			}else if (FTBType[3] != (byte)0x70){
				conformity = false;
			}
			//contents
			//brand
			byte[] FTBBrand = new byte[4];
			DInFile.read(FTBBrand);
			if (FTBBrand[0] != (byte)0x6A){
				conformity = false;
			}else if (FTBBrand[1] != (byte)0x70){
				conformity = false;
			}else if (FTBBrand[2] != (byte)0x32){
				conformity = false;
			}else if (FTBBrand[3] != (byte)0x20){
				conformity = false;
			}
			//minor version
			byte[] FTBMinVByte = new byte[4];
			DInFile.read(FTBMinVByte);
			int FTBMinV = byteArray2int(FTBMinVByte);
			if (FTBMinV > 0){
				conformity = false;
			}
			//compatibility list
			byte[] FTBCL = new byte[FTBLength-16];
			DInFile.read(FTBCL);
			int FTBCLitems = ((FTBLength - 16) / 4);
			for (int i=0; i<FTBCLitems; i++){
				if ((FTBCL[i] == (byte)0x6A) && (FTBCL[i+1] == (byte)0x70) && (FTBCL[i+2] == (byte)0x32) && (FTBCL[i+3] == (byte)0x20)){
					compatibility = true;
				}
			}
		}
		if ((!conformity) || (!compatibility)){
			throw new ErrorException("File Type Box does not conform the standard.");
		}
		}catch(Exception e){
			throw new ErrorException("I/O Error: " + e);
		}
	}

	/**
	 * Reads the Image Header Box and retrieves the parameters contained.
	 *
	 * @return bppVariation - definition in {@link #bppVariation}
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	boolean readImageHeaderBox() throws ErrorException{
		boolean conformity = true;
		bppVariation = false;
		//length
		byte[] IHBLengthByte = new byte[4];
		try{
		DInFile.read(IHBLengthByte);
		int IHBLength = byteArray2int(IHBLengthByte);
		if (IHBLength != 22){
			throw new ErrorException("Image Header Box parameters cannot be read.");
		}
		else{
			//type
			byte[] IHBType = new byte[4];
			DInFile.read(IHBType);
			if (IHBType[0] != (byte)0x69){
				conformity = false;
			}else if (IHBType[1] != (byte)0x68){
				conformity = false;
			}else if (IHBType[2] != (byte)0x64){
				conformity = false;
			}else if (IHBType[3] != (byte)0x72){
				conformity = false;
			}
			//contents
			byte[] height = new byte[4];
			DInFile.read(height);
			ySize = byteArray2int(height);
			byte[] width = new byte[4];
			DInFile.read(width);
			xSize = byteArray2int(width);
			byte[] NC = new byte[2];
			DInFile.read(NC);
			zSize = byteArray2int(NC);
			LSSignedComponents = new boolean[zSize];//initialization
			QComponentsBits = new int[zSize];//initialization
			byte[] BPC = new byte[1];
			DInFile.read(BPC);
			if (BPC[0] == (byte)0xFF){
				bppVariation = true;
			}
			else{
				int bitDepth = (BPC[0] & (byte)0x7F) + 1;
				for (int z=0; z<zSize; z++){
					QComponentsBits[z] = bitDepth;
				}
				boolean signedComponents = ((BPC[0] & (byte)0x80) > 0);
				for (int z=0; z<zSize; z++){
					LSSignedComponents[z] = signedComponents;
				}
			}
			int C = DInFile.read();
			if (C != 7){
				conformity = false;
			}
			byte[] UnkC = new byte[1];
			DInFile.read(UnkC);
			if ((UnkC[0] != (byte)0x00) && (UnkC[0] != (byte)0x01)){//An unknown colourspace is treated as if the colourspace interpretation method do accurately reproduce de image
				conformity = false;
			}
			byte[] IPR = new byte[1];
			DInFile.read(IPR);
			if (IPR[0] == (byte)0x01){
				throw new ErrorException("BOI cannot decode an image with intellectual property rights.");
			}else if (IPR[0] != (byte)0x00){
				conformity = false;
			}
		}
		if (!conformity){
			throw new ErrorException("Image Header Box does not conform the standard.");
		}
		}catch(Exception e){
			throw new ErrorException("I/O Error: " + e);
		}

		return(bppVariation);
	}

	/**
	 * Reads the Bits Per Component Box and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readBitsPerComponentBox() throws ErrorException{
		boolean conformity = true;
		//length
		byte[] BPCBLengthByte = new byte[4];
		try{
		DInFile.read(BPCBLengthByte);
		int BPCBLength = byteArray2int(BPCBLengthByte);
		if (BPCBLength != (8+zSize)){
			throw new ErrorException("Bits Per Component Box parameters cannot be read.");
		}
		else{
			//type
			byte[] BPCBType = new byte[4];
			DInFile.read(BPCBType);
			if (BPCBType[0] != (byte)0x62){
				conformity = false;
			}else if (BPCBType[1] != (byte)0x70){
				conformity = false;
			}else if (BPCBType[2] != (byte)0x63){
				conformity = false;
			}else if (BPCBType[3] != (byte)0x63){
				conformity = false;
			}
			//contents
			byte[] BPC = new byte[zSize];
			DInFile.read(BPC);
			for (int z=0; z<BPC.length; z++){
				QComponentsBits[z] = (BPC[z] & (byte)0x7F) + 1;
				LSSignedComponents[z] = ((BPC[z] & (byte)0x80) > 0);
			}
		}
		if (!conformity){
			throw new ErrorException("Bits Per Component Box does not conform the standard.");
		}
		}catch(Exception e){
			throw new ErrorException("I/O Error: " + e);
		}
	}

	/**
	 * Reads the Colour Specification Box and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readColourSpecificationBox() throws ErrorException{
		boolean conformity = true;
		//length
		byte[] CSBLengthByte = new byte[4];
		try{
		DInFile.read(CSBLengthByte);
		int CSBLength = byteArray2int(CSBLengthByte);
		//type
		byte[] CSBType = new byte[4];
		DInFile.read(CSBType);
		if (CSBType[0] != (byte)0x63){
			conformity = false;
		}else if (CSBType[1] != (byte)0x6F){
			conformity = false;
		}else if (CSBType[2] != (byte)0x6C){
			conformity = false;
		}else if (CSBType[3] != (byte)0x72){
			conformity = false;
		}
		//contents
		byte[] METHByte = new byte[1];
		DInFile.read(METHByte);
		int METH = byteArray2int(METHByte);
		if (METH != 1){
			throw new ErrorException("Colour specification method must be 'Enumerated Colourspace'");
		}
		else if (CSBLength != 15){
			throw new ErrorException("Bits Per Component Box does not conform the standard.");
		}
		byte[] PREC = new byte[1];
		DInFile.read(PREC);//ignored
		byte[] APPROX = new byte[1];
		DInFile.read(APPROX);//ignored
		byte[] EnumCSByte = new byte[4];
		DInFile.read(EnumCSByte);
		int EnumCS = byteArray2int(EnumCSByte);
		if ((EnumCS != 16) && (EnumCS != 17)){
			throw new ErrorException("EnumCS must be 16 for sRGB or 17 for greyscale.");
		}
		if (!conformity){
			throw new ErrorException("Bits Per Component Box does not conform the standard.");
		}
		}catch(Exception e){
			throw new ErrorException("I/O Error: " + e);
		}
	}

	/**
	 * Reads the JP2 Header SuperBox and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readJP2HeaderSuperBox() throws ErrorException{
		boolean JP2HSBFound = false;
		boolean conformity = true;
		byte[] JP2HSBType = new byte[4];
		JP2HSBType[0] = (byte)0x6A;
		JP2HSBType[1] = (byte)0x70;
		JP2HSBType[2] = (byte)0x32;
		JP2HSBType[3] = (byte)0x68;
		byte[] type = new byte[4];
		try{
		long length = DInFile.length() - DInFile.getFilePointer();
		long index = 0;
		while ((!JP2HSBFound) && (index < length)){
			DInFile.read(type);
			index += 4;
			JP2HSBFound = Arrays.equals(type, JP2HSBType);
		}
		if (index >= length){
			throw new ErrorException("No JP2 Header SuperBox found.");
		}
		else{
			long position = DInFile.getFilePointer();
			DInFile.seek(position-8);
			//length
			byte[] JP2HSBLengthByte = new byte[4];
			DInFile.read(JP2HSBLengthByte);
			int JP2HSBLength = byteArray2int(JP2HSBLengthByte);
			//type (already read)
			DInFile.read(type);
			//contents
			//Image Header Box
			bppVariation = readImageHeaderBox();
			//Bits Per Component Box
			if (bppVariation){
				readBitsPerComponentBox();
			}
			//Colour Specification Box
			readColourSpecificationBox();
		}
		}catch(Exception e){
			throw new ErrorException(e.toString());
		}
	 }

	/**
	 * Reads the Contiguous Codestream Box and checks its conformity.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	void readContiguousCodestreamBox() throws ErrorException{
		boolean CCBFound = false;
		boolean conformity = true;
		byte[] CCBType = new byte[4];
		CCBType[0] = (byte)0x6A;
		CCBType[1] = (byte)0x70;
		CCBType[2] = (byte)0x32;
		CCBType[3] = (byte)0x63;
		byte[] type = new byte[4];
		try{
		long length = DInFile.length() - DInFile.getFilePointer();
		long index = 0;
		while ((!CCBFound) && (index < length)){
			DInFile.read(type);
			index += 4;
			CCBFound = Arrays.equals(type, CCBType);
		}
		if (index >= length){
			throw new ErrorException("No Contiguous Codestream Box found.");
		}
		}catch(Exception e){
			throw new ErrorException(e.toString());
		}
	 }

	/**
	 * Converts a byte array with length 1 to 4 to an integer
	 *
	 * @param byteArray array to be converted
	 * @return returns the integer value
	 */
	int byteArray2int(byte[] byteArray){
		int value = 0;
		for (int b=0; b<byteArray.length; b++){
			value += (((int)byteArray[b] + ((int)byteArray[b] < 0 ? 256: 0)) << (8*(byteArray.length-1-b)));
		}
		return(value);
	}

}