/**
 * BOI Software - Set of imaging libraries with support for JPEG2000 (ISO/IEC 15444-1)
 * Copyright (C) 2010 - Francesc Auli-Llinas
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
package BOIDecode.FileRead;
import BOIException.ErrorException;
import BOIException.WarningException;
import BOIStream.BitStream;

import java.io.RandomAccessFile;
import java.io.IOException;


/**
 * This class finds out the JPEG2000 parameters from the J2C headers.
 *
 * @author Joan Bartrina-Rapesta, Francesc Auli-Llinas
 * @version 1.1
 */
public class JPCDeheading extends Deheading{

	/**
	 * Definition in {@link BOICode.Coder#zSize}
	 */
	int zSize;
	final int zSize_MAX = 16384;
	final int zSize_BITS = 16;

	/**
	 * Definition in {@link BOICode.Coder#ySize}
	 */
	int ySize;
	final int ySize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#xSize}
	 */
	int xSize;
	final int xSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#YTOsize}
	 */
	int YTOsize;
	final int YTOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#XTOsize}
	 */
	int XTOsize;
	final int XTOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#YOsize}
	 */
	int YOsize;
	final int YOsize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#XOsize}
	 */
	int XOsize;
	final int XOsize_BITS = 32;

	/**
	* Definition in {@link BOICode.Coder#TileYSize}
	 */
	int TileYSize;
	final int TileYSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Coder#TileXSize}
	 */
	int TileXSize;
	final int TileXSize_BITS = 32;

	/**
	 * Definition in {@link BOICode.Transform.LevelShift}
	 */
	boolean[] LSSignedComponents = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	int CTType;
	final int CTType_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	int[] WTTypes = null;
	final int WTTypes_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	int[] WTLevels = null;
	final int WTLevels_MAX = 32;
	final int WTLevels_BITS = 8;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	int[] QTypes = null;
	final int QTypes_BITS = 5;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	int[] QComponentsBits = null;
	final int QComponentsBits_BITS = 7;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QExponents}
	 */
	int[][][] QExponents = null;
	final int QExponents_BITS = 5;

	/**
	 * Definition in {@link BOICode.Transform.Quantization#QMantisas}
	 */
	int[][][] QMantisas = null;
	final int QMantisas_BITS = 11;

	/**
	 * Definition in {@link BOICode.Transform.Quantization} (see source code)
	 */
	int QGuardBits;
	final int QGuardBits_BITS = 3;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	int[] BDBlockHeights = null;
	final int BDBlockHeights_MAX = 10;
	final int BDBlockHeights_MIN = 2;
	final int BDBlockHeights_BITS = 8;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	int[] BDBlockWidths = null;
	final int BDBlockWidths_MAX = 10;
	final int BDBlockWidths_MIN = 2;
	final int BDBlockWidths_BITS = 8;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 */
	int[][] BDResolutionPrecinctHeights = null;
	final int BDResolutionPrecinctHeights_BITS = 4;

	/**
	 * Definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 */
	int[][] BDResolutionPrecinctWidths = null;
	final int BDResolutionPrecinctWidths_BITS = 4;


	/**
	 * Definition in {@link BOICode.Code.MQCoder} (see source code)
	 */
	boolean[] MQCFlags = null;

	/**
	 * Definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 */
	int LCAchievedNumLayers;
	final int LCAchievedNumLayers_BITS = 16;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWProgressionOrder}
	 */
	int FWProgressionOrder;
	final int FWProgressionOrder_BITS =  8;

	/**
	 * Definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	boolean[] FWPacketHeaders = null;

	/**
	 * Definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 */
	int[] CTComponents = null;

	/**
	 * Definition in {@link BOIDecode.Decoder#DInFile}
	 */
	RandomAccessFile DInFile = null;

	//INTERNAL VARIABLES OF THE CLASS

	/**
	 * Tells if precincts are defined within the headers or not (1 or 0).
	 * <p>
	 * Index in the array is component index
	 */
	int[] definedPrecincts = null;

	 /** Denotes capabilities that a decoder needs to properly decode the codestream
	 *
	 */
	int Rsiz;
	final int Rsiz_BITS = 16;

	/**
	 * Horitzontal separation of sample of ith component with respect to the reference grid. thre is one occurence of this paramter for each compoenent
	 *
	 */
	int XRsiz;
	final int XRsiz_BITS = 8;

	/**
	 * Horitzontal separation of sample of ith component with respect to the reference grid. thre is one occurence of this paramter for each compoenent
	 *
	 */
	int YRsiz;
	final int YRsiz_BITS = 8;

	/**
	 * Length of Image and Tile size marker (SIZ) segments in bytes
	 *
	 */
	int LSIZ;
	final int LSIZ_MAX = 49190;
	final int LSIZ_BITS = 16;

	/**
	 * Length of Image and Tile size marker (COD) segments in bytes
	 *
	 */
	int LCOD;
	final int LCOD_MAX = 45;
	final int LCOD_BITS = 16;

	/**
	 * Length of Component Image and Tile size marker (COC) segments in bytes
	 *
	 */
	int LCOC;
	int LCOC_MAX = 43;
	final int LCOC_BITS = 16;

	/**
	 * Index of the component to which COC marker segment relates
	 *
	 */
	int CCOC;
	final int CCOC_MAX = 16383;
	final int CCOC_BITS = 16;

	/**
	 * The length of index of the component to which QCD marker segment relates
	 *
	 */
	int LQCD;
	final int LQCD_MAX = 197;
	final int LQCD_BITS = 16;

	/**
	 * The length of index of the component to which QCC marker segment relates
	 *
	 */
	int LQCC;
	final int LQCC_MAX = 199;
	final int LQCC_BITS = 16;

	/**
	 * Index of the component to which QCC marker segment relates
	 *
	 */
	int CQCC;
	final int CQCC_MAX = 16383;
	final int CQCC_BITS = 16;

	/**
	 * Tells if colour transform has been applied.
	 */
	boolean colourTransform = true;

	/**
	 * To know if a marker has been found.
	 */
	boolean markerFound = false;

	/**
	 * To know if a SOT marker.
	 */
	boolean SOTFound = false;

	/**
	 * Temporal bitstream where the values are saved from the file.
	 * <p>
	 * Content is a buffer of bits.
	 */
	BitStream JPCHeading = null;

	/**
	 * Constructor of JPCDeheading. It receives the input file.
	 *
	 * @param DInFile definition in {@link BOIDecode.Decoder#DInFile}
	 */
	public JPCDeheading(RandomAccessFile DInFile){
		this.DInFile = DInFile;
	}

	/**
	 * Reads the J2C file headers and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the header information is ambiguous or incorrect
	 */
	public void run() throws ErrorException{
		JPCHeading = new BitStream();
		markerFound = false;
		readSOC();
		readSIZ();
		while(!SOTFound){
			while(!markerFound){
				markerFound = (readByte() == (byte)0xFF);
			}
			switch(readByte()){
			case (byte)0x52://COD marker
					readCOD();
					break;
			case (byte)0x53://COC marker
					readCOC();
					break;
			case (byte)0x5C://QCD marker
					readQCD();
					break;
			case (byte)0x5D://QCC marker
					readQCC();
					break;
			case (byte)0x90://SOT marker
					SOTFound = true;
					break;
			case (byte)0x5E://RGN marker
					throw new ErrorException("BOI cannot decode codestreams with a RGN marker.");
			case (byte)0x5F://POC marker
					throw new ErrorException("BOI cannot decode codestreams with a POC marker.");
			case (byte)0x55://TLM marker
					throw new ErrorException("BOI cannot decode codestreams with a TLM marker.");
			case (byte)0x57://PLM marker
					throw new ErrorException("BOI cannot decode codestreams with a PLM marker.");
			case (byte)0x58://PLT marker
				throw new ErrorException("PLT marker not is allowed in the main header.");
			case (byte)0x60://PPM marker
					throw new ErrorException("BOI cannot decode codestreams with a PPM marker.");
			case (byte)0x61://PPT marker
				throw new ErrorException("PPT marker not is allowed in the main header.");
			case (byte)0x63://CRG marker
				throw new ErrorException("BOI cannot decode codestreams with a CRG marker.");
			case (byte)0x91://SOP marker
				throw new ErrorException("SOP marker not is allowed in the main header.");
			default:
				markerFound = false;
			break;
			}
		}

	}

	/**
	 * Reads the SOC marker.
	 *
	 * @throws ErrorException when the SOC marker is not unique in the codestream
	 */
	void readSOC() throws ErrorException{
		while(!markerFound){
			markerFound = (readByte() == (byte)0xFF);
		}
		if (readByte() != (byte)0x4F){
			throw new ErrorException("First Marker is not a SOC.");
		}
		//End of SOC marker
		markerFound = false;
	}

	/**
	 * Reads the SIZ marker segment and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the SIZ marker or its marker segment is ambiguous or incorrect
	 */
	void readSIZ() throws ErrorException{
		while(!markerFound){
			markerFound = (readByte() == (byte)0xFF);
		}
		if (readByte() != (byte)0x51){
			throw new ErrorException("Second Marker is not a SIZ.");
		}
		//SIZ marker segment's Lsiz
		LSIZ = readInteger(LSIZ_BITS);
		if((LSIZ  < 41) || (LSIZ  > LSIZ_MAX)){
			throw new ErrorException("LSIZ of marker SIZ is wrong.");
		}
		Rsiz = readInteger(Rsiz_BITS);
		if (Rsiz > 0){
			throw new ErrorException("Rsiz parameter of SIZ marker segment must have value 0.");
		}

		xSize = readInteger(xSize_BITS);
		if((xSize < 1) || (xSize > (int)Math.pow(2, xSize_BITS)-1)){
			throw new ErrorException("xSize is wrong.");
		}
		//SIZ marker segment's Ysiz
		ySize = readInteger(ySize_BITS);
		if((ySize < 1) || (ySize > (int)Math.pow(2, ySize_BITS)-1)){
			throw new ErrorException("ySize is wrong.");
		}
		//SIZ marker segment's XOsiz
		XOsize = readInteger(XOsize_BITS);
		if((XOsize < 0) || (XOsize > (int)Math.pow(2, XOsize_BITS)-2)){
			throw new ErrorException("XOsize is wrong.");
		}
		//SIZ marker segment's YOsiz
		YOsize = readInteger(YOsize_BITS);
		if((YOsize < 0) || (YOsize > (int)Math.pow(2, YOsize_BITS)-2)){
			throw new ErrorException("YOsize is wrong.");
		}
		//SIZ marker segment's XTsiz
		TileXSize = readInteger(TileXSize_BITS);
		if((TileXSize < 1) || (TileXSize > (int)Math.pow(2, TileXSize_BITS)-1)){
			throw new ErrorException("TileXSize is wrong.");
		}
		//SIZ marker segment's YTsiz
		TileYSize = readInteger(TileYSize_BITS);
		if((TileYSize < 1) || (TileYSize > (int)Math.pow(2, TileYSize_BITS)-1)){
			throw new ErrorException("TileYSize is wrong.");
		}
		//SIZ marker segment's XTOsiz
		XTOsize = readInteger(XTOsize_BITS);
		if((XTOsize < 0) || (XTOsize > (int)Math.pow(2, XTOsize_BITS)-2)){
			throw new ErrorException("XTOsize is wrong.");
		}
		//SIZ marker segment's YTOsiz
		YTOsize = readInteger(YTOsize_BITS);
		if((YTOsize < 0) || (YTOsize > (int)Math.pow(2, YTOsize_BITS)-2)){
			throw new ErrorException("YTOsize is wrong.");
		}
		//SIZ marker segment's Csiz
		zSize = readInteger(zSize_BITS);
		if((zSize < 1) || (zSize > zSize_MAX)){
			throw new ErrorException("zSize is wrong.");
		}
		if (LSIZ != (38+(3*zSize))){
			throw new ErrorException("Wrong SIZ marker segment: Csiz is "+zSize+"and Lsiz is "+LSIZ+".");
		}
		LSSignedComponents = new boolean[zSize];
		QComponentsBits = new int[zSize];
		for(int z = 0; z < zSize; z++){
				LSSignedComponents[z] = readBoolean();
				QComponentsBits[z]=readInteger(7)+1;
				if( (QComponentsBits[z]-1 < 1) || (QComponentsBits[z]-1 > 38)){
					throw new ErrorException("QComponentsBits is wrong.");
				}
				XRsiz = readInteger(XRsiz_BITS);
				if( (XRsiz < 1) || (XRsiz > (int)Math.pow(2, XRsiz_BITS)-2)){
					throw new ErrorException("XRsiz is wrong.");
				}
				YRsiz = readInteger(YRsiz_BITS);
				if( (YRsiz < 1) || (YRsiz > (int)Math.pow(2, YRsiz_BITS)-2)){
					throw new ErrorException("YRsiz is wrong.");
				}
		}
		//Initializes definedPrecincts
		definedPrecincts = new int[zSize];
		for (int z = 0; z < zSize; z++){
			definedPrecincts[z] = -1;//value not set
		}
		//Initializes WTLevels
		WTLevels = new int[zSize];
		for (int z = 0; z < zSize; z++){
			WTLevels[z] = -1;//value not set
		}
		//Initializes WTTypes
		WTTypes = new int[zSize];
		for (int z = 0; z < zSize; z++){
			WTTypes[z] = -1;//value not set
		}
		//Initializes BDBlock dimensions
		BDBlockWidths = new int[zSize];
		BDBlockHeights = new int[zSize];
		for (int z = 0; z < zSize; z++){
			BDBlockWidths[z] = -1;//value not set
			BDBlockHeights[z] = -1;//value not set
		}
		//Initializes BDResolutionPrecinct dimensions partially
		BDResolutionPrecinctWidths = new int[zSize][];
		BDResolutionPrecinctHeights = new int[zSize][];
		//Initializes QTypes
		QTypes = new int[zSize];
		for (int z = 0; z < zSize; z++){
			QTypes[z] = -1;//value not set
		}
		//Initializes QExponents and QMantisas partially
		QExponents = new int[zSize][][];
		QMantisas = new int[zSize][][];
		//End of SIZ marker
		markerFound = false;

	}

	/**
	 * Reads a COD marker segment and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the COD marker segment is ambiguous or incorrect
	 */
	void readCOD() throws ErrorException{
		//COD marker segment's Lcod
		LCOD = readInteger(LCOD_BITS);
		if((LCOD < 12) || (LCOD  > LCOD_MAX)){
			throw new ErrorException("LCOD of marker COD is wrong.");
		}
		//COD marker segment's Scod
		FWPacketHeaders = new boolean[2];
		if(readInteger(6) == 1){
			FWPacketHeaders[1] = true;
		}else{
			FWPacketHeaders[1] = false;
		}
		//Start of packet header marker
		if(readInteger(1) == 1){
			FWPacketHeaders[0] = true;
		}else{
			FWPacketHeaders[0] = false;
		}
		//Precincts defined below
		definedPrecincts[0] = readInteger(1);
		//Progression Order
		FWProgressionOrder = readInteger(FWProgressionOrder_BITS);
		if((FWProgressionOrder < 0) || (FWProgressionOrder > 4)){
			throw new ErrorException("FWProgressionOrder is wrong.");
		}
		LCAchievedNumLayers = readInteger(LCAchievedNumLayers_BITS);
		if((LCAchievedNumLayers < 1) || (LCAchievedNumLayers > (int)Math.pow(2, LCAchievedNumLayers_BITS)-1)){
			throw new ErrorException("LCAchievedNumLayers is wrong.");
		}
		int colourTransformType = readInteger(CTType_BITS);
		if (colourTransformType > 1){
			throw new ErrorException("Multiple component transformation type does not exist.");
		}else{
			if (colourTransformType == 0){
				colourTransform = false;
			}else{
				if(zSize < 3){
					throw new ErrorException("Colour transform cannot have been applied to only "+zSize+" component(s).");
				}else{
					colourTransform = true;
				}
			}
		}
		int WTlevels = readInteger(WTLevels_BITS);
		if((WTlevels < 0) || (WTlevels > WTLevels_MAX)){
			throw new ErrorException("WTLevels is wrong.");
		}
		for (int z = 0; z < zSize; z++){
			if (WTLevels[z] < 0){//Does not set the value if it was already set by a COC marker
				WTLevels[z] = WTlevels;
				//Finishes the initialization of BDResolutionPrecinct dimensions for those defined by a COD marker
				BDResolutionPrecinctWidths[z] = new int[WTlevels+1];
				BDResolutionPrecinctHeights[z] = new int[WTlevels+1];
				for (int rLevel = 0; rLevel < (WTlevels+1); rLevel++){
					BDResolutionPrecinctWidths[z][rLevel] = -1;
					BDResolutionPrecinctHeights[z][rLevel] = -1;
				}
				//Finishes the initialization of QExponents and Qmantisas for those defined by a QCD marker
				QExponents[z] = new int[WTLevels[z]+1][];
				QMantisas[z] = new int[WTLevels[z]+1][];
				QExponents[z][0] = new int[1];
				QMantisas[z][0] = new int[1];
				QExponents[z][0][0] = -1;
				QMantisas[z][0][0] = -1;
				for(int rLevel = 1; rLevel < (WTLevels[z]+1); rLevel++){
					QExponents[z][rLevel] = new int[3];
					QMantisas[z][rLevel] = new int[3];
					for (int subband = 0; subband < 3; subband++){
						QExponents[z][rLevel][subband] = -1;
						QMantisas[z][rLevel][subband] = -1;
					}
				}
			}
		}
		BDBlockWidths[0] = readInteger(BDBlockWidths_BITS) + 2;
		BDBlockHeights[0] = readInteger(BDBlockHeights_BITS) + 2;
		for (int z = 1; z < zSize; z++){
			BDBlockWidths[z] = BDBlockWidths[0] ;
			BDBlockHeights[z] = BDBlockHeights[0];
		}
		//code-block width and height
		if((BDBlockWidths[0] < BDBlockWidths_MIN) || (BDBlockWidths[0] > BDBlockWidths_MAX) || (BDBlockHeights[0] < BDBlockHeights_MIN) || (BDBlockHeights[0] > BDBlockHeights_MAX) || ((BDBlockWidths[0]+BDBlockHeights[0]) > 12)){
			throw new ErrorException("Wrong block sizes.");
		}
		//Flags MQ

		readInteger(2);//just to seek
		MQCFlags = new boolean[6];
		for(int MQCFlag = MQCFlags.length - 1; MQCFlag >= 0; MQCFlag--){
			MQCFlags[MQCFlag] = readBoolean();
		}
		if((MQCFlags[0] == true) || (MQCFlags[1] == true) || (MQCFlags[3] == true) || (MQCFlags[4] == true) || (MQCFlags[5] == true)){
			throw new ErrorException("BOI can only work with the RESTART mode active an the others desactivated.");
		}
		WTTypes[0] = readInteger(WTTypes_BITS);
		if(WTTypes[0] == 1){
			WTTypes[0] = 1;
		}
		if(WTTypes[0] == 0){
			WTTypes[0] = 2;
		}
		if( ( WTTypes[0] < 1 ) || (WTTypes[0] > 2)){
			throw new ErrorException("WTTypes is wrong.");
		}
		for (int z = 0; z < zSize; z++){
			//Does not set the value if it was already set by a COC marker
			if (WTTypes[z] < 0){
				WTTypes[z] = WTTypes[0];
			}
		}
		//user-defined precincts
		if (definedPrecincts[0] == 1){
			int widthExp, heightExp;
			for (int rLevel = 0; rLevel < (WTlevels+1); rLevel++){
				heightExp = readInteger(BDResolutionPrecinctHeights_BITS);
				if( (heightExp < 0) && (heightExp > Math.pow(2,BDResolutionPrecinctHeights_BITS)-1)){
					throw new ErrorException("BDResolutionPrecinctHeights is wrong.");
				}
				widthExp = readInteger(BDResolutionPrecinctHeights_BITS);
				if( (widthExp < 0) && (widthExp > Math.pow(2,BDResolutionPrecinctWidths_BITS)-1)){
					throw new ErrorException("BDResolutionPrecinctWidths is wrong.");
				}
				for (int z = 0; z < zSize; z++){
					BDResolutionPrecinctHeights[z][rLevel] = heightExp - (rLevel == 0 ? 0: 1);
					BDResolutionPrecinctWidths[z][rLevel] = widthExp - (rLevel == 0 ? 0: 1);
				}
			}
		}else{//maximum precincts
			for (int rLevel = 0; rLevel < WTlevels+1; rLevel++){
			for (int z = 0; z < zSize; z++){
				BDResolutionPrecinctWidths[z][rLevel] = 15 - (rLevel == 0 ? 0: 1);
				BDResolutionPrecinctHeights[z][rLevel] = 15 - (rLevel == 0 ? 0: 1);
			}}
		}
		//End of COD marker
		markerFound = false;
	}

	/**
	 * Reads a COD/COC marker segment and retrieves the parameters contained.
	 *
	 *
	 * @throws ErrorException when the COD/COC marker segment is ambiguous or incorrect
	 */
	void readCOC() throws ErrorException{
		//COC marker segment's Lcoc
		LCOC = readInteger(LCOC_BITS);
		if((LCOC < 12) || (LCOC  > LCOC_MAX)){
			throw new ErrorException("LCOC of marker COC is wrong.");
		}
		//COC marker segment's Ccoc
		if(zSize < 257){
			CCOC= readInteger(CCOC_BITS/2);
		}else{
			CCOC = readInteger(CCOC_BITS);
		}
		if (CCOC > CCOC_MAX){
			throw new ErrorException("Ccoc in COC marker segment must be lower than " + CCOC_MAX+1 +" and is "+CCOC+".");
		}
		//COC marker segment's Scoc
		definedPrecincts[CCOC] = readInteger(8);
		//COC marker segment's SPcoc
		int WTlevels = readInteger(WTLevels_BITS);
		if( (WTlevels < 0) || (WTlevels > WTLevels_MAX)){
			throw new ErrorException("WTLevels is wrong.");
		}
		WTLevels[CCOC] = WTlevels;
		//Initializes BDResolutionPrecinct dimensions for this component
		BDResolutionPrecinctWidths[CCOC] = new int[WTlevels+1];
		BDResolutionPrecinctHeights[CCOC] = new int[WTlevels+1];
		//Initializes QExponents and QMantisas for this component
		QExponents[CCOC] = new int[WTlevels+1][];
		QMantisas[CCOC] = new int[WTlevels+1][];
		QExponents[CCOC][0] = new int[1];
		QMantisas[CCOC][0] = new int[1];
		QExponents[CCOC][0][0] = -1;
		QMantisas[CCOC][0][0] = -1;
		for(int rLevel = 1; rLevel < (WTlevels+1); rLevel++){
			QExponents[CCOC][rLevel] = new int[3];
			QMantisas[CCOC][rLevel] = new int[3];
			for (int subband = 0; subband < 3; subband++){
				QExponents[CCOC][rLevel][subband] = -1;
				QMantisas[CCOC][rLevel][subband] = -1;
			}
		}
		BDBlockWidths[CCOC] = readInteger(BDBlockWidths_BITS) + 2;
		BDBlockHeights[CCOC] = readInteger(BDBlockHeights_BITS) + 2;
		//code-block width and height
		if((BDBlockWidths[CCOC] < BDBlockWidths_MIN) || (BDBlockWidths[CCOC] > BDBlockWidths_MAX) || (BDBlockHeights[CCOC] < BDBlockHeights_MIN) || (BDBlockHeights[CCOC] > BDBlockHeights_MAX) || ((BDBlockWidths[CCOC]+BDBlockHeights[CCOC]) > 12)){
			throw new ErrorException("Wrong block sizes.");
		}
		//Flags MQ
		readInteger(2);//just to seek
		for(int MQCFlag = MQCFlags.length - 1; MQCFlag >= 0; MQCFlag--){
			MQCFlags[MQCFlag] = readBoolean();
		}
		if((MQCFlags[0] == true) || (MQCFlags[1] == true) || (MQCFlags[3] == true) || (MQCFlags[4] == true) || (MQCFlags[5] == true)){
			throw new ErrorException("BOI can only work with the RESTART mode active an the others desactivated.");
		}
		WTTypes[CCOC] = readInteger(WTTypes_BITS);
		if(WTTypes[CCOC] == 1){
			WTTypes[CCOC] = 1;
		}
		if(WTTypes[CCOC] == 0){
			WTTypes[CCOC] = 2;
		}
		if( ( WTTypes[CCOC] < 1 ) || (WTTypes[CCOC] > 2)){
			throw new ErrorException("WTTypes is wrong.");
		}
		if (definedPrecincts[CCOC] == 1){ //user-defined precincts
			int widthExp, heightExp;
			for (int rLevel = 0; rLevel < (WTlevels+1); rLevel++){
				heightExp = readInteger(BDResolutionPrecinctHeights_BITS);
				if( (heightExp < 0) && (heightExp > Math.pow(2,BDResolutionPrecinctHeights_BITS)-1)){
					throw new ErrorException("BDResolutionPrecinctHeights is wrong.");
				}
				widthExp = readInteger(BDResolutionPrecinctHeights_BITS);
				if( (widthExp < 0) && (widthExp > Math.pow(2,BDResolutionPrecinctWidths_BITS)-1)){
					throw new ErrorException("BDResolutionPrecinctWidths is wrong.");
				}
				BDResolutionPrecinctHeights[CCOC][rLevel] = heightExp - (rLevel == 0 ? 0: 1);
				BDResolutionPrecinctWidths[CCOC][rLevel] = widthExp - (rLevel == 0 ? 0: 1);
			}
		}else{//maximum precincts
			for (int rLevel = 0; rLevel < (WTlevels+1); rLevel++){
					BDResolutionPrecinctWidths[CCOC][rLevel] = 15 - (rLevel == 0 ? 0: 1);
					BDResolutionPrecinctHeights[CCOC][rLevel] = 15 - (rLevel == 0 ? 0: 1);
			}
		}
		//End of COC marker
		markerFound = false;
	}

	/**
	 * Reads a QCD/QCC marker segment and retrieves the parameters contained.
	 *
	 * @throws ErrorException when the QCD/QCC marker segment is ambiguous or incorrect
	 */
	void readQCD( ) throws ErrorException{
		//QCD marker segment's Lqcd
		LQCD = readInteger(LQCD_BITS);
		if((LQCD < 4) || (LQCD > LQCD_MAX)){
			throw new ErrorException("LQCD of marker QCD is wrong.");
		}
		if (LQCD == 5){
			for (int z=0; z<zSize; z++){
				//Does not set the value if it was already set by a QCC marker
				if (QTypes[z] < 0){
					//scalar quantization derived
					QTypes[z] = 1;
				}
			}
		}else{
			if (LQCD == 4+(3*WTLevels[0])){
				for (int z=0; z<zSize; z++){
					//Does not set the value if it was already set by a QCC marker
					if (QTypes[z] < 0){
						//no quantization
						QTypes[z] = 0;
					}
				}
			}else{
				if (LQCD == 5+(6*WTLevels[0])){
					for (int z=0; z<zSize; z++){
						//Does not set the value if it was already set by a QCC marker
						if (QTypes[z] < 0){
							//scalar quantization expounded
							QTypes[z] = 2;
						}
					}
				}else{
					throw new ErrorException("Lqcd parameter from QCD marker segment has a wrong value of "+LQCD+".");
				}
			}
		}
		QGuardBits = readInteger(QGuardBits_BITS);
		if (QGuardBits  > ( Math.pow(2,QGuardBits_BITS)-1)){
			throw new ErrorException("QGuardBits of Component 0 is wrong");
		}
		int Sqcd = readInteger(QTypes_BITS);
		if ((Sqcd == 0) && (QTypes[0] != 0)){
			throw new ErrorException("Sqcd parameter tells QType 0 whilst Lqcd tells QType "+QTypes[0]+".");
		}
		if ((Sqcd == 1) && (QTypes[0] != 1)){
			throw new ErrorException("Sqcd parameter tells QType 1 whilst Lqcd tells QType "+QTypes[0]+".");
		}
		if ((Sqcd == 2) && (QTypes[0] != 2)){
			throw new ErrorException("Sqcd parameter tells QType 2 whilst Lqcd tells QType "+QTypes[0]+".");
		}
		if (Sqcd  > 2){
			throw new ErrorException("QType "+ Sqcd +" unrecognized in Sqcd parameter.");
		}
		//QCD marker segment's SPqcd
		int exponent;
		int mantisa;
		switch(QTypes[0]){
		case 0://no quantization
			exponent = readInteger(QExponents_BITS);
			readInteger(3);//just to seek
			for (int z = 0; z < zSize; z++){
				//Does not set the value if it was already set by a QCC marker
				if (QExponents[z][0][0] < 0){
					QExponents[z][0][0] = exponent;
					QMantisas[z][0][0] = 0;
				}
			}

			for (int rLevel = 1; rLevel < (WTLevels[0] + 1); rLevel++){
				for(int subband = 0; subband < 3; subband++){
					exponent = readInteger(QExponents_BITS);
					readInteger(3);//just to seek
					for (int z = 0; z < zSize; z++){
						//Does not set the value if it was already set by a QCC marker
						if (QExponents[z][rLevel][subband] < 0){
							QExponents[z][rLevel][subband] = exponent;
							if ((QExponents[z][rLevel][subband] < 0) || (QExponents[z][rLevel][subband] > Math.pow(2,QExponents_BITS)-1)){
								throw new ErrorException("Invalid exponent.");
							}
							QMantisas[z][rLevel][subband] = 0;
						}
					}
				}
			}
			break;
		case 1://scalar quantization derived
			exponent = readInteger(QExponents_BITS);
			mantisa  = readInteger(QMantisas_BITS);
			for (int z = 0; z < zSize; z++){
				QExponents[z][0][0] = exponent;//LL
				QMantisas[z][0][0] = mantisa;
				if ((QExponents[z][0][0] < 0) || (QExponents[z][0][0] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[z][0][0] < 0) || QMantisas[z][0][0] > Math.pow(2,QMantisas_BITS)-1){
					throw new ErrorException("Invalid exponent or Mantisa.");
				}
				//rLevel = 1
				for(int subband = 0; subband < 3; subband++){
					if(WTLevels[z] >= 1){
						QExponents[z][1][subband] = QExponents[z][0][0];
						QMantisas[z][1][subband] = QMantisas[z][0][0];
					}
					//other rLevels
					if(WTLevels[z] >= 2){
						for(int rLevel = 2; rLevel < (WTLevels[z]+1); rLevel++){
							QExponents[z][rLevel][subband] = QExponents[z][rLevel-1][subband]-1;
							QMantisas[z][rLevel][subband] = QMantisas[z][0][0];
						}
					}
				}
			}
			break;
		case 2://scalar quantization expounded

			for(int rLevel = 0; rLevel <= WTLevels[0]; rLevel++){
				for(int subband = 0; subband < QExponents[0][rLevel].length; subband++){
					QExponents[0][rLevel][subband] = readInteger(QExponents_BITS);
					QMantisas[0][rLevel][subband] = readInteger(QMantisas_BITS);
					if ((QExponents[0][rLevel][subband] < 0) || (QExponents[0][rLevel][subband] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[0][rLevel][subband] < 0) || QMantisas[0][rLevel][subband] > Math.pow(2,QMantisas_BITS)-1){
						throw new ErrorException("Invalid exponent or Mantisa.");
					}

					for (int z = 1; z < zSize; z++){
					for(int rLevel_copy = 0; rLevel_copy <= WTLevels[z]; rLevel_copy++){
					for(int subband_copy = 0; subband_copy < QExponents[z][rLevel_copy].length; subband_copy++){
						QExponents[z][rLevel_copy][subband_copy] = QExponents[0][rLevel_copy][subband_copy];
						QMantisas[z][rLevel_copy][subband_copy] = QMantisas[0][rLevel_copy][subband_copy];
					}}}

				}
			}

			break;
		}
		markerFound = false;
	}

	/**
	 * Reads a QCD/QCC marker segment and retrieves the parameters contained.
	 *
	 *
	 * @throws ErrorException when the QCD/QCC marker segment is ambiguous or incorrect
	 */
	void readQCC() throws ErrorException{
		//QCC marker segment's Lqcd
		LQCC = readInteger(LQCC_BITS);
		if((LQCC < 5) || (LQCC > LQCC_MAX)){
			throw new ErrorException("LQCC of marker QCC is wrong.");
		}
		//QCC marker segment's Cqcc
		if(zSize < 257){
			CQCC = readInteger(CQCC_BITS/2);
		}else{
			CQCC = readInteger(CQCC_BITS);
		}
		if(CQCC > CQCC_MAX){
			throw new ErrorException(CQCC+" components are not allowed (maximum 16383).");
		}
		QGuardBits = readInteger(QGuardBits_BITS);
		if (QGuardBits  > ( Math.pow(2,QGuardBits_BITS)-1)){
			throw new ErrorException("QGuardBits of Component "+ CQCC+ " is wrong");
		}
		QTypes[CQCC] = readInteger(QTypes_BITS);
		if (QTypes[CQCC]  > 2){
			throw new ErrorException("QType "+(QTypes[CQCC])+" unrecognized in Sqcc parameter for component "+CQCC+".");
		}
		//QCC marker segment's SPqcd
		int exponent;
		int mantisa;
		switch(QTypes[CQCC]){
		case 0://no quantization
				for (int rLevel = 0; rLevel < WTLevels[CQCC]+1; rLevel++){
						for(int subband = 0; subband < QExponents[CQCC][rLevel].length; subband++){
						QExponents[CQCC][rLevel][subband] = readInteger(QExponents_BITS);
						if ((QExponents[CQCC][rLevel][subband] < 0) || (QExponents[CQCC][rLevel][subband] > Math.pow(2,QExponents_BITS)-1)){
							throw new ErrorException("Invalid exponent.");
						}
						readInteger(QMantisas_BITS);
						QMantisas[CQCC][rLevel][subband] = 0;
						}
				}
				break;
		case 1://scalar quantization derived
			exponent = readInteger(QExponents_BITS);
			mantisa  = readInteger(QMantisas_BITS);
			QExponents[CQCC][0][0] = exponent;
			QMantisas[CQCC][0][0] = mantisa;
			if ((QExponents[CQCC][0][0] < 0) || (QExponents[CQCC][0][0] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[CQCC][0][0] < 0) || QMantisas[CQCC][0][0] > Math.pow(2,QMantisas_BITS)-1){
				throw new ErrorException("Invalid exponent or Mantisa.");
			}
			//rLevel = 1
			for(int subband = 0; subband < 3; subband++){
				if(WTLevels[CQCC] >= 1){
						QExponents[CQCC][1][subband] = QExponents[CQCC][0][0];
						QMantisas[CQCC][1][subband] = QMantisas[CQCC][0][0];
				}
				//other rLevels
				if(WTLevels[CQCC] >= 2){
					for(int rLevel = 2; rLevel < (WTLevels[CQCC]+1); rLevel++){
								QExponents[CQCC][rLevel][subband] = QExponents[CQCC][rLevel-1][subband]-1;
								QMantisas[CQCC][rLevel][subband] = QMantisas[CQCC][0][0];
					}
				}
			}
			break;
		case 2://scalar quantization expounded
			for (int rLevel = 0; rLevel <= (WTLevels[CQCC]); rLevel++){
				for (int subband = 0; subband < QMantisas[CQCC][rLevel].length; subband++){
					QExponents[CQCC][rLevel][subband] = readInteger(QExponents_BITS);
					QMantisas[CQCC][rLevel][subband] = readInteger(QMantisas_BITS);
					if ((QExponents[CQCC][rLevel][subband] < 0) || (QExponents[CQCC][rLevel][subband] > Math.pow(2,QExponents_BITS)-1) || (QMantisas[CQCC][rLevel][subband] < 0) || QMantisas[CQCC][rLevel][subband] > Math.pow(2,QMantisas_BITS)-1){
						throw new ErrorException("Invalid exponent or Mantisa.");
					}
				}
			}
			break;
		}
		markerFound = false;
	}

	/**
	 * @return zSize definition in {@link BOICode.Coder#zSize}
	 */
	public int getzSize(){
		return(zSize);
	}

	/**
	 * @return ySize definition in {@link BOICode.Coder#ySize}
	 */
	public int getySize(){
		return(ySize);
	}

	/**
	 * @return definition in {@link BOICode.Coder#xSize}
	 */
	public int getxSize(){
		return(xSize);
	}

	/**
	 * @return LSSignedComponents definition in {@link BOICode.Transform.LevelShift} (see source code)
	 */
	public boolean[] getLSSignedComponents(){
		return(LSSignedComponents);
	}

	/**
	 * @return CTType definition in {@link BOICode.Transform.ColourTransform#CTType}
	 */
	public int getCTType(){
		if (!colourTransform){
			//no colour transform
			CTType = 0;
		}else{
			if (WTTypes[0] == 1){
				//reversible colour transform
				CTType = 1;
			}else{
				//irreversible colour transform
				CTType = 2;
			}
		}
		return(CTType);
	}

	/**
	 * @return CTComponents definition in {@link BOICode.Transform.ColourTransform#CTComponents}
	 */
	public int[] getCTComponents(){
		CTComponents = new int[3];
		if (colourTransform){
			for(int z = 0; z < 3; z++){
				//components 0, 1 and 2
				CTComponents[z] = z;
			}
		}
		return(CTComponents);
	}

	/**
	 * @return WTTypes definition in {@link BOICode.Transform.WaveletTransform#WTTypes}
	 */
	public int[] getWTTypes(){
		return(WTTypes);
	}

	/**
	 * @return WTLevels definition in {@link BOICode.Transform.WaveletTransform#WTLevels}
	 */
	public int[] getWTLevels(){
		return(WTLevels);
	}

	/**
	 * @return QTypes definition in {@link BOICode.Transform.Quantization#QTypes}
	 */
	public int[] getQTypes(){
		return(QTypes);
	}

	/**
	 * @return QComponentsBits definition in {@link BOICode.Transform.Quantization#QComponentsBits}
	 */
	public int[] getQComponentsBits(){
		return(QComponentsBits);
	}

	/**
	 * @return QExponents definition in {@link BOICode.Transform.Quantization#QExponents}
	 */
	public int[][][] getQExponents(){
		return(QExponents);
	}

	/**
	 * @return QMantisas ddefinition in {@link BOICode.Transform.Quantization#QMantisas}
	 */
	public int[][][] getQMantisas(){
		return(QMantisas);
	}

	/**
	 * @return QGuardBits definition in {@link BOICode.Transform.Quantization}
	 */
	public int getQGuardBits(){
		return(QGuardBits);
	}

	/**
	 * @return BDBlockWidths definition in {@link BOICode.Code.BlockDivision#BDBlockWidths}
	 */
	public int[] getBDBlockWidths(){
		return(BDBlockWidths);
	}

	/**
	 * @return BDBlockHeights definition in {@link BOICode.Code.BlockDivision#BDBlockHeights}
	 */
	public int[] getBDBlockHeights(){
		return(BDBlockHeights);
	}

	/**
	 * @return BDResolutionPrecinctWidths definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctWidths}
	 */
	public int[][] getBDResolutionPrecinctWidths(){
		return(BDResolutionPrecinctWidths);
	}

	/**
	 * @return BDResolutionPrecinctHeights definition in {@link BOICode.Code.BlockDivision#BDResolutionPrecinctHeights}
	 */
	public int[][] getBDResolutionPrecinctHeights(){
		return(BDResolutionPrecinctHeights);
	}

	/**
	 * @return MQCFlags definition in {@link BOICode.Code.MQCoder} (see source code)
	 */
	public boolean[] getMQCFlags(){
		return(MQCFlags);
	}

	/**
	 * @return LCAchievedNumLayers definition in {@link BOICode.RateDistortion.LayerCalculation#LCAchievedNumLayers}
	 */
	public int getLCAchievedNumLayers(){
		return(LCAchievedNumLayers);
	}

	/**
	 * @return FWProgressionOrder definition in {@link BOICode.FileWrite.FileWrite#FWProgressionOrder}
	 */
	public int getFWProgressionOrder(){
		return(FWProgressionOrder);
	}

	/**
	 * @return FWPacketHeaders definition in {@link BOICode.FileWrite.FileWrite#FWPacketHeaders}
	 */
	public boolean[] getFWPacketHeaders(){
		return(FWPacketHeaders);
	}

	/**
	 * Reads a boolean from the file and returns its value.
	 *
	 * @return the boolean read
	 *
	 * @throws ErrorException when some wrong bitstream or I/O operation occurs
	 */
	boolean readBoolean() throws ErrorException{
		boolean value;
		if(JPCHeading.getNumBits() < 1){
			try{
				JPCHeading.addByte((byte) DInFile.read());
			}catch(IOException e){
				throw new ErrorException("I/O error (" + e.toString() + ").");
			}
		}
		try{
			value = JPCHeading.getBit(0);
			JPCHeading.deleteBeginBit();
		}catch(WarningException e){
			throw new ErrorException("BitStream error.");
		}
		return(value);
	}

	/**
	 * Reads a boolean from the file and returns its value.
	 *
	 * @return the boolean read
	 *
	 * @throws ErrorException when some wrong bitstream or I/O operation occurs
	 */
	byte readByte() throws ErrorException{
		byte value;

		try{
			value = ((byte) DInFile.read());
		}catch(IOException e){
			throw new ErrorException("I/O error (" + e.toString() + ").");
		}
		return(value);
	}

	/**
	 * Reads an integer from the file and returns its value.
	 *
	 * @param numBits the number of bits of the integer
	 * @return the integer read
	 *
	 * @throws ErrorException when some wrong bitstream or I/O operation occurs
	 */
	int readInteger(int numBits) throws ErrorException{
		int value = 0;
		while(JPCHeading.getNumBits() < numBits){
			try{
				JPCHeading.addByte((byte) DInFile.read());
			}catch(IOException e){
				throw new ErrorException("I/O error (" + e.toString() + ").");
			}
		}
		int mask = (1 << (numBits-1));
		for(int bit = 0; bit < numBits; bit++){
			try{
				if(JPCHeading.getBit(bit)){
					value |= Math.abs(mask);
				}
			}catch(WarningException e){
				throw new ErrorException("BitStream reaches end of buffer.");
			}
					mask >>= 1;
		}
		value = Math.abs(value);
		try{
			JPCHeading.deleteBeginBits(numBits);
		}catch(WarningException e){
			throw new ErrorException("BitStream error.");
		}
		return(value);
	}
}